Add LoadPath "~/ba/casestudy".
Require Export stlc_num.


Definition value {n} (s : tm n) :=
  match s with
  | (lam _ ) => True
  | const _ => True
  | _ => False end.


Reserved Notation  "e1 › e2"  (at level 70).
Inductive step {n} : tm n -> tm n -> Prop :=
| appl (e1 : tm n) e1' e2 : e1 › e1' -> (app e1 e2) › (app e1' e2)
| appr (e1 : tm n) e2' e2 : value e1 -> e2 › e2' -> (app e1 e2) › (app e1 e2')
| plusl (e1 : tm n) e1' e2 : e1 › e1' -> (plus e1 e2) › (plus e1' e2)
| plusr (e1: tm n) e2' e2 : value e1 -> e2 › e2' -> (plus e1 e2) › (plus e1 e2')
| splus n1 n2 : (@plus n (const n1) (const n2)) › (const (n1+n2)) 
| sbeta( e1: tm (S n)) e2 : value e2 -> (app (lam e1) e2) › (subst_tm (e2.: var_tm) e1)
where "e1 › e2" := (step e1 e2).

Inductive star {X} (R : X -> X -> Prop) : X -> X -> Prop :=
| srefl e : star R e e
| strans {e1 e2 e3} : R e1 e2 -> star R e2 e3 -> star R e1 e3.
Notation "e1 ›* e2" := (star step e1 e2) (at level 70).
(*Hint Constructors star.*)

Lemma star_trans {X} (R : X -> X -> Prop) e1 e2 e3 :
  star R e1 e2 -> star R e2 e3 -> star R e1 e3.
Proof.
  induction 1. auto. induction 1; eauto using strans. 
Qed.

Inductive type : Type :=
| int : type
| arrow : type -> type -> type.
Notation "A ⤏ B" := (arrow A B) (at level 70).

Reserved Notation "Γ ⊢ x : A" (at level 70, x at level 99). 
Inductive types {m: nat} : (fin m -> type) -> tm m -> type -> Prop :=
| tvar (Γ : fin m -> type) (x : fin m) : Γ ⊢ (var_tm x) : (Γ x)
| tapp Γ (e1 : tm m) e2 A B :
    Γ ⊢ e1 : (A⤏B) -> Γ ⊢ e2 : A -> Γ ⊢ app e1 e2 :B
| tplus Γ (e1 : tm m) e2 :
    Γ ⊢ e1 : int -> Γ ⊢ e2 : int -> Γ ⊢ plus e1 e2 : int
| tlam Γ e A B :
    @types (S m) (scons A Γ) e B -> Γ ⊢ lam e : (A⤏B)
| tint Γ n : types Γ (const n) int
where "Γ ⊢ x : A" := (types Γ x A). 

Definition ctx m := fin m -> type.

Definition agree {n m} (Γ : ctx n) (Γ' : ctx m) (xi : fin m -> fin n) :=
  forall x,  Γ (xi x) = Γ' x.
Notation "Γ' ≼ Γ ; xi" := (agree Γ Γ' xi) (at level 70).

          

Definition agree_subst {n m} (Γ : ctx n) (Γ' : ctx m) (sigma : fin n -> tm m) :=
  forall (x : fin n) , Γ' ⊢  ( sigma x) : (Γ x).
Notation "Γ' ≼ Γ , σ" := (agree_subst Γ Γ' σ) (at level 70). 


(*TODO name induction arguments*)


Lemma morphism_ren {m }  (Γ : ctx m)   (s : tm m) A :
  Γ ⊢ s : A -> forall m' Γ' (ξ : fin m -> fin m'),
     (Γ ≼ Γ' ; ξ ) -> Γ' ⊢ ren_tm ξ s : A.
Proof. induction 1; intros; asimpl; try econstructor; eauto.
        + rewrite <- H. econstructor. 
        + apply IHtypes. intros [|]; eauto. apply (H0 f).
Qed.

Theorem morphism_subst {m} (Γ : ctx m) (s : tm m) A :
  Γ ⊢ s : A -> forall m' (σ: fin m -> tm m') Γ',
      (Γ' ≼ Γ , σ) -> Γ' ⊢ (subst_tm σ s) : A.
Proof. induction 1; intros; asimpl in *; try econstructor; eauto.
       + apply IHtypes.
         intros [|]; cbn.
         * eapply morphism_ren. eauto. intros x. eauto.
         * constructor. 
Qed.           
         
Definition empty : fin 0 -> type := fun x => match x with | end.
Notation "∅" := empty.


Lemma preservation e A : ∅ ⊢ e : A -> forall e', e › e' -> ∅ ⊢ e' : A.
  Proof.
    induction 1; intros e' H1; asimpl; inversion H1; subst; try econstructor; eauto.
   + eapply morphism_subst. 
      inversion H; subst. eauto. 
      intros [|]. constructor.  eauto.
  Qed.

Fixpoint V (A : type){n} : tm n -> Prop :=
    match A with
    | int =>
      fun x => match x with
               | const n => True
               | _ => False end
    | arrow A B =>
      fun x => match x with
               | (lam e) => forall v,
                   V A v -> exists v', (subst_tm (scons v var_tm) e) ›* v' /\ V B v'
               | _ => False end end.
  
Definition E (A : type) {n} := fun e => exists v, e ›* v /\ @V A n v.

Definition G {n m} (Γ : ctx n) : (fin n -> tm m) -> Prop :=
  fun σ => forall (x : fin n), V (Γ x) (σ x) .

Definition sem_types {n} (Γ : ctx n) (e : tm n) A :=
  forall m (σ : fin n -> tm m), G Γ σ -> E A (subst_tm σ e).

Notation "Γ |= e : A " := (sem_types Γ e A) (at level 70, e at level 99). 

Lemma val_inclusion {n} e A : @V A n e -> @E A n e.
Proof.
  intros. exists e. split. constructor. eauto.
Qed.

Lemma compat_var {n} (Γ : ctx n) A x : Γ ⊢ var_tm x : A -> Γ |= var_tm x : A. 
Proof.
  intros H m sigma H1.
  apply val_inclusion. asimpl. inversion H; subst. 
  eauto.
Qed.

Lemma val A {n} (v: tm n) :  V A v -> value v.
Proof.
  intros H. destruct A; destruct v; try contradiction H; constructor.
Qed.   

Lemma compat_lam  {n} (Γ : ctx n) A B e : A.:Γ |= e : B -> Γ |= lam e : (A ⤏ B).
Proof.
  intros H m σ H1. 
  asimpl. apply val_inclusion.
  intros v H2. asimpl.
  assert (H3 : G (A.:Γ) (v .: σ) ).
  { intros [|]; eauto. } 
  destruct (H _ _ H3). eauto. 
Qed. 

Lemma star_appl {n} {e1 : tm n} {e1' e2} :
  e1 ›* e1' -> (app e1 e2) ›* (app e1' e2).
Proof. induction 1. constructor. eauto using strans, appl.
Qed.

Lemma star_appr {n} {e1 : tm n} {e2' e2} :
  value e1 -> e2 ›* e2' -> (app e1 e2) ›* (app e1 e2').
Proof. intros H. induction 1; eauto using strans, appr. constructor. 
Qed.

Lemma star_sbeta {n} {e1 : tm (S n)} {e2} :
  value e2 -> app (lam e1) e2 ›* (subst_tm (e2.: var_tm) e1).
Proof. intros H. econstructor; eauto using strans, sbeta, srefl.
Qed.

Lemma compat_app {n} (Γ : ctx n) (A B : type) e1 e2 :
  Γ |= e1 : (A⤏B) -> Γ |= e2 : A -> Γ |= app e1 e2 : B. 
Proof.
  intros H H' m sigma H1. asimpl. 
  destruct (H _ sigma H1) as [v1 [H2 H3]].
  destruct (H' _ sigma H1) as [v2 [H4 H5]].
  destruct v1; cbn in H3; try contradiction H3.
  destruct (H3 v2 H5) as [x [H6 H7]].
  exists x. split; eauto.
  eapply star_trans. eauto using star_appl.  
  eapply star_trans. refine (star_appr _ H4).
  eapply (val (arrow A B)). cbn. intros. eauto.
    eauto using star_trans, star_sbeta, val. 
Qed.

Lemma agree_plus {n} {e1 e2 : tm n} m1 m2 :
  e1 ›* const m1 -> e2 ›* const m2 -> plus e1 e2 ›* const (m1+m2).
Proof. remember (const m1) as e3. induction 1; subst.
       remember (const m2) as e4. induction 1; subst.
       * eapply strans; eauto using splus, strans, srefl.
       * eapply strans. eapply plusr. exact I. eauto.
         eauto.
       * eauto using plusl, strans.
Qed.


Lemma compat_plus {n} (Γ : ctx n) e1 e2 :
  Γ |= e1 : int -> Γ |= e2 : int -> Γ |= plus e1 e2 : int. 
Proof.
  intros H H' m sigma H1. asimpl.
  destruct (H _ _ H1) as [x [h0 h]]. destruct (H' _ _ H1) as [y [h0' h']].
  destruct x; cbn in h; try contradiction h.
  destruct y; cbn in h'; try contradiction h'. clear h. clear h'.
  exists (const (n0+n1)).
  split. apply agree_plus; eauto. now cbn. 
Qed.


Lemma sem_soundness {n} (Γ : ctx n) A (e : tm n) : Γ ⊢ e : A -> Γ |= e : A.
Proof.
  induction 1.
  * eapply compat_var. econstructor.
  * eapply compat_app; eauto using IHtypes1, IHtypes2.
  * eapply compat_plus; eauto using IHtypes1, IHtypes2.
  * eapply compat_lam; eauto using IHtypes.
   * intros x sigma H. asimpl. exists (const n). split. econstructor. now cbn.
Qed.

Lemma wn_cbv e A : ∅ ⊢ e : A -> exists v, value v /\ e ›* v.
Proof.
  intros. assert (H' := sem_soundness _ _ _ H).
  enough (E A e).
  * destruct H0 as [x []]. exists x. eauto using val.
  * specialize (H' _ var_tm). asimpl in H'.
    apply H'. intros []. 
Qed.

Reserved Notation " e1 ≻ e2" (at level 70). 
Inductive eval {n} : tm n -> tm n -> Prop :=
| eappl (e1 : tm n) e1' e2 : e1 ≻ e1' -> app e1 e2 ≻ app e1' e2
| eappr (e1 : tm n) e2 e2' : e2 ≻ e2' -> app e1 e2 ≻ app e1 e2'
| elam (e1 : tm (S n)) e2 : e1 ≻ e2 -> lam e1 ≻ lam e2
| eplusl (e1 : tm n) e1' e2 : e1 ≻ e1' -> plus e1 e2 ≻ plus e1' e2
| eplusr (e1 : tm n) e2 e2' : e2 ≻ e2' -> plus e1 e2 ≻ plus e1 e2'
| eplus n1 n2 : @plus n (const n1) (const n2) ≻ const (n1+n2)
| ebeta (e1 : tm (S n)) e2 : app (lam e1) e2 ≻ subst_tm (e2 .: var_tm) e1
    
where "e1 ≻ e2" := (eval e1 e2).

Notation "e1 ≻* e2" := (star eval e1 e2) (at level 70).


Lemma agree_cbv {n} (e1 : tm n) e2 : e1 › e2 -> e1 ≻ e2.
Proof.
  induction 1; try econstructor; eauto. 
Qed.

Lemma agree_cbv' {n} (e1 : tm n) e2 : e1 ›* e2 -> e1 ≻* e2.
Proof.
  induction 1; try econstructor; eauto using agree_cbv, star, eval, step.
  + apply (agree_cbv _ _ H).
  + apply IHstar.
Qed.

    
Lemma wn_lambda (e : tm 0) A : ∅ ⊢ e : A -> exists v, value v /\ e ≻* v.
Proof.
  intros. destruct (wn_cbv e A H) as [x [h1 h2]].
  exists x. split; eauto using agree_cbv'.
Qed.


Inductive SN {n} (R : tm n -> tm n -> Prop) : tm n -> Prop := 
| sn_step (e1 : tm n) : (forall e2, R e1 e2 -> SN R e2) -> SN R e1.
(*
Lemma step_det {n} (e1: tm n) e2 e3 : e1 › e2 -> e1 › e3 -> e2 = e3.
Proof.
  intros H1. revert e3. induction H1; intros e3 H2.
  - induction e1; inversion H2; subst; try (rewrite (IHstep e1'0 H4); reflexivity ); cbn in *; try (contradiction H3).
    (* specialize (IHe1 _ _ H1).*) admit. admit . 
    admit.
  - inversion H2; subst.  erewrite IHstep. *)
    
 (* Lemma sn_cbv e A : ∅ ⊢ e : A -> SN (star step) e.*)

 (* Lemma sn_lambda e A : ∅ ⊢ e : A -> SN (star eval) e.*)
