open nat
run_cmd mk_simp_attr `substLemmasFin
run_cmd mk_simp_attr `defs

def ap {X Y} (f:X → Y) {x y : X} (p : x = y) : f x = f y 
:= by induction p; refl 

def apc {X Y} {f g : X -> Y} {x y : X} (p : f = g) (q : x = y) : f x = g y 
:= by induction p; induction q; refl

@[defs]
def comp {X Y Z} (g: Y → Z) (f : X → Y):= λ (x : X), g (f x)
notation f`>>`g:50 := comp g f

inductive Fin : ℕ → Type
| fz {n} : Fin (succ n)
| fs {n} : Fin n → Fin (succ n)
open Fin 
attribute reducible Fin 

def Fin2 : nat → Type
| 0 := empty
| (n+1) := option (Fin2 n)

@[defs]
def scons  {X : Type} {n : nat} (x : X) (f : (Fin n) → X) (m : Fin (nat.succ n)) : X  
:= match m with 
| fz := x
| (fs x) := (f x) end
attribute [reducible] scons

notation x `.:` f := @scons _ _ x f

def ren (m n : nat) : Type := Fin m → Fin n

def bound {n : nat} : Fin (succ n) := fz 

def null {T} (i : Fin 0) : T := by cases i


def shift {n : nat} : ren n (succ n) := fs

@[defs]
def var_zero {n : nat} : Fin (succ n) := fz

@[defs]
def up_ren {m n} (ξ : ren m n) : ren (succ m) (succ n)  :=  var_zero .: (ξ >> shift)

def idren {k: nat} : ren k k := λ x, x

@[defs]
lemma up_ren_ren {k l m} (ξ : ren k l) (ζ : ren l m) (ρ : ren k m) (E : ∀ x, (ξ >> ζ) x = ρ x) : (∀ x, (up_ren ξ >> up_ren ζ) x = up_ren ρ x) := 
begin 
  intro x, cases x, 
  { refl, }, specialize (E x_a), simp [up_ren, scons,shift,comp], rw ← E, refl   
end

lemma scons_eta {T} {n : nat} (f : Fin (succ n) → T) :
  ((f var_zero) .: (shift >> f)) = f :=
begin
  apply funext, simp [ shift, var_zero], intro x, cases x; refl 
end 

lemma scons_eta' {T} {n:nat} {f f' f'': Fin (succ n) → T} {g : Fin n→ Fin (succ n)} {x : Fin (succ n)}
 (H1 : x = var_zero) (H2 : ∀ x, g x = shift x) (H3 : ∀ x, f'' x = f' x) : (∀ x, f x = f' x)
  → ∀ y, ((f'' x) .: (g >> f)) y = f' y :=
begin intros r, rw H1, simp[shift,var_zero], rw H3, intro x, cases x, refl, simp[scons, comp], rw ← r, rw H2, refl end 

lemma scons_eta_id {n : nat} : (@scons _ n var_zero shift = id) := 
begin 
  apply funext, intro x, cases x; refl 
end 

lemma scons_eta_id' {n : nat} {f : Fin n → Fin (succ n)} {y: Fin (succ n)}
(H1 : y = var_zero) (H2 : ∀ x, f x = shift x) : ∀ x, (y .: f) x = id x :=
begin  intros x, rw H1, cases x, refl, unfold scons, rw (H2 x_a), refl 
end 

lemma scons_comp {T1 T2} {n : nat} (x : T1) (f : Fin n → T1) (g : T1 → T2) : 
  (x .: f) >> g = ((g x) .: (f >> g)) := 
begin apply funext, intro x, cases x; refl end

lemma scons_comp' {T1 T2} {n : nat} {x : T1} {x' : T2} {f : Fin n → T1} {f' : Fin (succ n) → T2} {g : T1 → T2}
 (H1 : ∀ x', ((g x) .: (f >> g)) x' = f' x') : ∀ x', ((x .: f) >> g) x' = f' x':=
begin intros x, cases x, rw ← H1, refl, rw ← H1, refl end

lemma comp' {T1 T2} {n : nat} {f f': Fin n → T1} {g g': T1 → T2} (H1 : ∀ x, f x = f' x) (H2 : ∀ x, g x = g' x) :  ∀ x, (f>>g) x = (f'>>g') x := 
by intros x; simp[comp, (H1 x), (H2 _)]

lemma scons' {X : Type} {n : nat} {x x': X} {f f': (Fin n) → X}
  {m m': Fin (nat.succ n)} (H1 : x = x') (H2 : f = f') : ∀ y, (x .: f) y = (x' .: f') y:=
by intros y; simp[scons, H1,H2]

lemma fCompIdL {X Y} (f: X → Y) : id >> f = f := by refl

lemma fCompIdR {X Y} (f: X → Y) : f >> id = f := by refl 

lemma fId {X} (s : X) : id s = s := by refl

lemma fIdComp {X Y Z} (f: X → Y)(g: Y → Z) : comp g f = f >> g := by refl

lemma fComp {X Y Z P}  (f: X → Y)(g: Y → Z)  (h : Z → P) : (f>>g)>>h = f>>(g>>h) := by refl

lemma fSconsZero {X} {n} (f : Fin n → X) (s : X) : (s.:f) var_zero = s := by refl

lemma fSconsBound {X} {n} (f : Fin n → X) (s : X) : (s.:f) bound = s := by refl

lemma fSconsShift {X} {n} (f : Fin n → X) (s : X) {n} : (s.:f) (shift n) = f n := by refl

lemma boundVarZero {n} : bound = @var_zero n := by refl

lemma fCompIdrL  {k m} (f: ren k m) : idren >> f = f := by refl

lemma fCompIdrR {k m} (f: ren k m) : f >> idren = f := by refl 

lemma fCompScons {X} {n}  (g : Fin n → X) (x) : shift >> (x.:g) = g := by refl

lemma fSconsEta1 {n} {Y}  (g : Fin (nat.succ n) → Y) (f : Fin (nat.succ n) → Y): 
  ((f bound).:(shift >> f)) = f := by rw boundVarZero;  rw scons_eta 

attribute [substLemmasFin] fCompIdL fCompIdR fId fIdComp  
  fComp fSconsZero fSconsBound fSconsShift boundVarZero fCompIdrL fCompIdrR
  fCompScons fSconsEta1

meta def RwLemmasFin := 
  [``(scons_comp), ``(scons_eta_id), ``(scons_eta) ,
  ``(fCompIdL), ``(fCompIdR), ``(fId), --TODO ``(fIdComp), 
  ``(fComp),``(fSconsZero), ``(fSconsBound), ``(fSconsShift),
  ``(boundVarZero), ``(fCompIdL), ``(fCompIdR),``(fCompScons), ``(fSconsEta1)]

def Inst := 1
def Rinst := 2
def Comp := 3 
def Up := 4
def Cons := 5
meta def ex (p:pexpr) := tactic.to_expr p


