import .step
open tm type nat eval


/- Strong Normalization predicate and properties -/

inductive SN {n} (R : tm n -> tm n -> Prop ) : tm n -> Prop  
| sn_step (e1 : tm n) : (forall e2, R e1 e2 -> SN e2) -> SN e1.
/-inductive SN' {n} :  (tm n -> tm n -> Prop )→  tm n -> Prop  
| sn_step (R: tm n → tm n → Prop) (e1 : tm n) : (forall e2, R e1 e2 -> SN' R e2) -> SN' R e1.-- R must be a parameter-/

definition SNe {n} := @SN n eval.

lemma closed_star {n} (s t): s ≻* t -> SNe s -> @SNe n t :=
begin intros h1 h2,  induction h1 with s' s' t' e h3 h4 ih, 
      { aauto },
      { apply ih, cases h2, aauto  }, 
end

lemma SN₀ {n} (x : Fin n) : SNe (var_tm x) := 
begin 
   constructor, intros x h, cases h
end



theorem SN₁ {n} (s t : tm n) : SN eval (app s t) → SN eval s ∧ SN eval t := 
begin 
  intro h₁, split, 
  { generalize h₂ : (app s t) = st, 
  rw h₂ at h₁, revert s t h₂,
  induction h₁ ,  intros, rw ←   h₂ at *, 
  constructor,  intros s' hs', apply (h₁_ih (app s' t)),
  rw ← h₂ , constructor, aauto, refl
  },
  { generalize h₂ : (app s t) = st, revert h₁, rw h₂, intro h₁, revert s t h₂,
  induction h₁ ,  intros, rw ←   h₂ at *, 
  constructor,  intros s' hs', apply (h₁_ih (app s s')), 
  rw ← h₂ , constructor, aauto,  refl
  }
end 

lemma SN₁' {n} (s t : tm n) : SNe (app s t) → SNe s := 
  λ h, begin exact (SN₁ _ _ h).left  end

lemma SN₂ {n m} (s : tm n) (σ  : Fin n → tm m) :
  SNe (s[σ]) → SNe s :=
begin intro h, generalize hs : (s[σ]) = s', revert h, 
  rw hs, intro h, revert n s σ hs, induction h, intros n s σ hs, 
  constructor, intros t ht, cases ht; try
  { aauto , rw ← hs, arw, constructor,
  aauto_w ``(substitutivity) }, 

  { constructor, intros c Hc, cases Hc },
  { apply (h_ih (ht_e₁[ht_e₂..] [σ])); aauto,  rw ← hs, 
  aauto_w ``(substitutivity) , 
} end 

lemma SN₃r {n m} (s : tm n) (ξ : Fin n → Fin m) : 
  SNe (s.⟨ξ⟩) → SNe s :=
begin substify, apply SN₂, aauto  end 


lemma SN₃l' {n m} (s t) (ξ : Fin n → Fin m) :
  s.⟨ξ⟩ ≻ t → ∃ t', t = (t'.⟨ξ⟩) ∧  (s ≻ t') :=
begin revert m ξ t, induction s with _ _ m' s' t' iht ihs
m' s' ihs m m' m s' t' ihs iht; intros m ξ t h,  
  { arw_at `h, cases h, }, 
  { dedup,  arw_at `h, 
  generalize hs : s'.⟨ξ⟩ = s'', rw hs at h,
  generalize ht : t'.⟨ξ⟩ = t'', rw ht at h,
  cases h, any_goals { subst hs; subst ht, },
    { destruct (iht ξ h_e₁' h_a), intros, apply (exists.intro (app w t')), 
      split, {simp with substLemmas, exact h_1.left,}, constructor, exact h_1.right},
    { destruct (ihs ξ h_e₂' h_a), intros w hw, apply (exists.intro (app s' w)), split, 
    { rw hw.left, now_arw,}, constructor, exact hw.right,
  },
    { cases s'; try {simp with substLemmas at hs; exfalso; solve1 { aauto} },
    apply (exists.intro (s'_a[t'..])), subst ht,
    split, simp with substLemmas at hs, rw ← hs, unfold up_ren , arw, refl,
    constructor,
     },}, 
  { simp with substLemmas at *, cases h,
    specialize (ihs (upRen_tm_tm ξ) h_e2 _), cases ihs with w hw, 
    apply (exists.intro (lam w)),  
    split, 
      { rw hw.left, now_arw, }, 
      { constructor, exact hw.right,  }, 
      cases h, aauto
  },
  { arw_at `h, cases h,} , 
  { arw_at `h, generalize hs : s'.⟨ξ⟩ = s'',
  generalize ht : t'.⟨ξ⟩ = t'', rw hs at h, rw ht at h, cases h, 
  all_goals { rw ← hs at*, rw← ht at *, } , 
    { destruct (ihs ξ h_e₁' h_a), intros w hw, apply (exists.intro (plus w t')) , 
    split, { rw hw.left, now_arw, }, constructor, exact hw.right, },
    { destruct (iht ξ h_e₂' h_a), intros w hw, apply (exists.intro (plus s' w)), 
    split, {rw hw.left, now_arw, }, constructor, exact hw.right, }, 
    { cases s'; cases t'; try {simp with substLemmas at *; exfalso; solve1 { aauto }, },
    apply (exists.intro (const (h_n₁+h_n₂))) , 
    split; simp with substLemmas at *, rw ← ht, rw ← hs, constructor, },}, end

lemma SN₃l {n m} (s : tm n) (ξ : Fin n → Fin m) : SNe s → SNe (s.⟨ξ⟩) :=
begin
  intro h, revert m ξ , induction h with s hs ihs, intros, 
  constructor, intros s' hs', 
  have h := SN₃l' s s' ξ hs', cases h with t ht,
  rw ht.left, apply ihs, exact ht.right
end

lemma SN₃ {n m} (s : tm n) (ξ : Fin n → Fin m) : SNe s ↔ SNe (s.⟨ξ⟩) :=
begin 
  split,  aauto_w ``( SN₃l), aauto_w  ``( SN₃r) 
end

lemma SN₄ {n} (s) (t: tm n) : SNe s → SNe t → SNe (s[t..]) → SNe (app (lam s) t) :=
begin 
  intro h, revert t, induction h  with s' hs ihs ,
  intros t ht, revert s hs ihs, induction ht with t' ht' iht, 
  intros s hs ihs hsnt,   constructor, intros e₃ h₃ , 
  cases h₃,  cases h₃_a , 
  { apply ihs, aauto, constructor, apply ht', 
  cases hsnt with x y, apply y, apply substitutivity, aauto, },
  { apply iht; try { assumption }, cases hsnt, 
  apply closed_star, apply subst_eval_head, 
  exact h₃_a, constructor, assumption }, 
  { assumption }
end

lemma SNplus {n} (s t : tm n) : SNe s → SNe t → SNe (plus s t) :=
begin 
  intro h, revert t, 
  induction h with s hs ihs, 
  intros t h', revert s hs ihs, induction h' with t' ht iht,
  intros s' hs ihs, constructor,
  intros s' h, cases h, apply ihs, assumption, 
  constructor, assumption, apply iht; assumption, constructor,
  intros e he, cases he end 

lemma SN_appzero {n} (s : tm n) :
  SNe (app (s.⟨↑⟩) (var_tm var_zero)) → SNe s :=
begin 
  intro h, have h' := h, cases h, apply SN₃r, apply SN₁',aauto
end
