import .fintype
import .wftactic
run_cmd mk_simp_attr `substLemmas

@[substLemmas]
mutual inductive tm,vl 
with tm  : nat -> Type 
  | val : Π (nvl : nat), vl nvl -> tm nvl
  | lam : Π (nvl : nat), tm (nat.succ nvl) -> tm nvl
 with vl  : nat -> Type 
  | var_vl : Π (nvl : nat), Fin nvl -> vl nvl
  | app : Π (nvl : nat), vl nvl -> tm nvl -> vl nvl
open tm vl

@[substLemmas]
lemma congr_val { mvl : nat } { s0 : vl mvl } { t0 : vl mvl } : s0 = t0 -> val mvl s0 = val mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam { mvl : nat } { s0 : tm (nat.succ mvl) } { t0 : tm (nat.succ mvl) } : s0 = t0 -> lam mvl s0 = lam mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_val' { mvl : nat } { s0 : vl mvl } { t0 : vl mvl } (h0 : s0 = t0) : val mvl s0 = val mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam' { mvl : nat } { s0 : tm (nat.succ mvl) } { t0 : tm (nat.succ mvl) } (h0 : s0 = t0) : lam mvl s0 = lam mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_app { mvl : nat } { s0 : vl mvl } { s1 : tm mvl } { t0 : vl mvl } { t1 : tm mvl } : s0 = t0 -> s1 = t1 -> app mvl s0 s1 = app mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_app' { mvl : nat } { s0 : vl mvl } { s1 : tm mvl } { t0 : vl mvl } { t1 : tm mvl } (h0 : s0 = t0) (h1 : s1 = t1) : app mvl s0 s1 = app mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
def upRen_vl_vl { m } { n } (xi : Fin m -> Fin n) : _ :=
  up_ren xi

@[substLemmas]
mutual def ren_tm,ren_vl
with ren_tm : Π { mvl : nat } { nvl : nat } (xivl : Fin mvl -> Fin nvl) (s : tm mvl), tm nvl
    | mvl nvl xivl  (val _ s0) := val nvl (ren_vl xivl s0)
    | mvl nvl xivl  (lam _ s0) := lam nvl (ren_tm (upRen_vl_vl xivl) s0)
 with ren_vl : Π { mvl : nat } { nvl : nat } (xivl : Fin mvl -> Fin nvl) (s : vl mvl), vl nvl
    | mvl nvl xivl  (var_vl _ s) := (var_vl nvl) (xivl s)
    | mvl nvl xivl  (app _ s0 s1) := app nvl (ren_vl xivl s0) (ren_tm xivl s1)
using_well_founded wf_single_arg

@[substLemmas]
def up_vl_vl { m } { nvl : nat } (sigma : Fin m -> vl nvl) : _ :=
  scons ((var_vl (nat.succ nvl)) var_zero) (comp (ren_vl shift) sigma)

@[substLemmas]
mutual def subst_tm,subst_vl
with subst_tm : Π { mvl : nat } { nvl : nat } (sigmavl : Fin mvl -> vl nvl) (s : tm mvl), tm nvl
    | mvl nvl sigmavl  (val _ s0) := val nvl (subst_vl sigmavl s0)
    | mvl nvl sigmavl  (lam _ s0) := lam nvl (subst_tm (up_vl_vl sigmavl) s0)
 with subst_vl : Π { mvl : nat } { nvl : nat } (sigmavl : Fin mvl -> vl nvl) (s : vl mvl), vl nvl
    | mvl nvl sigmavl  (var_vl _ s) := sigmavl s
    | mvl nvl sigmavl  (app _ s0 s1) := app nvl (subst_vl sigmavl s0) (subst_tm sigmavl s1)
using_well_founded wf_single_arg

@[substLemmas]
def upId_vl_vl { mvl : nat } { sigma : Fin mvl -> vl mvl } (Eq : forall x, sigma x = (var_vl mvl) x) : forall x, (up_vl_vl sigma) x = (var_vl (nat.succ mvl)) x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_vl shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
mutual def idSubst_tm,idSubst_vl
with idSubst_tm : Π { mvl : nat } (sigmavl : Fin mvl -> vl mvl) (Eqvl : forall x, sigmavl x = (var_vl mvl) x) (s : tm mvl), subst_tm sigmavl s = s
    | mvl sigmavl Eqvl 
    (val _ s0) := have h : _, from congr_val (idSubst_vl sigmavl Eqvl s0), begin simp with substLemmas at *, assumption end
    | mvl sigmavl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (idSubst_tm (up_vl_vl sigmavl) (upId_vl_vl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with idSubst_vl : Π { mvl : nat } (sigmavl : Fin mvl -> vl mvl) (Eqvl : forall x, sigmavl x = (var_vl mvl) x) (s : vl mvl), subst_vl sigmavl s = s
    | mvl sigmavl Eqvl  (var_vl _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | mvl sigmavl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (idSubst_vl sigmavl Eqvl s0) (idSubst_tm sigmavl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma idSubst'_tm { mvl : nat } { sigmavl : Fin mvl -> vl mvl } { Eqvl : forall x, sigmavl x = (var_vl mvl) x } { s s' : tm mvl } (Eqvl : forall x, sigmavl x = (var_vl mvl) x) (Eqstm : s = s') : subst_tm sigmavl s = s' :=
eq.trans (idSubst_tm sigmavl Eqvl s) Eqstm

@[substLemmas]
lemma idSubst'_vl { mvl : nat } { sigmavl : Fin mvl -> vl mvl } { Eqvl : forall x, sigmavl x = (var_vl mvl) x } { s s' : vl mvl } (Eqvl : forall x, sigmavl x = (var_vl mvl) x) (Eqsvl : s = s') : subst_vl sigmavl s = s' :=
eq.trans (idSubst_vl sigmavl Eqvl s) Eqsvl

@[substLemmas]
def upExtRen_vl_vl { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_vl_vl xi) x = (upRen_vl_vl zeta) x :=
  λ n , match n with
  | (some n) := ap shift (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def extRen_tm,extRen_vl
with extRen_tm : Π { mvl : nat } { nvl : nat } (xivl : Fin mvl -> Fin nvl) (zetavl : Fin mvl -> Fin nvl) (Eqvl : forall x, xivl x = zetavl x) (s : tm mvl), ren_tm xivl s = ren_tm zetavl s
    | mvl nvl xivl zetavl Eqvl 
    (val _ s0) := have h : _, from congr_val (extRen_vl xivl zetavl Eqvl s0), begin simp with substLemmas at *, assumption end
    | mvl nvl xivl zetavl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (extRen_tm (upRen_vl_vl xivl) (upRen_vl_vl zetavl) (upExtRen_vl_vl xivl zetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with extRen_vl : Π { mvl : nat } { nvl : nat } (xivl : Fin mvl -> Fin nvl) (zetavl : Fin mvl -> Fin nvl) (Eqvl : forall x, xivl x = zetavl x) (s : vl mvl), ren_vl xivl s = ren_vl zetavl s
    | mvl nvl xivl zetavl Eqvl  (var_vl _ s) := have h : _, from ap (var_vl nvl) (Eqvl s), begin simp with substLemmas at *, assumption end
    | mvl nvl xivl zetavl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (extRen_vl xivl zetavl Eqvl s0) (extRen_tm xivl zetavl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma extRen'_tm { mvl : nat } { nvl : nat } { xivl : Fin mvl -> Fin nvl } { zetavl : Fin mvl -> Fin nvl } { s s' : tm mvl } (Eqvl : forall x, xivl x = zetavl x) (Eqstm : s = s') : ren_tm xivl s = ren_tm zetavl s' :=
have h : _, from extRen_tm xivl zetavl Eqvl s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma extRen'_vl { mvl : nat } { nvl : nat } { xivl : Fin mvl -> Fin nvl } { zetavl : Fin mvl -> Fin nvl } { s s' : vl mvl } (Eqvl : forall x, xivl x = zetavl x) (Eqsvl : s = s') : ren_vl xivl s = ren_vl zetavl s' :=
have h : _, from extRen_vl xivl zetavl Eqvl s, begin rw <- Eqsvl, assumption end

@[substLemmas]
def upExt_vl_vl { m } { nvl : nat } (sigma : Fin m -> vl nvl) (tau : Fin m -> vl nvl) (Eq : forall x, sigma x = tau x) : forall x, (up_vl_vl sigma) x = (up_vl_vl tau) x :=
  λ n , match n with
  | (some n) := ap (ren_vl shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def ext_tm,ext_vl
with ext_tm : Π { mvl : nat } { nvl : nat } (sigmavl : Fin mvl -> vl nvl) (tauvl : Fin mvl -> vl nvl) (Eqvl : forall x, sigmavl x = tauvl x) (s : tm mvl), subst_tm sigmavl s = subst_tm tauvl s
    | mvl nvl sigmavl tauvl Eqvl 
    (val _ s0) := have h : _, from congr_val (ext_vl sigmavl tauvl Eqvl s0), begin simp with substLemmas at *, assumption end
    | mvl nvl sigmavl tauvl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (ext_tm (up_vl_vl sigmavl) (up_vl_vl tauvl) (upExt_vl_vl sigmavl tauvl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with ext_vl : Π { mvl : nat } { nvl : nat } (sigmavl : Fin mvl -> vl nvl) (tauvl : Fin mvl -> vl nvl) (Eqvl : forall x, sigmavl x = tauvl x) (s : vl mvl), subst_vl sigmavl s = subst_vl tauvl s
    | mvl nvl sigmavl tauvl Eqvl  (var_vl _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | mvl nvl sigmavl tauvl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (ext_vl sigmavl tauvl Eqvl s0) (ext_tm sigmavl tauvl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma ext'_tm { mvl : nat } { nvl : nat } { sigmavl : Fin mvl -> vl nvl } { tauvl : Fin mvl -> vl nvl } { s s' : tm mvl } (Eqvl : forall x, sigmavl x = tauvl x) (Eqstm : s = s') : subst_tm sigmavl s = subst_tm tauvl s' :=
have h : _, from ext_tm sigmavl tauvl Eqvl s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma ext'_vl { mvl : nat } { nvl : nat } { sigmavl : Fin mvl -> vl nvl } { tauvl : Fin mvl -> vl nvl } { s s' : vl mvl } (Eqvl : forall x, sigmavl x = tauvl x) (Eqsvl : s = s') : subst_vl sigmavl s = subst_vl tauvl s' :=
have h : _, from ext_vl sigmavl tauvl Eqvl s, begin rw <- Eqsvl, assumption end

@[substLemmas]
mutual def compRenRen_tm,compRenRen_vl
with compRenRen_tm : Π { kvl : nat } { lvl : nat } { mvl : nat } (xivl : Fin mvl -> Fin kvl) (zetavl : Fin kvl -> Fin lvl) (rhovl : Fin mvl -> Fin lvl) (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (s : tm mvl), ren_tm zetavl (ren_tm xivl s) = ren_tm rhovl s
    | kvl lvl mvl xivl zetavl rhovl Eqvl 
    (val _ s0) := have h : _, from congr_val (compRenRen_vl xivl zetavl rhovl Eqvl s0), begin simp with substLemmas at *, assumption end
    | kvl lvl mvl xivl zetavl rhovl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (compRenRen_tm (upRen_vl_vl xivl) (upRen_vl_vl zetavl) (upRen_vl_vl rhovl) (up_ren_ren xivl zetavl rhovl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with compRenRen_vl : Π { kvl : nat } { lvl : nat } { mvl : nat } (xivl : Fin mvl -> Fin kvl) (zetavl : Fin kvl -> Fin lvl) (rhovl : Fin mvl -> Fin lvl) (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (s : vl mvl), ren_vl zetavl (ren_vl xivl s) = ren_vl rhovl s
    | kvl lvl mvl xivl zetavl rhovl Eqvl 
    (var_vl _ s) := have h : _, from ap (var_vl lvl) (Eqvl s), begin simp with substLemmas at *, assumption end
    | kvl lvl mvl xivl zetavl rhovl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compRenRen_vl xivl zetavl rhovl Eqvl s0) (compRenRen_tm xivl zetavl rhovl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compRenRen'_tm { kvl : nat } { lvl : nat } { mvl : nat } { xivl : Fin mvl -> Fin kvl } { zetavl : Fin kvl -> Fin lvl } { rhovl : Fin mvl -> Fin lvl } { s : tm mvl } { s' : tm kvl } (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (Eqstm : s' = ren_tm xivl s) : ren_tm zetavl s' = ren_tm rhovl s :=
have h : _, from compRenRen_tm xivl zetavl rhovl Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compRenRen'_vl { kvl : nat } { lvl : nat } { mvl : nat } { xivl : Fin mvl -> Fin kvl } { zetavl : Fin kvl -> Fin lvl } { rhovl : Fin mvl -> Fin lvl } { s : vl mvl } { s' : vl kvl } (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (Eqsvl : s' = ren_vl xivl s) : ren_vl zetavl s' = ren_vl rhovl s :=
have h : _, from compRenRen_vl xivl zetavl rhovl Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def up_ren_subst_vl_vl { k } { l } { mvl : nat } (xi : Fin k -> Fin l) (tau : Fin l -> vl mvl) (theta : Fin k -> vl mvl) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_vl_vl tau) (upRen_vl_vl xi)) x = (up_vl_vl theta) x :=
  λ n , match n with
  | (some n) := ap (ren_vl shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def compRenSubst_tm,compRenSubst_vl
with compRenSubst_tm : Π { kvl : nat } { lvl : nat } { mvl : nat } (xivl : Fin mvl -> Fin kvl) (tauvl : Fin kvl -> vl lvl) (thetavl : Fin mvl -> vl lvl) (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (s : tm mvl), subst_tm tauvl (ren_tm xivl s) = subst_tm thetavl s
    | kvl lvl mvl xivl tauvl thetavl Eqvl 
    (val _ s0) := have h : _, from congr_val (compRenSubst_vl xivl tauvl thetavl Eqvl s0), begin simp with substLemmas at *, assumption end
    | kvl lvl mvl xivl tauvl thetavl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (compRenSubst_tm (upRen_vl_vl xivl) (up_vl_vl tauvl) (up_vl_vl thetavl) (up_ren_subst_vl_vl xivl tauvl thetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with compRenSubst_vl : Π { kvl : nat } { lvl : nat } { mvl : nat } (xivl : Fin mvl -> Fin kvl) (tauvl : Fin kvl -> vl lvl) (thetavl : Fin mvl -> vl lvl) (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (s : vl mvl), subst_vl tauvl (ren_vl xivl s) = subst_vl thetavl s
    | kvl lvl mvl xivl tauvl thetavl Eqvl  (var_vl _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | kvl lvl mvl xivl tauvl thetavl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compRenSubst_vl xivl tauvl thetavl Eqvl s0) (compRenSubst_tm xivl tauvl thetavl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compRenSubst'_tm { kvl : nat } { lvl : nat } { mvl : nat } { xivl : Fin mvl -> Fin kvl } { tauvl : Fin kvl -> vl lvl } { thetavl : Fin mvl -> vl lvl } { s : tm mvl } { s' : tm kvl } (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (Eqstm : s' = ren_tm xivl s) : subst_tm tauvl s' = subst_tm thetavl s :=
have h : _, from compRenSubst_tm xivl tauvl thetavl Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compRenSubst'_vl { kvl : nat } { lvl : nat } { mvl : nat } { xivl : Fin mvl -> Fin kvl } { tauvl : Fin kvl -> vl lvl } { thetavl : Fin mvl -> vl lvl } { s : vl mvl } { s' : vl kvl } (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (Eqsvl : s' = ren_vl xivl s) : subst_vl tauvl s' = subst_vl thetavl s :=
have h : _, from compRenSubst_vl xivl tauvl thetavl Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def up_subst_ren_vl_vl { k } { lvl : nat } { mvl : nat } (sigma : Fin k -> vl lvl) (zetavl : Fin lvl -> Fin mvl) (theta : Fin k -> vl mvl) (Eq : forall x, (comp (ren_vl zetavl) sigma) x = theta x) : forall x, (comp (ren_vl (upRen_vl_vl zetavl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenRen_vl shift (upRen_vl_vl zetavl) (comp shift zetavl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_vl zetavl shift (comp shift zetavl) (λ x , by refl) (sigma n))) (ap (ren_vl shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
mutual def compSubstRen_tm,compSubstRen_vl
with compSubstRen_tm : Π { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : Fin mvl -> vl kvl) (zetavl : Fin kvl -> Fin lvl) (thetavl : Fin mvl -> vl lvl) (Eqvl : forall x, (comp (ren_vl zetavl) sigmavl) x = thetavl x) (s : tm mvl), ren_tm zetavl (subst_tm sigmavl s) = subst_tm thetavl s
    | kvl lvl mvl sigmavl zetavl thetavl Eqvl 
    (val _ s0) := have h : _, from congr_val (compSubstRen_vl sigmavl zetavl thetavl Eqvl s0), begin simp with substLemmas at *, assumption end
    | kvl lvl mvl sigmavl zetavl thetavl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (compSubstRen_tm (up_vl_vl sigmavl) (upRen_vl_vl zetavl) (up_vl_vl thetavl) (up_subst_ren_vl_vl sigmavl zetavl thetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with compSubstRen_vl : Π { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : Fin mvl -> vl kvl) (zetavl : Fin kvl -> Fin lvl) (thetavl : Fin mvl -> vl lvl) (Eqvl : forall x, (comp (ren_vl zetavl) sigmavl) x = thetavl x) (s : vl mvl), ren_vl zetavl (subst_vl sigmavl s) = subst_vl thetavl s
    | kvl lvl mvl sigmavl zetavl thetavl Eqvl  (var_vl _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | kvl lvl mvl sigmavl zetavl thetavl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compSubstRen_vl sigmavl zetavl thetavl Eqvl s0) (compSubstRen_tm sigmavl zetavl thetavl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compSubstRen'_tm { kvl : nat } { lvl : nat } { mvl : nat } { sigmavl : Fin mvl -> vl kvl } { zetavl : Fin kvl -> Fin lvl } { thetavl : Fin mvl -> vl lvl } { s : tm mvl } { s' : tm kvl } (Eqvl : forall x, (comp (ren_vl zetavl) sigmavl) x = thetavl x) (Eqstm : s' = subst_tm sigmavl s) : ren_tm zetavl s' = subst_tm thetavl s :=
have h : _, from compSubstRen_tm sigmavl zetavl thetavl Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compSubstRen'_vl { kvl : nat } { lvl : nat } { mvl : nat } { sigmavl : Fin mvl -> vl kvl } { zetavl : Fin kvl -> Fin lvl } { thetavl : Fin mvl -> vl lvl } { s : vl mvl } { s' : vl kvl } (Eqvl : forall x, (comp (ren_vl zetavl) sigmavl) x = thetavl x) (Eqsvl : s' = subst_vl sigmavl s) : ren_vl zetavl s' = subst_vl thetavl s :=
have h : _, from compSubstRen_vl sigmavl zetavl thetavl Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def up_subst_subst_vl_vl { k } { lvl : nat } { mvl : nat } (sigma : Fin k -> vl lvl) (tauvl : Fin lvl -> vl mvl) (theta : Fin k -> vl mvl) (Eq : forall x, (comp (subst_vl tauvl) sigma) x = theta x) : forall x, (comp (subst_vl (up_vl_vl tauvl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenSubst_vl shift (up_vl_vl tauvl) (comp (up_vl_vl tauvl) shift) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_vl tauvl shift (comp (ren_vl shift) tauvl) (λ x , by refl) (sigma n))) (ap (ren_vl shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
mutual def compSubstSubst_tm,compSubstSubst_vl
with compSubstSubst_tm : Π { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : Fin mvl -> vl kvl) (tauvl : Fin kvl -> vl lvl) (thetavl : Fin mvl -> vl lvl) (Eqvl : forall x, (comp (subst_vl tauvl) sigmavl) x = thetavl x) (s : tm mvl), subst_tm tauvl (subst_tm sigmavl s) = subst_tm thetavl s
    | kvl lvl mvl sigmavl tauvl thetavl Eqvl 
    (val _ s0) := have h : _, from congr_val (compSubstSubst_vl sigmavl tauvl thetavl Eqvl s0), begin simp with substLemmas at *, assumption end
    | kvl lvl mvl sigmavl tauvl thetavl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (compSubstSubst_tm (up_vl_vl sigmavl) (up_vl_vl tauvl) (up_vl_vl thetavl) (up_subst_subst_vl_vl sigmavl tauvl thetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with compSubstSubst_vl : Π { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : Fin mvl -> vl kvl) (tauvl : Fin kvl -> vl lvl) (thetavl : Fin mvl -> vl lvl) (Eqvl : forall x, (comp (subst_vl tauvl) sigmavl) x = thetavl x) (s : vl mvl), subst_vl tauvl (subst_vl sigmavl s) = subst_vl thetavl s
    | kvl lvl mvl sigmavl tauvl thetavl Eqvl  (var_vl _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | kvl lvl mvl sigmavl tauvl thetavl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compSubstSubst_vl sigmavl tauvl thetavl Eqvl s0) (compSubstSubst_tm sigmavl tauvl thetavl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compSubstSubst'_tm { kvl : nat } { lvl : nat } { mvl : nat } { sigmavl : Fin mvl -> vl kvl } { tauvl : Fin kvl -> vl lvl } { thetavl : Fin mvl -> vl lvl } { s : tm mvl } { s' : tm kvl } (Eqvl : forall x, (comp (subst_vl tauvl) sigmavl) x = thetavl x) (Eqstm : s' = subst_tm sigmavl s) : subst_tm tauvl s' = subst_tm thetavl s :=
have h : _, from compSubstSubst_tm sigmavl tauvl thetavl Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compSubstSubst'_vl { kvl : nat } { lvl : nat } { mvl : nat } { sigmavl : Fin mvl -> vl kvl } { tauvl : Fin kvl -> vl lvl } { thetavl : Fin mvl -> vl lvl } { s : vl mvl } { s' : vl kvl } (Eqvl : forall x, (comp (subst_vl tauvl) sigmavl) x = thetavl x) (Eqsvl : s' = subst_vl sigmavl s) : subst_vl tauvl s' = subst_vl thetavl s :=
have h : _, from compSubstSubst_vl sigmavl tauvl thetavl Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def rinstInst_up_vl_vl { m } { nvl : nat } (xi : Fin m -> Fin nvl) (sigma : Fin m -> vl nvl) (Eq : forall x, (comp (var_vl nvl) xi) x = sigma x) : forall x, (comp (var_vl (nat.succ nvl)) (upRen_vl_vl xi)) x = (up_vl_vl sigma) x :=
  λ n , match n with
  | (some n) := ap (ren_vl shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def rinst_inst_tm,rinst_inst_vl
with rinst_inst_tm : Π { mvl : nat } { nvl : nat } (xivl : Fin mvl -> Fin nvl) (sigmavl : Fin mvl -> vl nvl) (Eqvl : forall x, (comp (var_vl nvl) xivl) x = sigmavl x) (s : tm mvl), ren_tm xivl s = subst_tm sigmavl s
    | mvl nvl xivl sigmavl Eqvl 
    (val _ s0) := have h : _, from congr_val (rinst_inst_vl xivl sigmavl Eqvl s0), begin simp with substLemmas at *, assumption end
    | mvl nvl xivl sigmavl Eqvl 
    (lam _ s0) := have h : _, from congr_lam (rinst_inst_tm (upRen_vl_vl xivl) (up_vl_vl sigmavl) (rinstInst_up_vl_vl xivl sigmavl Eqvl) s0), begin simp with substLemmas at *, assumption end
 with rinst_inst_vl : Π { mvl : nat } { nvl : nat } (xivl : Fin mvl -> Fin nvl) (sigmavl : Fin mvl -> vl nvl) (Eqvl : forall x, (comp (var_vl nvl) xivl) x = sigmavl x) (s : vl mvl), ren_vl xivl s = subst_vl sigmavl s
    | mvl nvl xivl sigmavl Eqvl  (var_vl _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | mvl nvl xivl sigmavl Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (rinst_inst_vl xivl sigmavl Eqvl s0) (rinst_inst_tm xivl sigmavl Eqvl s1), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma rinst_inst'_tm { mvl : nat } { nvl : nat } { sigmavl : Fin mvl -> vl nvl } { xivl : Fin mvl -> Fin nvl } { s s' : tm mvl } (Eqvl : forall x, (comp (var_vl nvl) xivl) x = sigmavl x) (Eqstm : s = s') : ren_tm xivl s' = subst_tm sigmavl s :=
have h : _, from rinst_inst_tm xivl sigmavl Eqvl s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma rinst_inst'_vl { mvl : nat } { nvl : nat } { sigmavl : Fin mvl -> vl nvl } { xivl : Fin mvl -> Fin nvl } { s s' : vl mvl } (Eqvl : forall x, (comp (var_vl nvl) xivl) x = sigmavl x) (Eqsvl : s = s') : ren_vl xivl s' = subst_vl sigmavl s :=
have h : _, from rinst_inst_vl xivl sigmavl Eqvl s, begin rw <- Eqsvl, assumption end

meta def is_val_tm : expr -> bool
  | `(val _ %%s0) := tt
  | `(congr_val' %%s0) := tt
  | `(idSubst'_tm %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_val_tm : expr -> tactic expr
  | `(val _ %%s0) := return s0
  | `(congr_val' %%s0) := return s0
  | `(idSubst'_tm %%Eqvl %%Eqstm) := destruct_val_tm Eqstm
  | e := return e

meta def is_lam_tm : expr -> bool
  | `(lam _ %%s0) := tt
  | `(congr_lam' %%s0) := tt
  | `(idSubst'_tm %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_lam_tm : expr -> tactic expr
  | `(lam _ %%s0) := return s0
  | `(congr_lam' %%s0) := return s0
  | `(idSubst'_tm %%Eqvl %%Eqstm) := destruct_lam_tm Eqstm
  | e := return e

meta def is_app_vl : expr -> bool
  | `(app _ %%s0 %%s1) := tt
  | `(congr_app' %%s0 %%s1) := tt
  | `(idSubst'_vl %%Eqvl %%Eqsvl) := tt
  | _ := ff

meta def destruct_app_vl : expr -> tactic (expr × expr)
  | `(app _ %%s0 %%s1) := return (s0, s1)
  | `(congr_app' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_vl %%Eqvl %%Eqsvl) := destruct_app_vl Eqsvl
  | e := return (e, e)

meta def is_InstRen_tm : expr -> bool
  | `(extRen'_tm %%Eqvl %%Eqstm) := tt
  | `(compRenRen'_tm %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_InstRen_tm : expr -> tactic (expr × expr)
  | `(extRen'_tm %%Eqvl %%Eqstm) := return (Eqvl, Eqstm)
  | `(compRenRen'_tm %%Eqvl %%Eqstm) := destruct_InstRen_tm Eqstm
  | `(%%e) := return (e,e)

meta def is_InstRen_vl : expr -> bool
  | `(extRen'_vl %%Eqvl %%Eqsvl) := tt
  | `(compRenRen'_vl %%Eqvl %%Eqsvl) := tt
  | _ := ff

meta def destruct_InstRen_vl : expr -> tactic (expr × expr)
  | `(extRen'_vl %%Eqvl %%Eqsvl) := return (Eqvl, Eqsvl)
  | `(compRenRen'_vl %%Eqvl %%Eqsvl) := destruct_InstRen_vl Eqsvl
  | `(%%e) := return (e,e)

meta def is_Inst_tm : expr -> bool
  | `(compRenRen'_tm %%Eqvl %%Eqstm) := tt
  | `(compRenSubst'_tm %%Eqvl %%Eqstm) := tt
  | `(compSubstRen'_tm %%Eqvl %%Eqstm) := tt
  | `(compSubstSubst'_tm %%Eqvl %%Eqstm) := tt
  | `(idSubst'_tm %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_Inst_tm : expr -> tactic (expr × expr)
  | `(ext'_tm %%Eqvl %%Eqstm) := return (Eqvl, Eqstm)
  | `(compRenRen'_tm %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compRenSubst'_tm %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstRen'_tm %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstSubst'