Require Export fintype.

Inductive tm (nvl : nat) : Type :=
  | val  : vl nvl -> tm nvl
  | lam  : tm (S nvl) -> tm nvl
 with vl (nvl : nat) : Type :=
  | var_vl  : fin nvl -> vl nvl
  | app  : vl nvl -> tm nvl -> vl nvl.

Lemma congr_val { mvl : nat } { s0 : vl mvl } { t0 : vl mvl } : s0 = t0 -> val mvl s0 = val mvl t0 .
Proof. congruence. Qed.

Lemma congr_lam { mvl : nat } { s0 : tm (S mvl) } { t0 : tm (S mvl) } : s0 = t0 -> lam mvl s0 = lam mvl t0 .
Proof. congruence. Qed.

Lemma congr_app { mvl : nat } { s0 : vl mvl } { s1 : tm mvl } { t0 : vl mvl } { t1 : tm mvl } : s0 = t0 -> s1 = t1 -> app mvl s0 s1 = app mvl t0 t1 .
Proof. congruence. Qed.

Definition upRen_vl_vl { m } { n } (xi : fin m -> fin n) : _ :=
  up_ren xi.

Fixpoint ren_tm { mvl : nat } { nvl : nat } (xivl : fin mvl -> fin nvl) (s : tm mvl) : _ :=
    match s with
    | val s0 => val nvl (ren_vl xivl s0)
    | lam s0 => lam nvl (ren_tm (upRen_vl_vl xivl) s0)
    end
 with ren_vl { mvl : nat } { nvl : nat } (xivl : fin mvl -> fin nvl) (s : vl mvl) : _ :=
    match s with
    | var_vl s => (var_vl nvl) (xivl s)
    | app s0 s1 => app nvl (ren_vl xivl s0) (ren_tm xivl s1)
    end.

Definition up_vl_vl { m } { nvl : nat } (sigma : fin m -> vl nvl) : _ :=
  scons ((var_vl (S nvl)) var_zero) (comp tt (ren_vl shift) sigma).

Fixpoint subst_tm { mvl : nat } { nvl : nat } (sigmavl : fin mvl -> vl nvl) (s : tm mvl) : _ :=
    match s with
    | val s0 => val nvl (subst_vl sigmavl s0)
    | lam s0 => lam nvl (subst_tm (up_vl_vl sigmavl) s0)
    end
 with subst_vl { mvl : nat } { nvl : nat } (sigmavl : fin mvl -> vl nvl) (s : vl mvl) : _ :=
    match s with
    | var_vl s => sigmavl s
    | app s0 s1 => app nvl (subst_vl sigmavl s0) (subst_tm sigmavl s1)
    end.

Definition upId_vl_vl { mvl : nat } (sigma : fin mvl -> vl mvl) (Eq : forall x, sigma x = (var_vl mvl) x) : forall x, (up_vl_vl sigma) x = (var_vl (S mvl)) x :=
  fun n => match n with
  | Some n => ap (ren_vl shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint idSubst_tm { mvl : nat } (sigmavl : fin mvl -> vl mvl) (Eqvl : forall x, sigmavl x = (var_vl mvl) x) (s : tm mvl) : subst_tm sigmavl s = s :=
    match s with
    | val s0 => congr_val (idSubst_vl sigmavl Eqvl s0)
    | lam s0 => congr_lam (idSubst_tm (up_vl_vl sigmavl) (upId_vl_vl (_) Eqvl) s0)
    end
 with idSubst_vl { mvl : nat } (sigmavl : fin mvl -> vl mvl) (Eqvl : forall x, sigmavl x = (var_vl mvl) x) (s : vl mvl) : subst_vl sigmavl s = s :=
    match s with
    | var_vl s => Eqvl s
    | app s0 s1 => congr_app (idSubst_vl sigmavl Eqvl s0) (idSubst_tm sigmavl Eqvl s1)
    end.

Definition upExtRen_vl_vl { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_vl_vl xi) x = (upRen_vl_vl zeta) x :=
  fun n => match n with
  | Some n => ap shift (Eq n)
  | None => eq_refl
  end.

Fixpoint extRen_tm { mvl : nat } { nvl : nat } (xivl : fin mvl -> fin nvl) (zetavl : fin mvl -> fin nvl) (Eqvl : forall x, xivl x = zetavl x) (s : tm mvl) : ren_tm xivl s = ren_tm zetavl s :=
    match s with
    | val s0 => congr_val (extRen_vl xivl zetavl Eqvl s0)
    | lam s0 => congr_lam (extRen_tm (upRen_vl_vl xivl) (upRen_vl_vl zetavl) (upExtRen_vl_vl xivl zetavl Eqvl) s0)
    end
 with extRen_vl { mvl : nat } { nvl : nat } (xivl : fin mvl -> fin nvl) (zetavl : fin mvl -> fin nvl) (Eqvl : forall x, xivl x = zetavl x) (s : vl mvl) : ren_vl xivl s = ren_vl zetavl s :=
    match s with
    | var_vl s => ap (var_vl nvl) (Eqvl s)
    | app s0 s1 => congr_app (extRen_vl xivl zetavl Eqvl s0) (extRen_tm xivl zetavl Eqvl s1)
    end.

Definition upExt_vl_vl { m } { nvl : nat } (sigma : fin m -> vl nvl) (tau : fin m -> vl nvl) (Eq : forall x, sigma x = tau x) : forall x, (up_vl_vl sigma) x = (up_vl_vl tau) x :=
  fun n => match n with
  | Some n => ap (ren_vl shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint ext_tm { mvl : nat } { nvl : nat } (sigmavl : fin mvl -> vl nvl) (tauvl : fin mvl -> vl nvl) (Eqvl : forall x, sigmavl x = tauvl x) (s : tm mvl) : subst_tm sigmavl s = subst_tm tauvl s :=
    match s with
    | val s0 => congr_val (ext_vl sigmavl tauvl Eqvl s0)
    | lam s0 => congr_lam (ext_tm (up_vl_vl sigmavl) (up_vl_vl tauvl) (upExt_vl_vl sigmavl tauvl Eqvl) s0)
    end
 with ext_vl { mvl : nat } { nvl : nat } (sigmavl : fin mvl -> vl nvl) (tauvl : fin mvl -> vl nvl) (Eqvl : forall x, sigmavl x = tauvl x) (s : vl mvl) : subst_vl sigmavl s = subst_vl tauvl s :=
    match s with
    | var_vl s => Eqvl s
    | app s0 s1 => congr_app (ext_vl sigmavl tauvl Eqvl s0) (ext_tm sigmavl tauvl Eqvl s1)
    end.

Fixpoint compRenRen_tm { kvl : nat } { lvl : nat } { mvl : nat } (xivl : fin mvl -> fin kvl) (zetavl : fin kvl -> fin lvl) (rhovl : fin mvl -> fin lvl) (Eqvl : forall x, (comp tt zetavl xivl) x = rhovl x) (s : tm mvl) : ren_tm zetavl (ren_tm xivl s) = ren_tm rhovl s :=
    match s with
    | val s0 => congr_val (compRenRen_vl xivl zetavl rhovl Eqvl s0)
    | lam s0 => congr_lam (compRenRen_tm (upRen_vl_vl xivl) (upRen_vl_vl zetavl) (upRen_vl_vl rhovl) (up_ren_ren xivl zetavl rhovl Eqvl) s0)
    end
 with compRenRen_vl { kvl : nat } { lvl : nat } { mvl : nat } (xivl : fin mvl -> fin kvl) (zetavl : fin kvl -> fin lvl) (rhovl : fin mvl -> fin lvl) (Eqvl : forall x, (comp tt zetavl xivl) x = rhovl x) (s : vl mvl) : ren_vl zetavl (ren_vl xivl s) = ren_vl rhovl s :=
    match s with
    | var_vl s => ap (var_vl lvl) (Eqvl s)
    | app s0 s1 => congr_app (compRenRen_vl xivl zetavl rhovl Eqvl s0) (compRenRen_tm xivl zetavl rhovl Eqvl s1)
    end.

Definition up_ren_subst_vl_vl { k } { l } { mvl : nat } (xi : fin k -> fin l) (tau : fin l -> vl mvl) (theta : fin k -> vl mvl) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_vl_vl tau) (upRen_vl_vl xi)) x = (up_vl_vl theta) x :=
  fun n => match n with
  | Some n => ap (ren_vl shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint compRenSubst_tm { kvl : nat } { lvl : nat } { mvl : nat } (xivl : fin mvl -> fin kvl) (tauvl : fin kvl -> vl lvl) (thetavl : fin mvl -> vl lvl) (Eqvl : forall x, (comp tt tauvl xivl) x = thetavl x) (s : tm mvl) : subst_tm tauvl (ren_tm xivl s) = subst_tm thetavl s :=
    match s with
    | val s0 => congr_val (compRenSubst_vl xivl tauvl thetavl Eqvl s0)
    | lam s0 => congr_lam (compRenSubst_tm (upRen_vl_vl xivl) (up_vl_vl tauvl) (up_vl_vl thetavl) (up_ren_subst_vl_vl xivl tauvl thetavl Eqvl) s0)
    end
 with compRenSubst_vl { kvl : nat } { lvl : nat } { mvl : nat } (xivl : fin mvl -> fin kvl) (tauvl : fin kvl -> vl lvl) (thetavl : fin mvl -> vl lvl) (Eqvl : forall x, (comp tt tauvl xivl) x = thetavl x) (s : vl mvl) : subst_vl tauvl (ren_vl xivl s) = subst_vl thetavl s :=
    match s with
    | var_vl s => Eqvl s
    | app s0 s1 => congr_app (compRenSubst_vl xivl tauvl thetavl Eqvl s0) (compRenSubst_tm xivl tauvl thetavl Eqvl s1)
    end.

Definition up_subst_ren_vl_vl { k } { lvl : nat } { mvl : nat } (sigma : fin k -> vl lvl) (zetavl : fin lvl -> fin mvl) (theta : fin k -> vl mvl) (Eq : forall x, (comp tt (ren_vl zetavl) sigma) x = theta x) : forall x, (comp tt (ren_vl (upRen_vl_vl zetavl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenRen_vl shift (upRen_vl_vl zetavl) (comp tt shift zetavl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_vl zetavl shift (comp tt shift zetavl) (fun x => eq_refl) (sigma n))) (ap (ren_vl shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstRen__tm { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : fin mvl -> vl kvl) (zetavl : fin kvl -> fin lvl) (thetavl : fin mvl -> vl lvl) (Eqvl : forall x, (comp tt (ren_vl zetavl) sigmavl) x = thetavl x) (s : tm mvl) : ren_tm zetavl (subst_tm sigmavl s) = subst_tm thetavl s :=
    match s with
    | val s0 => congr_val (compSubstRen__vl sigmavl zetavl thetavl Eqvl s0)
    | lam s0 => congr_lam (compSubstRen__tm (up_vl_vl sigmavl) (upRen_vl_vl zetavl) (up_vl_vl thetavl) (up_subst_ren_vl_vl sigmavl zetavl thetavl Eqvl) s0)
    end
 with compSubstRen__vl { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : fin mvl -> vl kvl) (zetavl : fin kvl -> fin lvl) (thetavl : fin mvl -> vl lvl) (Eqvl : forall x, (comp tt (ren_vl zetavl) sigmavl) x = thetavl x) (s : vl mvl) : ren_vl zetavl (subst_vl sigmavl s) = subst_vl thetavl s :=
    match s with
    | var_vl s => Eqvl s
    | app s0 s1 => congr_app (compSubstRen__vl sigmavl zetavl thetavl Eqvl s0) (compSubstRen__tm sigmavl zetavl thetavl Eqvl s1)
    end.

Definition up_subst_subst_vl_vl { k } { lvl : nat } { mvl : nat } (sigma : fin k -> vl lvl) (tauvl : fin lvl -> vl mvl) (theta : fin k -> vl mvl) (Eq : forall x, (comp tt (subst_vl tauvl) sigma) x = theta x) : forall x, (comp tt (subst_vl (up_vl_vl tauvl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenSubst_vl shift (up_vl_vl tauvl) (comp tt (up_vl_vl tauvl) shift) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen__vl tauvl shift (comp tt (ren_vl shift) tauvl) (fun x => eq_refl) (sigma n))) (ap (ren_vl shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstSubst_tm { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : fin mvl -> vl kvl) (tauvl : fin kvl -> vl lvl) (thetavl : fin mvl -> vl lvl) (Eqvl : forall x, (comp tt (subst_vl tauvl) sigmavl) x = thetavl x) (s : tm mvl) : subst_tm tauvl (subst_tm sigmavl s) = subst_tm thetavl s :=
    match s with
    | val s0 => congr_val (compSubstSubst_vl sigmavl tauvl thetavl Eqvl s0)
    | lam s0 => congr_lam (compSubstSubst_tm (up_vl_vl sigmavl) (up_vl_vl tauvl) (up_vl_vl thetavl) (up_subst_subst_vl_vl sigmavl tauvl thetavl Eqvl) s0)
    end
 with compSubstSubst_vl { kvl : nat } { lvl : nat } { mvl : nat } (sigmavl : fin mvl -> vl kvl) (tauvl : fin kvl -> vl lvl) (thetavl : fin mvl -> vl lvl) (Eqvl : forall x, (comp tt (subst_vl tauvl) sigmavl) x = thetavl x) (s : vl mvl) : subst_vl tauvl (subst_vl sigmavl s) = subst_vl thetavl s :=
    match s with
    | var_vl s => Eqvl s
    | app s0 s1 => congr_app (compSubstSubst_vl sigmavl tauvl thetavl Eqvl s0) (compSubstSubst_tm sigmavl tauvl thetavl Eqvl s1)
    end.

Definition rinstInst_up_vl_vl { m } { nvl : nat } (xi : fin m -> fin nvl) (sigma : fin m -> vl nvl) (Eq : forall x, (comp tt (var_vl nvl) xi) x = sigma x) : forall x, (comp tt (var_vl (S nvl)) (upRen_vl_vl xi)) x = (up_vl_vl sigma) x :=
  fun n => match n with
  | Some n => ap (ren_vl shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint rinst_inst_tm { mvl : nat } { nvl : nat } (xivl : fin mvl -> fin nvl) (sigmavl : fin mvl -> vl nvl) (Eqvl : forall x, (comp tt (var_vl nvl) xivl) x = sigmavl x) (s : tm mvl) : ren_tm xivl s = subst_tm sigmavl s :=
    match s with
    | val s0 => congr_val (rinst_inst_vl xivl sigmavl Eqvl s0)
    | lam s0 => congr_lam (rinst_inst_tm (upRen_vl_vl xivl) (up_vl_vl sigmavl) (rinstInst_up_vl_vl xivl sigmavl Eqvl) s0)
    end
 with rinst_inst_vl { mvl : nat } { nvl : nat } (xivl : fin mvl -> fin nvl) (sigmavl : fin mvl -> vl nvl) (Eqvl : forall x, (comp tt (var_vl nvl) xivl) x = sigmavl x) (s : vl mvl) : ren_vl xivl s = subst_vl sigmavl s :=
    match s with
    | var_vl s => Eqvl s
    | app s0 s1 => congr_app (rinst_inst_vl xivl sigmavl Eqvl s0) (rinst_inst_tm xivl sigmavl Eqvl s1)
    end.
