open nat
def ap {X Y} (f:X → Y) {x y : X} (p : x = y) : f x = f y 
  := by induction p; refl 

def apc {X Y} {f g : X -> Y} {x y : X} (p : f = g) (q : x = y) : f x = g y :=
  by induction p; induction q; refl

def funcomp {X Y Z} (g: Y → Z) (f : X → Y):= λ (x : X), g (f x)
notation f`>>`g:50 := funcomp g f

def Fin : nat → Type
  | 0 := empty
  | (n+1) := option (Fin n)

def scons  {X : Type} {n : nat} (x : X) (f : (Fin n) → X) 
  (m : Fin (nat.succ n)) : X  :=
  match m with 
  | none := x  
  | (some i) := (f i) end

notation x `.:` f := @scons _ _ x f

def ren (m n : nat) : Type := Fin m → Fin n

--instance coe_ren {n m :nat }: has_coe  (ren m n)(Fin m → Fin n) := ⟨λ x, x⟩

def bound {n : nat} : Fin (succ n) := none

def null {T} (i : Fin 0) : T := by cases i

def shift {n : nat} : ren n (succ n) := some

def up_ren {m n} (ξ : ren m n) : ren (succ m) (succ n)  :=  bound .: (ξ >> shift)

def var_zero {n : nat} : Fin (succ n) := none

def idren {k: nat} : ren k k := λ x, x

def comp := @funcomp

lemma up_ren_ren {k l m} (ξ : ren k l) (ζ : ren l m) (ρ : ren k m) (E : ∀ x, (ξ >> ζ) x = ρ x) : (∀ x, (up_ren ξ >> up_ren ζ) x = up_ren ρ x) := 
begin 
  intro x, cases x, {refl,}, specialize (E a), simp [up_ren, scons,shift,funcomp], rw ← E, refl   
end

lemma scons_eta {T} {n : nat} (f : Fin (succ n) → T) :
  ((f bound) .: (shift >> f)) = f :=
begin
  apply funext, simp [ shift, bound], intro x, cases x; refl end 

lemma scons_eta' {T} {n:nat} {f f' f'': Fin (succ n) → T} {g : Fin n→ Fin (succ n)} {x : Fin (succ n)}
 (H1 : x = bound) (H2 : ∀ x, g x = shift x) (H3 : ∀ x, f'' x = f' x) : (∀ x, f x = f' x)
  → ∀ y, ((f'' x) .: (g >> f)) y = f' y :=
begin intros r, rw H1, simp[shift,bound], rw H3, intro x, cases x, refl, simp[scons, funcomp], rw ← r, rw H2, refl end 

lemma scons_eta_id {n : nat} : (@scons _ n bound shift = id) := 
begin 
  apply funext, intro x, cases x; refl end 
lemma scons_eta_id' {n : nat} {f : Fin n → Fin (succ n)} {y: Fin (succ n)}
(H1 : y = bound) (H2 : ∀ x, f x = shift x) : ∀ x, (y .: f) x = id x
  := begin  intros x, rw H1, cases x, refl, unfold scons, rw (H2 a), refl end 

lemma scons_comp {T1 T2} {n : nat} (x : T1) (f : Fin n → T1) (g : T1 → T2) : 
  (x .: f) >> g = ((g x) .: (f >> g)) := 
begin apply funext, intro x, cases x; refl end

lemma scons_comp' {T1 T2} {n : nat} {x : T1} {x' : T2} {f : Fin n → T1} {f' : Fin (succ n) → T2} {g : T1 → T2}
 (H1 : ∀ x', ((g x) .: (f >> g)) x' = f' x') : ∀ x', ((x .: f) >> g) x' = f' x':=
begin intros x, cases x; rw ← H1; refl  end

lemma comp' {T1 T2} {n : nat} {f f': Fin n → T1} {g g': T1 → T2} (H1 : ∀ x, f x = f' x) (H2 : ∀ x, g x = g' x) :  ∀ x, (f>>g) x = (f'>>g') x := 
by intros x; simp[funcomp, (H1 x), (H2 _)]

lemma scons' {X : Type} {n : nat} {x x': X} {f f': (Fin n) → X}
  {m m': Fin (nat.succ n)} (H1 : x = x') (H2 : f = f') : ∀ y, (x .: f) y = (x' .: f') y:=
by intros y; simp[scons, H1,H2]

def Inst := 1
def Rinst := 2
def Comp := 3 
def Up := 4
def Cons := 5
meta def ex (p:pexpr) := tactic.to_expr p



meta def simp_normalized (normalize : bool → expr → tactic expr ): tactic unit := do t <- tactic.target,  e <- normalize tt t, tactic.rewrite_target e --todo in context?
