Require Export fintype.

Inductive ty (nty : nat) : Type :=
  | var_ty  : fin nty -> ty nty
  | arr  : ty nty -> ty nty -> ty nty
  | all  : ty (S nty) -> ty nty.

Lemma congr_arr { mty : nat } { s0 : ty mty } { s1 : ty mty } { t0 : ty mty } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> arr mty s0 s1 = arr mty t0 t1 .
Proof. congruence. Qed.

Lemma congr_all { mty : nat } { s0 : ty (S mty) } { t0 : ty (S mty) } : s0 = t0 -> all mty s0 = all mty t0 .
Proof. congruence. Qed.

Definition upRen_ty_ty { m } { n } (xi : fin m -> fin n) : _ :=
  up_ren xi.

Fixpoint ren_ty { mty : nat } { nty : nat } (xity : fin mty -> fin nty) (s : ty mty) : _ :=
    match s with
    | var_ty s => (var_ty nty) (xity s)
    | arr s0 s1 => arr nty (ren_ty xity s0) (ren_ty xity s1)
    | all s0 => all nty (ren_ty (upRen_ty_ty xity) s0)
    end.

Definition up_ty_ty { m } { nty : nat } (sigma : fin m -> ty nty) : _ :=
  scons ((var_ty (S nty)) var_zero) (comp tt (ren_ty shift) sigma).

Fixpoint subst_ty { mty : nat } { nty : nat } (sigmaty : fin mty -> ty nty) (s : ty mty) : _ :=
    match s with
    | var_ty s => sigmaty s
    | arr s0 s1 => arr nty (subst_ty sigmaty s0) (subst_ty sigmaty s1)
    | all s0 => all nty (subst_ty (up_ty_ty sigmaty) s0)
    end.

Definition upId_ty_ty { mty : nat } (sigma : fin mty -> ty mty) (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_ty_ty sigma) x = (var_ty (S mty)) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint idSubst_ty { mty : nat } (sigmaty : fin mty -> ty mty) (Eqty : forall x, sigmaty x = (var_ty mty) x) (s : ty mty) : subst_ty sigmaty s = s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (idSubst_ty sigmaty Eqty s0) (idSubst_ty sigmaty Eqty s1)
    | all s0 => congr_all (idSubst_ty (up_ty_ty sigmaty) (upId_ty_ty (_) Eqty) s0)
    end.

Definition upExtRen_ty_ty { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_ty xi) x = (upRen_ty_ty zeta) x :=
  fun n => match n with
  | Some n => ap shift (Eq n)
  | None => eq_refl
  end.

Fixpoint extRen_ty { mty : nat } { nty : nat } (xity : fin mty -> fin nty) (zetaty : fin mty -> fin nty) (Eqty : forall x, xity x = zetaty x) (s : ty mty) : ren_ty xity s = ren_ty zetaty s :=
    match s with
    | var_ty s => ap (var_ty nty) (Eqty s)
    | arr s0 s1 => congr_arr (extRen_ty xity zetaty Eqty s0) (extRen_ty xity zetaty Eqty s1)
    | all s0 => congr_all (extRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upExtRen_ty_ty xity zetaty Eqty) s0)
    end.

Definition upExt_ty_ty { m } { nty : nat } (sigma : fin m -> ty nty) (tau : fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_ty sigma) x = (up_ty_ty tau) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint ext_ty { mty : nat } { nty : nat } (sigmaty : fin mty -> ty nty) (tauty : fin mty -> ty nty) (Eqty : forall x, sigmaty x = tauty x) (s : ty mty) : subst_ty sigmaty s = subst_ty tauty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (ext_ty sigmaty tauty Eqty s0) (ext_ty sigmaty tauty Eqty s1)
    | all s0 => congr_all (ext_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (upExt_ty_ty sigmaty tauty Eqty) s0)
    end.

Fixpoint compRenRen_ty { kty : nat } { lty : nat } { mty : nat } (xity : fin mty -> fin kty) (zetaty : fin kty -> fin lty) (rhoty : fin mty -> fin lty) (Eqty : forall x, (comp tt zetaty xity) x = rhoty x) (s : ty mty) : ren_ty zetaty (ren_ty xity s) = ren_ty rhoty s :=
    match s with
    | var_ty s => ap (var_ty lty) (Eqty s)
    | arr s0 s1 => congr_arr (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_ty xity zetaty rhoty Eqty s1)
    | all s0 => congr_all (compRenRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upRen_ty_ty rhoty) (up_ren_ren xity zetaty rhoty Eqty) s0)
    end.

Definition up_ren_subst_ty_ty { k } { l } { mty : nat } (xi : fin k -> fin l) (tau : fin l -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_ty_ty tau) (upRen_ty_ty xi)) x = (up_ty_ty theta) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint compRenSubst_ty { kty : nat } { lty : nat } { mty : nat } (xity : fin mty -> fin kty) (tauty : fin kty -> ty lty) (thetaty : fin mty -> ty lty) (Eqty : forall x, (comp tt tauty xity) x = thetaty x) (s : ty mty) : subst_ty tauty (ren_ty xity s) = subst_ty thetaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_ty xity tauty thetaty Eqty s1)
    | all s0 => congr_all (compRenSubst_ty (upRen_ty_ty xity) (up_ty_ty tauty) (up_ty_ty thetaty) (up_ren_subst_ty_ty xity tauty thetaty Eqty) s0)
    end.

Definition up_subst_ren_ty_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (zetaty : fin lty -> fin mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (ren_ty zetaty) sigma) x = theta x) : forall x, (comp tt (ren_ty (upRen_ty_ty zetaty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenRen_ty shift (upRen_ty_ty zetaty) (comp tt shift zetaty) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_ty zetaty shift (comp tt shift zetaty) (fun x => eq_refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstRen_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : fin mty -> ty kty) (zetaty : fin kty -> fin lty) (thetaty : fin mty -> ty lty) (Eqty : forall x, (comp tt (ren_ty zetaty) sigmaty) x = thetaty x) (s : ty mty) : ren_ty zetaty (subst_ty sigmaty s) = subst_ty thetaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (compSubstRen_ty sigmaty zetaty thetaty Eqty s0) (compSubstRen_ty sigmaty zetaty thetaty Eqty s1)
    | all s0 => congr_all (compSubstRen_ty (up_ty_ty sigmaty) (upRen_ty_ty zetaty) (up_ty_ty thetaty) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) s0)
    end.

Definition up_subst_subst_ty_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (tauty : fin lty -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (subst_ty tauty) sigma) x = theta x) : forall x, (comp tt (subst_ty (up_ty_ty tauty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenSubst_ty shift (up_ty_ty tauty) (comp tt (up_ty_ty tauty) shift) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen_ty tauty shift (comp tt (ren_ty shift) tauty) (fun x => eq_refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstSubst_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : fin mty -> ty kty) (tauty : fin kty -> ty lty) (thetaty : fin mty -> ty lty) (Eqty : forall x, (comp tt (subst_ty tauty) sigmaty) x = thetaty x) (s : ty mty) : subst_ty tauty (subst_ty sigmaty s) = subst_ty thetaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1)
    | all s0 => congr_all (compSubstSubst_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (up_ty_ty thetaty) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) s0)
    end.

Definition rinstInst_up_ty_ty { m } { nty : nat } (xi : fin m -> fin nty) (sigma : fin m -> ty nty) (Eq : forall x, (comp tt (var_ty nty) xi) x = sigma x) : forall x, (comp tt (var_ty (S nty)) (upRen_ty_ty xi)) x = (up_ty_ty sigma) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint rinst_inst_ty { mty : nat } { nty : nat } (xity : fin mty -> fin nty) (sigmaty : fin mty -> ty nty) (Eqty : forall x, (comp tt (var_ty nty) xity) x = sigmaty x) (s : ty mty) : ren_ty xity s = subst_ty sigmaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_ty xity sigmaty Eqty s1)
    | all s0 => congr_all (rinst_inst_ty (upRen_ty_ty xity) (up_ty_ty sigmaty) (rinstInst_up_ty_ty xity sigmaty Eqty) s0)
    end.

Inductive tm (nty ntm : nat) : Type :=
  | var_tm  : fin ntm -> tm nty ntm
  | app  : tm nty ntm -> tm nty ntm -> tm nty ntm
  | lam  : ty nty -> tm nty (S ntm) -> tm nty ntm
  | tapp  : tm nty ntm -> ty nty -> tm nty ntm
  | tlam  : tm (S nty) ntm -> tm nty ntm.

Lemma congr_app { mty mtm : nat } { s0 : tm mty mtm } { s1 : tm mty mtm } { t0 : tm mty mtm } { t1 : tm mty mtm } : s0 = t0 -> s1 = t1 -> app mty mtm s0 s1 = app mty mtm t0 t1 .
Proof. congruence. Qed.

Lemma congr_lam { mty mtm : nat } { s0 : ty mty } { s1 : tm mty (S mtm) } { t0 : ty mty } { t1 : tm mty (S mtm) } : s0 = t0 -> s1 = t1 -> lam mty mtm s0 s1 = lam mty mtm t0 t1 .
Proof. congruence. Qed.

Lemma congr_tapp { mty mtm : nat } { s0 : tm mty mtm } { s1 : ty mty } { t0 : tm mty mtm } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> tapp mty mtm s0 s1 = tapp mty mtm t0 t1 .
Proof. congruence. Qed.

Lemma congr_tlam { mty mtm : nat } { s0 : tm (S mty) mtm } { t0 : tm (S mty) mtm } : s0 = t0 -> tlam mty mtm s0 = tlam mty mtm t0 .
Proof. congruence. Qed.

Definition upRen_ty_tm { m } { n } (xi : fin m -> fin n) : _ :=
  xi.

Definition upRen_tm_ty { m } { n } (xi : fin m -> fin n) : _ :=
  xi.

Definition upRen_tm_tm { m } { n } (xi : fin m -> fin n) : _ :=
  up_ren xi.

Fixpoint ren_tm { mty mtm : nat } { nty ntm : nat } (xity : fin mty -> fin nty) (xitm : fin mtm -> fin ntm) (s : tm mty mtm) : _ :=
    match s with
    | var_tm s => (var_tm nty ntm) (xitm s)
    | app s0 s1 => app nty ntm (ren_tm xity xitm s0) (ren_tm xity xitm s1)
    | lam s0 s1 => lam nty ntm (ren_ty xity s0) (ren_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) s1)
    | tapp s0 s1 => tapp nty ntm (ren_tm xity xitm s0) (ren_ty xity s1)
    | tlam s0 => tlam nty ntm (ren_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) s0)
    end.

Definition up_ty_tm { m } { nty ntm : nat } (sigma : fin m -> tm nty ntm) : _ :=
  comp tt (ren_tm shift id) sigma.

Definition up_tm_ty { m } { nty : nat } (sigma : fin m -> ty nty) : _ :=
  comp tt (ren_ty id) sigma.

Definition up_tm_tm { m } { nty ntm : nat } (sigma : fin m -> tm nty ntm) : _ :=
  scons ((var_tm nty (S ntm)) var_zero) (comp tt (ren_tm id shift) sigma).

Fixpoint subst_tm { mty mtm : nat } { nty ntm : nat } (sigmaty : fin mty -> ty nty) (sigmatm : fin mtm -> tm nty ntm) (s : tm mty mtm) : _ :=
    match s with
    | var_tm s => sigmatm s
    | app s0 s1 => app nty ntm (subst_tm sigmaty sigmatm s0) (subst_tm sigmaty sigmatm s1)
    | lam s0 s1 => lam nty ntm (subst_ty sigmaty s0) (subst_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) s1)
    | tapp s0 s1 => tapp nty ntm (subst_tm sigmaty sigmatm s0) (subst_ty sigmaty s1)
    | tlam s0 => tlam nty ntm (subst_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) s0)
    end.

Definition upId_ty_tm { mty mtm : nat } (sigma : fin mtm -> tm mty mtm) (Eq : forall x, sigma x = (var_tm mty mtm) x) : forall x, (up_ty_tm sigma) x = (var_tm (S mty) mtm) x :=
  fun n => ap (ren_tm shift id) (Eq n).

Definition upId_tm_ty { mty : nat } (sigma : fin mty -> ty mty) (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_tm_ty sigma) x = (var_ty mty) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition upId_tm_tm { mty mtm : nat } (sigma : fin mtm -> tm mty mtm) (Eq : forall x, sigma x = (var_tm mty mtm) x) : forall x, (up_tm_tm sigma) x = (var_tm mty (S mtm)) x :=
  fun n => match n with
  | Some n => ap (ren_tm id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint idSubst_tm { mty mtm : nat } (sigmaty : fin mty -> ty mty) (sigmatm : fin mtm -> tm mty mtm) (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqtm : forall x, sigmatm x = (var_tm mty mtm) x) (s : tm mty mtm) : subst_tm sigmaty sigmatm s = s :=
    match s with
    | var_tm s => Eqtm s
    | app s0 s1 => congr_app (idSubst_tm sigmaty sigmatm Eqty Eqtm s0) (idSubst_tm sigmaty sigmatm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (idSubst_ty sigmaty Eqty s0) (idSubst_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (upId_tm_ty (_) Eqty) (upId_tm_tm (_) Eqtm) s1)
    | tapp s0 s1 => congr_tapp (idSubst_tm sigmaty sigmatm Eqty Eqtm s0) (idSubst_ty sigmaty Eqty s1)
    | tlam s0 => congr_tlam (idSubst_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (upId_ty_ty (_) Eqty) (upId_ty_tm (_) Eqtm) s0)
    end.

Definition upExtRen_ty_tm { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_tm xi) x = (upRen_ty_tm zeta) x :=
  fun n => Eq n.

Definition upExtRen_tm_ty { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_tm_ty xi) x = (upRen_tm_ty zeta) x :=
  fun n => Eq n.

Definition upExtRen_tm_tm { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_tm_tm xi) x = (upRen_tm_tm zeta) x :=
  fun n => match n with
  | Some n => ap shift (Eq n)
  | None => eq_refl
  end.

Fixpoint extRen_tm { mty mtm : nat } { nty ntm : nat } (xity : fin mty -> fin nty) (xitm : fin mtm -> fin ntm) (zetaty : fin mty -> fin nty) (zetatm : fin mtm -> fin ntm) (Eqty : forall x, xity x = zetaty x) (Eqtm : forall x, xitm x = zetatm x) (s : tm mty mtm) : ren_tm xity xitm s = ren_tm zetaty zetatm s :=
    match s with
    | var_tm s => ap (var_tm nty ntm) (Eqtm s)
    | app s0 s1 => congr_app (extRen_tm xity xitm zetaty zetatm Eqty Eqtm s0) (extRen_tm xity xitm zetaty zetatm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (extRen_ty xity zetaty Eqty s0) (extRen_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (upExtRen_tm_ty xity zetaty Eqty) (upExtRen_tm_tm xitm zetatm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (extRen_tm xity xitm zetaty zetatm Eqty Eqtm s0) (extRen_ty xity zetaty Eqty s1)
    | tlam s0 => congr_tlam (extRen_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (upExtRen_ty_ty xity zetaty Eqty) (upExtRen_ty_tm xitm zetatm Eqtm) s0)
    end.

Definition upExt_ty_tm { m } { nty ntm : nat } (sigma : fin m -> tm nty ntm) (tau : fin m -> tm nty ntm) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_tm sigma) x = (up_ty_tm tau) x :=
  fun n => ap (ren_tm shift id) (Eq n).

Definition upExt_tm_ty { m } { nty : nat } (sigma : fin m -> ty nty) (tau : fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_tm_ty sigma) x = (up_tm_ty tau) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition upExt_tm_tm { m } { nty ntm : nat } (sigma : fin m -> tm nty ntm) (tau : fin m -> tm nty ntm) (Eq : forall x, sigma x = tau x) : forall x, (up_tm_tm sigma) x = (up_tm_tm tau) x :=
  fun n => match n with
  | Some n => ap (ren_tm id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint ext_tm { mty mtm : nat } { nty ntm : nat } (sigmaty : fin mty -> ty nty) (sigmatm : fin mtm -> tm nty ntm) (tauty : fin mty -> ty nty) (tautm : fin mtm -> tm nty ntm) (Eqty : forall x, sigmaty x = tauty x) (Eqtm : forall x, sigmatm x = tautm x) (s : tm mty mtm) : subst_tm sigmaty sigmatm s = subst_tm tauty tautm s :=
    match s with
    | var_tm s => Eqtm s
    | app s0 s1 => congr_app (ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s0) (ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (ext_ty sigmaty tauty Eqty s0) (ext_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (up_tm_ty tauty) (up_tm_tm tautm) (upExt_tm_ty sigmaty tauty Eqty) (upExt_tm_tm sigmatm tautm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (ext_tm sigmaty sigmatm tauty tautm Eqty Eqtm s0) (ext_ty sigmaty tauty Eqty s1)
    | tlam s0 => congr_tlam (ext_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (up_ty_ty tauty) (up_ty_tm tautm) (upExt_ty_ty sigmaty tauty Eqty) (upExt_ty_tm sigmatm tautm Eqtm) s0)
    end.

Fixpoint compRenRen_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (xity : fin mty -> fin kty) (xitm : fin mtm -> fin ktm) (zetaty : fin kty -> fin lty) (zetatm : fin ktm -> fin ltm) (rhoty : fin mty -> fin lty) (rhotm : fin mtm -> fin ltm) (Eqty : forall x, (comp tt zetaty xity) x = rhoty x) (Eqtm : forall x, (comp tt zetatm xitm) x = rhotm x) (s : tm mty mtm) : ren_tm zetaty zetatm (ren_tm xity xitm s) = ren_tm rhoty rhotm s :=
    match s with
    | var_tm s => ap (var_tm lty ltm) (Eqtm s)
    | app s0 s1 => congr_app (compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s0) (compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (upRen_tm_ty rhoty) (upRen_tm_tm rhotm) Eqty (up_ren_ren xitm zetatm rhotm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (compRenRen_tm xity xitm zetaty zetatm rhoty rhotm Eqty Eqtm s0) (compRenRen_ty xity zetaty rhoty Eqty s1)
    | tlam s0 => congr_tlam (compRenRen_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (upRen_ty_ty rhoty) (upRen_ty_tm rhotm) (up_ren_ren xity zetaty rhoty Eqty) Eqtm s0)
    end.

Definition up_ren_subst_ty_tm { k } { l } { mty mtm : nat } (xi : fin k -> fin l) (tau : fin l -> tm mty mtm) (theta : fin k -> tm mty mtm) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_ty_tm tau) (upRen_ty_tm xi)) x = (up_ty_tm theta) x :=
  fun n => ap (ren_tm shift id) (Eq n).

Definition up_ren_subst_tm_ty { k } { l } { mty : nat } (xi : fin k -> fin l) (tau : fin l -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_tm_ty tau) (upRen_tm_ty xi)) x = (up_tm_ty theta) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition up_ren_subst_tm_tm { k } { l } { mty mtm : nat } (xi : fin k -> fin l) (tau : fin l -> tm mty mtm) (theta : fin k -> tm mty mtm) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_tm_tm tau) (upRen_tm_tm xi)) x = (up_tm_tm theta) x :=
  fun n => match n with
  | Some n => ap (ren_tm id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint compRenSubst_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (xity : fin mty -> fin kty) (xitm : fin mtm -> fin ktm) (tauty : fin kty -> ty lty) (tautm : fin ktm -> tm lty ltm) (thetaty : fin mty -> ty lty) (thetatm : fin mtm -> tm lty ltm) (Eqty : forall x, (comp tt tauty xity) x = thetaty x) (Eqtm : forall x, (comp tt tautm xitm) x = thetatm x) (s : tm mty mtm) : subst_tm tauty tautm (ren_tm xity xitm s) = subst_tm thetaty thetatm s :=
    match s with
    | var_tm s => Eqtm s
    | app s0 s1 => congr_app (compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s0) (compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (up_tm_ty tauty) (up_tm_tm tautm) (up_tm_ty thetaty) (up_tm_tm thetatm) (up_ren_subst_tm_ty xity tauty thetaty Eqty) (up_ren_subst_tm_tm xitm tautm thetatm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (compRenSubst_tm xity xitm tauty tautm thetaty thetatm Eqty Eqtm s0) (compRenSubst_ty xity tauty thetaty Eqty s1)
    | tlam s0 => congr_tlam (compRenSubst_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (up_ty_ty tauty) (up_ty_tm tautm) (up_ty_ty thetaty) (up_ty_tm thetatm) (up_ren_subst_ty_ty xity tauty thetaty Eqty) (up_ren_subst_ty_tm xitm tautm thetatm Eqtm) s0)
    end.

Definition up_subst_ren_ty_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : fin k -> tm lty ltm) (zetaty : fin lty -> fin mty) (zetatm : fin ltm -> fin mtm) (theta : fin k -> tm mty mtm) (Eq : forall x, (comp tt (ren_tm zetaty zetatm) sigma) x = theta x) : forall x, (comp tt (ren_tm (upRen_ty_ty zetaty) (upRen_ty_tm zetatm)) (up_ty_tm sigma)) x = (up_ty_tm theta) x :=
  fun n => eq_trans (compRenRen_tm shift id (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (comp tt shift zetaty) (comp tt id zetatm) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_tm zetaty zetatm shift id (comp tt shift zetaty) (comp tt id zetatm) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_tm shift id) (Eq n))).

Definition up_subst_ren_tm_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (zetaty : fin lty -> fin mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (ren_ty zetaty) sigma) x = theta x) : forall x, (comp tt (ren_ty (upRen_tm_ty zetaty)) (up_tm_ty sigma)) x = (up_tm_ty theta) x :=
  fun n => eq_trans (compRenRen_ty id (upRen_tm_ty zetaty) (comp tt id zetaty) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_ty zetaty id (comp tt id zetaty) (fun x => eq_refl) (sigma n))) (ap (ren_ty id) (Eq n))).

Definition up_subst_ren_tm_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : fin k -> tm lty ltm) (zetaty : fin lty -> fin mty) (zetatm : fin ltm -> fin mtm) (theta : fin k -> tm mty mtm) (Eq : forall x, (comp tt (ren_tm zetaty zetatm) sigma) x = theta x) : forall x, (comp tt (ren_tm (upRen_tm_ty zetaty) (upRen_tm_tm zetatm)) (up_tm_tm sigma)) x = (up_tm_tm theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenRen_tm id shift (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (comp tt id zetaty) (comp tt shift zetatm) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_tm zetaty zetatm id shift (comp tt id zetaty) (comp tt shift zetatm) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_tm id shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstRen_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (sigmaty : fin mty -> ty kty) (sigmatm : fin mtm -> tm kty ktm) (zetaty : fin kty -> fin lty) (zetatm : fin ktm -> fin ltm) (thetaty : fin mty -> ty lty) (thetatm : fin mtm -> tm lty ltm) (Eqty : forall x, (comp tt (ren_ty zetaty) sigmaty) x = thetaty x) (Eqtm : forall x, (comp tt (ren_tm zetaty zetatm) sigmatm) x = thetatm x) (s : tm mty mtm) : ren_tm zetaty zetatm (subst_tm sigmaty sigmatm s) = subst_tm thetaty thetatm s :=
    match s with
    | var_tm s => Eqtm s
    | app s0 s1 => congr_app (compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s0) (compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (compSubstRen_ty sigmaty zetaty thetaty Eqty s0) (compSubstRen_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (upRen_tm_ty zetaty) (upRen_tm_tm zetatm) (up_tm_ty thetaty) (up_tm_tm thetatm) (up_subst_ren_tm_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_tm_tm sigmatm zetaty zetatm thetatm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (compSubstRen_tm sigmaty sigmatm zetaty zetatm thetaty thetatm Eqty Eqtm s0) (compSubstRen_ty sigmaty zetaty thetaty Eqty s1)
    | tlam s0 => congr_tlam (compSubstRen_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (upRen_ty_ty zetaty) (upRen_ty_tm zetatm) (up_ty_ty thetaty) (up_ty_tm thetatm) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_ty_tm sigmatm zetaty zetatm thetatm Eqtm) s0)
    end.

Definition up_subst_subst_ty_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : fin k -> tm lty ltm) (tauty : fin lty -> ty mty) (tautm : fin ltm -> tm mty mtm) (theta : fin k -> tm mty mtm) (Eq : forall x, (comp tt (subst_tm tauty tautm) sigma) x = theta x) : forall x, (comp tt (subst_tm (up_ty_ty tauty) (up_ty_tm tautm)) (up_ty_tm sigma)) x = (up_ty_tm theta) x :=
  fun n => eq_trans (compRenSubst_tm shift id (up_ty_ty tauty) (up_ty_tm tautm) (comp tt (up_ty_ty tauty) shift) (comp tt (up_ty_tm tautm) id) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen_tm tauty tautm shift id (comp tt (ren_ty shift) tauty) (comp tt (ren_tm shift id) tautm) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_tm shift id) (Eq n))).

Definition up_subst_subst_tm_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (tauty : fin lty -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (subst_ty tauty) sigma) x = theta x) : forall x, (comp tt (subst_ty (up_tm_ty tauty)) (up_tm_ty sigma)) x = (up_tm_ty theta) x :=
  fun n => eq_trans (compRenSubst_ty id (up_tm_ty tauty) (comp tt (up_tm_ty tauty) id) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen_ty tauty id (comp tt (ren_ty id) tauty) (fun x => eq_refl) (sigma n))) (ap (ren_ty id) (Eq n))).

Definition up_subst_subst_tm_tm { k } { lty ltm : nat } { mty mtm : nat } (sigma : fin k -> tm lty ltm) (tauty : fin lty -> ty mty) (tautm : fin ltm -> tm mty mtm) (theta : fin k -> tm mty mtm) (Eq : forall x, (comp tt (subst_tm tauty tautm) sigma) x = theta x) : forall x, (comp tt (subst_tm (up_tm_ty tauty) (up_tm_tm tautm)) (up_tm_tm sigma)) x = (up_tm_tm theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenSubst_tm id shift (up_tm_ty tauty) (up_tm_tm tautm) (comp tt (up_tm_ty tauty) id) (comp tt (up_tm_tm tautm) shift) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen_tm tauty tautm id shift (comp tt (ren_ty id) tauty) (comp tt (ren_tm id shift) tautm) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_tm id shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstSubst_tm { kty ktm : nat } { lty ltm : nat } { mty mtm : nat } (sigmaty : fin mty -> ty kty) (sigmatm : fin mtm -> tm kty ktm) (tauty : fin kty -> ty lty) (tautm : fin ktm -> tm lty ltm) (thetaty : fin mty -> ty lty) (thetatm : fin mtm -> tm lty ltm) (Eqty : forall x, (comp tt (subst_ty tauty) sigmaty) x = thetaty x) (Eqtm : forall x, (comp tt (subst_tm tauty tautm) sigmatm) x = thetatm x) (s : tm mty mtm) : subst_tm tauty tautm (subst_tm sigmaty sigmatm s) = subst_tm thetaty thetatm s :=
    match s with
    | var_tm s => Eqtm s
    | app s0 s1 => congr_app (compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s0) (compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_tm (up_tm_ty sigmaty) (up_tm_tm sigmatm) (up_tm_ty tauty) (up_tm_tm tautm) (up_tm_ty thetaty) (up_tm_tm thetatm) (up_subst_subst_tm_ty sigmaty tauty thetaty Eqty) (up_subst_subst_tm_tm sigmatm tauty tautm thetatm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (compSubstSubst_tm sigmaty sigmatm tauty tautm thetaty thetatm Eqty Eqtm s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1)
    | tlam s0 => congr_tlam (compSubstSubst_tm (up_ty_ty sigmaty) (up_ty_tm sigmatm) (up_ty_ty tauty) (up_ty_tm tautm) (up_ty_ty thetaty) (up_ty_tm thetatm) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) (up_subst_subst_ty_tm sigmatm tauty tautm thetatm Eqtm) s0)
    end.

Definition rinstInst_up_ty_tm { m } { nty ntm : nat } (xi : fin m -> fin ntm) (sigma : fin m -> tm nty ntm) (Eq : forall x, (comp tt (var_tm nty ntm) xi) x = sigma x) : forall x, (comp tt (var_tm (S nty) ntm) (upRen_ty_tm xi)) x = (up_ty_tm sigma) x :=
  fun n => ap (ren_tm shift id) (Eq n).

Definition rinstInst_up_tm_ty { m } { nty : nat } (xi : fin m -> fin nty) (sigma : fin m -> ty nty) (Eq : forall x, (comp tt (var_ty nty) xi) x = sigma x) : forall x, (comp tt (var_ty nty) (upRen_tm_ty xi)) x = (up_tm_ty sigma) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition rinstInst_up_tm_tm { m } { nty ntm : nat } (xi : fin m -> fin ntm) (sigma : fin m -> tm nty ntm) (Eq : forall x, (comp tt (var_tm nty ntm) xi) x = sigma x) : forall x, (comp tt (var_tm nty (S ntm)) (upRen_tm_tm xi)) x = (up_tm_tm sigma) x :=
  fun n => match n with
  | Some n => ap (ren_tm id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint rinst_inst_tm { mty mtm : nat } { nty ntm : nat } (xity : fin mty -> fin nty) (xitm : fin mtm -> fin ntm) (sigmaty : fin mty -> ty nty) (sigmatm : fin mtm -> tm nty ntm) (Eqty : forall x, (comp tt (var_ty nty) xity) x = sigmaty x) (Eqtm : forall x, (comp tt (var_tm nty ntm) xitm) x = sigmatm x) (s : tm mty mtm) : ren_tm xity xitm s = subst_tm sigmaty sigmatm s :=
    match s with
    | var_tm s => Eqtm s
    | app s0 s1 => congr_app (rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s0) (rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s1)
    | lam s0 s1 => congr_lam (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_tm (upRen_tm_ty xity) (upRen_tm_tm xitm) (up_tm_ty sigmaty) (up_tm_tm sigmatm) (rinstInst_up_tm_ty xity sigmaty Eqty) (rinstInst_up_tm_tm xitm sigmatm Eqtm) s1)
    | tapp s0 s1 => congr_tapp (rinst_inst_tm xity xitm sigmaty sigmatm Eqty Eqtm s0) (rinst_inst_ty xity sigmaty Eqty s1)
    | tlam s0 => congr_tlam (rinst_inst_tm (upRen_ty_ty xity) (upRen_ty_tm xitm) (up_ty_ty sigmaty) (up_ty_tm sigmatm) (rinstInst_up_ty_ty xity sigmaty Eqty) (rinstInst_up_ty_tm xitm sigmatm Eqtm) s0)
    end.
