import .fintype
import .wftactic
run_cmd mk_simp_attr `substLemmas

@[substLemmas]
inductive ty  : nat -> Type 
  | var_ty : Π (nty : nat), Fin nty -> ty nty
  | arr : Π (nty : nat), ty nty -> ty nty -> ty nty
  | all : Π (nty : nat), ty (nat.succ nty) -> ty nty
open ty 

@[substLemmas]
lemma congr_arr { mty : nat } { s0 : ty mty } { s1 : ty mty } { t0 : ty mty } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> arr mty s0 s1 = arr mty t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_all { mty : nat } { s0 : ty (nat.succ mty) } { t0 : ty (nat.succ mty) } : s0 = t0 -> all mty s0 = all mty t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_arr' { mty : nat } { s0 : ty mty } { s1 : ty mty } { t0 : ty mty } { t1 : ty mty } (h0 : s0 = t0) (h1 : s1 = t1) : arr mty s0 s1 = arr mty t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_all' { mty : nat } { s0 : ty (nat.succ mty) } { t0 : ty (nat.succ mty) } (h0 : s0 = t0) : all mty s0 = all mty t0 :=
begin intros, congr; assumption end

@[substLemmas]
def upRen_ty_ty { m } { n } (xi : Fin m -> Fin n) : _ :=
  up_ren xi

@[substLemmas]
def ren_ty : Π { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) (s : ty mty), ty nty
    | mty nty xity  (var_ty _ s) := (var_ty nty) (xity s)
    | mty nty xity  (arr _ s0 s1) := arr nty (ren_ty xity s0) (ren_ty xity s1)
    | mty nty xity  (all _ s0) := all nty (ren_ty (upRen_ty_ty xity) s0)

@[substLemmas]
def up_ty_ty { m } { nty : nat } (sigma : Fin m -> ty nty) : _ :=
  scons ((var_ty (nat.succ nty)) var_zero) (comp (ren_ty shift) sigma)

@[substLemmas]
def subst_ty : Π { mty : nat } { nty : nat } (sigmaty : Fin mty -> ty nty) (s : ty mty), ty nty
    | mty nty sigmaty  (var_ty _ s) := sigmaty s
    | mty nty sigmaty  (arr _ s0 s1) := arr nty (subst_ty sigmaty s0) (subst_ty sigmaty s1)
    | mty nty sigmaty  (all _ s0) := all nty (subst_ty (up_ty_ty sigmaty) s0)

@[substLemmas]
def upId_ty_ty { mty : nat } { sigma : Fin mty -> ty mty } (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_ty_ty sigma) x = (var_ty (nat.succ mty)) x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_ty shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
def idSubst_ty : Π { mty : nat } (sigmaty : Fin mty -> ty mty) (Eqty : forall x, sigmaty x = (var_ty mty) x) (s : ty mty), subst_ty sigmaty s = s
    | mty sigmaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | mty sigmaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (idSubst_ty sigmaty Eqty s0) (idSubst_ty sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty sigmaty Eqty 
    (all _ s0) := have h : _, from congr_all (idSubst_ty (up_ty_ty sigmaty) (upId_ty_ty (_) Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma idSubst'_ty { mty : nat } { sigmaty : Fin mty -> ty mty } { Eqty : forall x, sigmaty x = (var_ty mty) x } { s s' : ty mty } (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqsty : s = s') : subst_ty sigmaty s = s' :=
eq.trans (idSubst_ty sigmaty Eqty s) Eqsty

@[substLemmas]
def upExtRen_ty_ty { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_ty xi) x = (upRen_ty_ty zeta) x :=
  λ n , match n with
  | (some n) := ap shift (Eq n)
  | none := by refl
  end

@[substLemmas]
def extRen_ty : Π { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) (zetaty : Fin mty -> Fin nty) (Eqty : forall x, xity x = zetaty x) (s : ty mty), ren_ty xity s = ren_ty zetaty s
    | mty nty xity zetaty Eqty  (var_ty _ s) := have h : _, from ap (var_ty nty) (Eqty s), begin simp with substLemmas at *, assumption end
    | mty nty xity zetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (extRen_ty xity zetaty Eqty s0) (extRen_ty xity zetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty nty xity zetaty Eqty 
    (all _ s0) := have h : _, from congr_all (extRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upExtRen_ty_ty xity zetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma extRen'_ty { mty : nat } { nty : nat } { xity : Fin mty -> Fin nty } { zetaty : Fin mty -> Fin nty } { s s' : ty mty } (Eqty : forall x, xity x = zetaty x) (Eqsty : s = s') : ren_ty xity s = ren_ty zetaty s' :=
have h : _, from extRen_ty xity zetaty Eqty s, begin rw <- Eqsty, assumption end

@[substLemmas]
def upExt_ty_ty { m } { nty : nat } (sigma : Fin m -> ty nty) (tau : Fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_ty sigma) x = (up_ty_ty tau) x :=
  λ n , match n with
  | (some n) := ap (ren_ty shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def ext_ty : Π { mty : nat } { nty : nat } (sigmaty : Fin mty -> ty nty) (tauty : Fin mty -> ty nty) (Eqty : forall x, sigmaty x = tauty x) (s : ty mty), subst_ty sigmaty s = subst_ty tauty s
    | mty nty sigmaty tauty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | mty nty sigmaty tauty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (ext_ty sigmaty tauty Eqty s0) (ext_ty sigmaty tauty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty nty sigmaty tauty Eqty 
    (all _ s0) := have h : _, from congr_all (ext_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (upExt_ty_ty sigmaty tauty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma ext'_ty { mty : nat } { nty : nat } { sigmaty : Fin mty -> ty nty } { tauty : Fin mty -> ty nty } { s s' : ty mty } (Eqty : forall x, sigmaty x = tauty x) (Eqsty : s = s') : subst_ty sigmaty s = subst_ty tauty s' :=
have h : _, from ext_ty sigmaty tauty Eqty s, begin rw <- Eqsty, assumption end

@[substLemmas]
def compRenRen_ty : Π { kty : nat } { lty : nat } { mty : nat } (xity : Fin mty -> Fin kty) (zetaty : Fin kty -> Fin lty) (rhoty : Fin mty -> Fin lty) (Eqty : forall x, (comp zetaty xity) x = rhoty x) (s : ty mty), ren_ty zetaty (ren_ty xity s) = ren_ty rhoty s
    | kty lty mty xity zetaty rhoty Eqty 
    (var_ty _ s) := have h : _, from ap (var_ty lty) (Eqty s), begin simp with substLemmas at *, assumption end
    | kty lty mty xity zetaty rhoty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_ty xity zetaty rhoty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty xity zetaty rhoty Eqty 
    (all _ s0) := have h : _, from congr_all (compRenRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upRen_ty_ty rhoty) (up_ren_ren xity zetaty rhoty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenRen'_ty { kty : nat } { lty : nat } { mty : nat } { xity : Fin mty -> Fin kty } { zetaty : Fin kty -> Fin lty } { rhoty : Fin mty -> Fin lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqsty : s' = ren_ty xity s) : ren_ty zetaty s' = ren_ty rhoty s :=
have h : _, from compRenRen_ty xity zetaty rhoty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def up_ren_subst_ty_ty { k } { l } { mty : nat } (xi : Fin k -> Fin l) (tau : Fin l -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_ty_ty tau) (upRen_ty_ty xi)) x = (up_ty_ty theta) x :=
  λ n , match n with
  | (some n) := ap (ren_ty shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def compRenSubst_ty : Π { kty : nat } { lty : nat } { mty : nat } (xity : Fin mty -> Fin kty) (tauty : Fin kty -> ty lty) (thetaty : Fin mty -> ty lty) (Eqty : forall x, (comp tauty xity) x = thetaty x) (s : ty mty), subst_ty tauty (ren_ty xity s) = subst_ty thetaty s
    | kty lty mty xity tauty thetaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | kty lty mty xity tauty thetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_ty xity tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty xity tauty thetaty Eqty 
    (all _ s0) := have h : _, from congr_all (compRenSubst_ty (upRen_ty_ty xity) (up_ty_ty tauty) (up_ty_ty thetaty) (up_ren_subst_ty_ty xity tauty thetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compRenSubst'_ty { kty : nat } { lty : nat } { mty : nat } { xity : Fin mty -> Fin kty } { tauty : Fin kty -> ty lty } { thetaty : Fin mty -> ty lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqsty : s' = ren_ty xity s) : subst_ty tauty s' = subst_ty thetaty s :=
have h : _, from compRenSubst_ty xity tauty thetaty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def up_subst_ren_ty_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (zetaty : Fin lty -> Fin mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (ren_ty zetaty) sigma) x = theta x) : forall x, (comp (ren_ty (upRen_ty_ty zetaty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenRen_ty shift (upRen_ty_ty zetaty) (comp shift zetaty) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_ty zetaty shift (comp shift zetaty) (λ x , by refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstRen_ty : Π { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (zetaty : Fin kty -> Fin lty) (thetaty : Fin mty -> ty lty) (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (s : ty mty), ren_ty zetaty (subst_ty sigmaty s) = subst_ty thetaty s
    | kty lty mty sigmaty zetaty thetaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty zetaty thetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compSubstRen_ty sigmaty zetaty thetaty Eqty s0) (compSubstRen_ty sigmaty zetaty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty zetaty thetaty Eqty 
    (all _ s0) := have h : _, from congr_all (compSubstRen_ty (up_ty_ty sigmaty) (upRen_ty_ty zetaty) (up_ty_ty thetaty) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstRen'_ty { kty : nat } { lty : nat } { mty : nat } { sigmaty : Fin mty -> ty kty } { zetaty : Fin kty -> Fin lty } { thetaty : Fin mty -> ty lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqsty : s' = subst_ty sigmaty s) : ren_ty zetaty s' = subst_ty thetaty s :=
have h : _, from compSubstRen_ty sigmaty zetaty thetaty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def up_subst_subst_ty_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (tauty : Fin lty -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (subst_ty tauty) sigma) x = theta x) : forall x, (comp (subst_ty (up_ty_ty tauty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenSubst_ty shift (up_ty_ty tauty) (comp (up_ty_ty tauty) shift) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_ty tauty shift (comp (ren_ty shift) tauty) (λ x , by refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
def compSubstSubst_ty : Π { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (tauty : Fin kty -> ty lty) (thetaty : Fin mty -> ty lty) (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (s : ty mty), subst_ty tauty (subst_ty sigmaty s) = subst_ty thetaty s
    | kty lty mty sigmaty tauty thetaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty tauty thetaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty lty mty sigmaty tauty thetaty Eqty 
    (all _ s0) := have h : _, from congr_all (compSubstSubst_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (up_ty_ty thetaty) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma compSubstSubst'_ty { kty : nat } { lty : nat } { mty : nat } { sigmaty : Fin mty -> ty kty } { tauty : Fin kty -> ty lty } { thetaty : Fin mty -> ty lty } { s : ty mty } { s' : ty kty } (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqsty : s' = subst_ty sigmaty s) : subst_ty tauty s' = subst_ty thetaty s :=
have h : _, from compSubstSubst_ty sigmaty tauty thetaty Eqty s, begin rw Eqsty, assumption end

@[substLemmas]
def rinstInst_up_ty_ty { m } { nty : nat } (xi : Fin m -> Fin nty) (sigma : Fin m -> ty nty) (Eq : forall x, (comp (var_ty nty) xi) x = sigma x) : forall x, (comp (var_ty (nat.succ nty)) (upRen_ty_ty xi)) x = (up_ty_ty sigma) x :=
  λ n , match n with
  | (some n) := ap (ren_ty shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
def rinst_inst_ty : Π { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) (sigmaty : Fin mty -> ty nty) (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (s : ty mty), ren_ty xity s = subst_ty sigmaty s
    | mty nty xity sigmaty Eqty  (var_ty _ s) := have h : _, from Eqty s, begin simp with substLemmas at *, assumption end
    | mty nty xity sigmaty Eqty 
    (arr _ s0 s1) := have h : _, from congr_arr (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_ty xity sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty nty xity sigmaty Eqty 
    (all _ s0) := have h : _, from congr_all (rinst_inst_ty (upRen_ty_ty xity) (up_ty_ty sigmaty) (rinstInst_up_ty_ty xity sigmaty Eqty) s0), begin simp with substLemmas at *, assumption end

@[substLemmas]
lemma rinst_inst'_ty { mty : nat } { nty : nat } { sigmaty : Fin mty -> ty nty } { xity : Fin mty -> Fin nty } { s s' : ty mty } (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqsty : s = s') : ren_ty xity s' = subst_ty sigmaty s :=
have h : _, from rinst_inst_ty xity sigmaty Eqty s, begin rw <- Eqsty, assumption end

@[substLemmas]
lemma instId_ty { mty : nat } : subst_ty (var_ty mty) = id :=
 (λ x , idSubst_ty (var_ty mty) (λ n , by refl) (id x))

@[substLemmas]
lemma varL_ty { mty : nat } { nty : nat } (sigmaty : Fin mty -> ty nty) : comp (subst_ty sigmaty) (var_ty mty) = sigmaty :=
 (λ x , by refl)

@[substLemmas]
lemma rinstInst_ty { mty : nat } { nty : nat } (xity : Fin mty -> Fin nty) : ren_ty xity = subst_ty (comp (var_ty nty) xity) :=
 (λ x , rinst_inst_ty xity (_) (λ n , by refl) x)

@[substLemmas]
lemma compComp_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (tauty : Fin kty -> ty lty) (s : ty mty) : subst_ty tauty (subst_ty sigmaty s) = subst_ty (comp (subst_ty tauty) sigmaty) s :=
compSubstSubst_ty sigmaty tauty (_) (λ n , by refl) s

@[substLemmas]
lemma compComp'_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : Fin mty -> ty kty) (tauty : Fin kty -> ty lty) : comp (subst_ty tauty) (subst_ty sigmaty) = subst_ty (comp (subst_ty tauty) sigmaty) :=
 (λ n , compComp_ty sigmaty tauty n)

meta def is_arr_ty : expr -> bool
  | `(arr _ %%s0 %%s1) := tt
  | `(congr_arr' %%s0 %%s1) := tt
  | `(idSubst'_ty %%Eqty %%Eqsty) := tt
  | _ := ff

meta def destruct_arr_ty : expr -> tactic (expr × expr)
  | `(arr _ %%s0 %%s1) := return (s0, s1)
  | `(congr_arr' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_ty %%Eqty %%Eqsty) := destruct_arr_ty Eqsty
  | e := return (e, e)

meta def is_all_ty : expr -> bool
  | `(all _ %%s0) := tt
  | `(congr_all' %%s0) := tt
  | `(idSubst'_ty %%Eqty %%Eqsty) := tt
  | _ := ff

meta def destruct_all_ty : expr -> tactic expr
  | `(all _ %%s0) := return s0
  | `(congr_all' %%s0) := return s0
  | `(idSubst'_ty %%Eqty %%Eqsty) := destruct_all_ty Eqsty
  | e := return e

meta def is_InstRen_ty : expr -> bool
  | `(extRen'_ty %%Eqty %%Eqsty) := tt
  | `(compRenRen'_ty %%Eqty %%Eqsty) := tt
  | _ := ff

meta def destruct_InstRen_ty : expr -> tactic (expr × expr)
  | `(extRen'_ty %%Eqty %%Eqsty) := return (Eqty, Eqsty)
  | `(compRenRen'_ty %%Eqty %%Eqsty) := destruct_InstRen_ty Eqsty
  | `(%%e) := return (e,e)

meta def is_Inst_ty : expr -> bool
  | `(compRenRen'_ty %%Eqty %%Eqsty) := tt
  | `(compRenSubst'_ty %%Eqty %%Eqsty) := tt
  | `(compSubstRen'_ty %%Eqty %%Eqsty) := tt
  | `(compSubstSubst'_ty %%Eqty %%Eqsty) := tt
  | `(idSubst'_ty %%Eqty %%Eqsty) := tt
  | `(rinst_inst'_ty %%Eqty %%Eqsty) := tt
  | _ := ff

meta def destruct_Inst_ty : expr -> tactic (expr × expr)
  | `(ext'_ty %%Eqty %%Eqsty) := return (Eqty, Eqsty)
  | `(compRenRen'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(compRenSubst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(compSubstRen'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(compSubstSubst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(idSubst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(rinst_inst'_ty %%Eqty %%Eqsty) := destruct_Inst_ty Eqsty
  | `(%%e) := return (e,e)

meta def is_scons : expr -> bool
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(scons %%H1 %%H2) := tt
  | _ := ff

meta def is_comp : expr -> bool
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(comp %%H1 %%H2) := tt
  | _ := ff

meta def is_var_ty : expr -> bool
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(var_ty %%H1 %%H2) := tt
  | _ := ff

meta def destruct_scons : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := destruct_scons Eq
  | `(scons %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_comp : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_ty %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_ty_ty %%sigma %%zetaty %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_ty_ty %%sigma %%tauty %%theta %%Eq) := destruct_comp Eq
  | `(comp %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def normalize'_ty : nat -> expr -> expr -> tactic expr
  | n e1 e2 := if (n = Inst)
  then if (e2 = var_ty)
  then idSubst'_ty e1 e2
  else subst_ty e1 e2
  else if (n = Rinst)
  then if (e2 = var_ty)
  then idSubst'_ty e1 e2
  else if (is_App_ty e1)
  then TODO
  else subst_ty e1 e2
  else if (n = Comp)
  then TODO
  else if (n = Up)
  then TODO
  else ex ``(%%e1=%%e2)

@[substLemmas]
mutual inductive tm,vl 
with tm  : nat -> nat -> Type 
  | app : Π (nty nvl : nat), tm nty nvl -> tm nty nvl -> tm nty nvl
  | tapp : Π (nty nvl : nat), tm nty nvl -> ty nty -> tm nty nvl
  | vt : Π (nty nvl : nat), vl nty nvl -> tm nty nvl
 with vl  : nat -> nat -> Type 
  | var_vl : Π (nty nvl : nat), Fin nvl -> vl nty nvl
  | lam : Π (nty nvl : nat), ty nty -> tm nty (nat.succ nvl) -> vl nty nvl
  | tlam : Π (nty nvl : nat), tm (nat.succ nty) nvl -> vl nty nvl
open tm vl

@[substLemmas]
lemma congr_app { mty mvl : nat } { s0 : tm mty mvl } { s1 : tm mty mvl } { t0 : tm mty mvl } { t1 : tm mty mvl } : s0 = t0 -> s1 = t1 -> app mty mvl s0 s1 = app mty mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tapp { mty mvl : nat } { s0 : tm mty mvl } { s1 : ty mty } { t0 : tm mty mvl } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> tapp mty mvl s0 s1 = tapp mty mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_vt { mty mvl : nat } { s0 : vl mty mvl } { t0 : vl mty mvl } : s0 = t0 -> vt mty mvl s0 = vt mty mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_app' { mty mvl : nat } { s0 : tm mty mvl } { s1 : tm mty mvl } { t0 : tm mty mvl } { t1 : tm mty mvl } (h0 : s0 = t0) (h1 : s1 = t1) : app mty mvl s0 s1 = app mty mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tapp' { mty mvl : nat } { s0 : tm mty mvl } { s1 : ty mty } { t0 : tm mty mvl } { t1 : ty mty } (h0 : s0 = t0) (h1 : s1 = t1) : tapp mty mvl s0 s1 = tapp mty mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_vt' { mty mvl : nat } { s0 : vl mty mvl } { t0 : vl mty mvl } (h0 : s0 = t0) : vt mty mvl s0 = vt mty mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam { mty mvl : nat } { s0 : ty mty } { s1 : tm mty (nat.succ mvl) } { t0 : ty mty } { t1 : tm mty (nat.succ mvl) } : s0 = t0 -> s1 = t1 -> lam mty mvl s0 s1 = lam mty mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tlam { mty mvl : nat } { s0 : tm (nat.succ mty) mvl } { t0 : tm (nat.succ mty) mvl } : s0 = t0 -> tlam mty mvl s0 = tlam mty mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_lam' { mty mvl : nat } { s0 : ty mty } { s1 : tm mty (nat.succ mvl) } { t0 : ty mty } { t1 : tm mty (nat.succ mvl) } (h0 : s0 = t0) (h1 : s1 = t1) : lam mty mvl s0 s1 = lam mty mvl t0 t1 :=
begin intros, congr; assumption end

@[substLemmas]
lemma congr_tlam' { mty mvl : nat } { s0 : tm (nat.succ mty) mvl } { t0 : tm (nat.succ mty) mvl } (h0 : s0 = t0) : tlam mty mvl s0 = tlam mty mvl t0 :=
begin intros, congr; assumption end

@[substLemmas]
def upRen_ty_vl { m } { n } (xi : Fin m -> Fin n) : _ :=
  xi

@[substLemmas]
def upRen_vl_ty { m } { n } (xi : Fin m -> Fin n) : _ :=
  xi

@[substLemmas]
def upRen_vl_vl { m } { n } (xi : Fin m -> Fin n) : _ :=
  up_ren xi

@[substLemmas]
mutual def ren_tm,ren_vl
with ren_tm : Π { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) (s : tm mty mvl), tm nty nvl
    | mty mvl nty nvl xity xivl  (app _ s0 s1) := app nty nvl (ren_tm xity xivl s0) (ren_tm xity xivl s1)
    | mty mvl nty nvl xity xivl  (tapp _ s0 s1) := tapp nty nvl (ren_tm xity xivl s0) (ren_ty xity s1)
    | mty mvl nty nvl xity xivl  (vt _ s0) := vt nty nvl (ren_vl xity xivl s0)
 with ren_vl : Π { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) (s : vl mty mvl), vl nty nvl
    | mty mvl nty nvl xity xivl  (var_vl _ _ s) := (var_vl nty nvl) (xivl s)
    | mty mvl nty nvl xity xivl  (lam _ s0 s1) := lam nty nvl (ren_ty xity s0) (ren_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) s1)
    | mty mvl nty nvl xity xivl  (tlam _ s0) := tlam nty nvl (ren_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) s0)
using_well_founded wf_single_arg

@[substLemmas]
def up_ty_vl { m } { nty nvl : nat } (sigma : Fin m -> vl nty nvl) : _ :=
  comp (ren_vl shift id) sigma

@[substLemmas]
def up_vl_ty { m } { nty : nat } (sigma : Fin m -> ty nty) : _ :=
  comp (ren_ty id) sigma

@[substLemmas]
def up_vl_vl { m } { nty nvl : nat } (sigma : Fin m -> vl nty nvl) : _ :=
  scons ((var_vl nty (nat.succ nvl)) var_zero) (comp (ren_vl id shift) sigma)

@[substLemmas]
mutual def subst_tm,subst_vl
with subst_tm : Π { mty mvl : nat } { nty nvl : nat } (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) (s : tm mty mvl), tm nty nvl
    | mty mvl nty nvl sigmaty sigmavl  (app _ s0 s1) := app nty nvl (subst_tm sigmaty sigmavl s0) (subst_tm sigmaty sigmavl s1)
    | mty mvl nty nvl sigmaty sigmavl  (tapp _ s0 s1) := tapp nty nvl (subst_tm sigmaty sigmavl s0) (subst_ty sigmaty s1)
    | mty mvl nty nvl sigmaty sigmavl  (vt _ s0) := vt nty nvl (subst_vl sigmaty sigmavl s0)
 with subst_vl : Π { mty mvl : nat } { nty nvl : nat } (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) (s : vl mty mvl), vl nty nvl
    | mty mvl nty nvl sigmaty sigmavl  (var_vl _ _ s) := sigmavl s
    | mty mvl nty nvl sigmaty sigmavl 
    (lam _ s0 s1) := lam nty nvl (subst_ty sigmaty s0) (subst_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) s1)
    | mty mvl nty nvl sigmaty sigmavl  (tlam _ s0) := tlam nty nvl (subst_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) s0)
using_well_founded wf_single_arg

@[substLemmas]
def upId_ty_vl { mty mvl : nat } { sigma : Fin mvl -> vl mty mvl } (Eq : forall x, sigma x = (var_vl mty mvl) x) : forall x, (up_ty_vl sigma) x = (var_vl (nat.succ mty) mvl) x :=
  λ n , have h : _, from ap (ren_vl shift id) (Eq n), begin simp with substLemmas at *, assumption end

@[substLemmas]
def upId_vl_ty { mty : nat } { sigma : Fin mty -> ty mty } (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_vl_ty sigma) x = (var_ty mty) x :=
  λ n , have h : _, from ap (ren_ty id) (Eq n), begin simp with substLemmas at *, assumption end

@[substLemmas]
def upId_vl_vl { mty mvl : nat } { sigma : Fin mvl -> vl mty mvl } (Eq : forall x, sigma x = (var_vl mty mvl) x) : forall x, (up_vl_vl sigma) x = (var_vl mty (nat.succ mvl)) x :=
  λ n , match n with
  | (some n) := have h : _, from ap (ren_vl id shift) (Eq n), begin simp with substLemmas at *, assumption end
  | none := by refl
  end

@[substLemmas]
mutual def idSubst_tm,idSubst_vl
with idSubst_tm : Π { mty mvl : nat } (sigmaty : Fin mty -> ty mty) (sigmavl : Fin mvl -> vl mty mvl) (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqvl : forall x, sigmavl x = (var_vl mty mvl) x) (s : tm mty mvl), subst_tm sigmaty sigmavl s = s
    | mty mvl sigmaty sigmavl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (idSubst_tm sigmaty sigmavl Eqty Eqvl s0) (idSubst_tm sigmaty sigmavl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | mty mvl sigmaty sigmavl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (idSubst_tm sigmaty sigmavl Eqty Eqvl s0) (idSubst_ty sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mvl sigmaty sigmavl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (idSubst_vl sigmaty sigmavl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with idSubst_vl : Π { mty mvl : nat } (sigmaty : Fin mty -> ty mty) (sigmavl : Fin mvl -> vl mty mvl) (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqvl : forall x, sigmavl x = (var_vl mty mvl) x) (s : vl mty mvl), subst_vl sigmaty sigmavl s = s
    | mty mvl sigmaty sigmavl Eqty Eqvl  (var_vl _ _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | mty mvl sigmaty sigmavl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (idSubst_ty sigmaty Eqty s0) (idSubst_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (upId_vl_ty (_) Eqty) (upId_vl_vl (_) Eqvl) s1), begin simp with substLemmas at *, assumption end
    | mty mvl sigmaty sigmavl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (idSubst_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (upId_ty_ty (_) Eqty) (upId_ty_vl (_) Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma idSubst'_tm { mty mvl : nat } { sigmaty : Fin mty -> ty mty } { sigmavl : Fin mvl -> vl mty mvl } { Eqty : forall x, sigmaty x = (var_ty mty) x } { Eqvl : forall x, sigmavl x = (var_vl mty mvl) x } { s s' : tm mty mvl } (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqvl : forall x, sigmavl x = (var_vl mty mvl) x) (Eqstm : s = s') : subst_tm sigmaty sigmavl s = s' :=
eq.trans (idSubst_tm sigmaty sigmavl Eqty Eqvl s) Eqstm

@[substLemmas]
lemma idSubst'_vl { mty mvl : nat } { sigmaty : Fin mty -> ty mty } { sigmavl : Fin mvl -> vl mty mvl } { Eqty : forall x, sigmaty x = (var_ty mty) x } { Eqvl : forall x, sigmavl x = (var_vl mty mvl) x } { s s' : vl mty mvl } (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqvl : forall x, sigmavl x = (var_vl mty mvl) x) (Eqsvl : s = s') : subst_vl sigmaty sigmavl s = s' :=
eq.trans (idSubst_vl sigmaty sigmavl Eqty Eqvl s) Eqsvl

@[substLemmas]
def upExtRen_ty_vl { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_vl xi) x = (upRen_ty_vl zeta) x :=
  λ n , Eq n

@[substLemmas]
def upExtRen_vl_ty { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_vl_ty xi) x = (upRen_vl_ty zeta) x :=
  λ n , Eq n

@[substLemmas]
def upExtRen_vl_vl { m } { n } (xi : Fin m -> Fin n) (zeta : Fin m -> Fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_vl_vl xi) x = (upRen_vl_vl zeta) x :=
  λ n , match n with
  | (some n) := ap shift (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def extRen_tm,extRen_vl
with extRen_tm : Π { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) (zetaty : Fin mty -> Fin nty) (zetavl : Fin mvl -> Fin nvl) (Eqty : forall x, xity x = zetaty x) (Eqvl : forall x, xivl x = zetavl x) (s : tm mty mvl), ren_tm xity xivl s = ren_tm zetaty zetavl s
    | mty mvl nty nvl xity xivl zetaty zetavl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (extRen_tm xity xivl zetaty zetavl Eqty Eqvl s0) (extRen_tm xity xivl zetaty zetavl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl zetaty zetavl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (extRen_tm xity xivl zetaty zetavl Eqty Eqvl s0) (extRen_ty xity zetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl zetaty zetavl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (extRen_vl xity xivl zetaty zetavl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with extRen_vl : Π { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) (zetaty : Fin mty -> Fin nty) (zetavl : Fin mvl -> Fin nvl) (Eqty : forall x, xity x = zetaty x) (Eqvl : forall x, xivl x = zetavl x) (s : vl mty mvl), ren_vl xity xivl s = ren_vl zetaty zetavl s
    | mty mvl nty nvl xity xivl zetaty zetavl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from ap (var_vl nty nvl) (Eqvl s), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl zetaty zetavl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (extRen_ty xity zetaty Eqty s0) (extRen_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (upExtRen_vl_ty xity zetaty Eqty) (upExtRen_vl_vl xivl zetavl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl zetaty zetavl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (extRen_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (upExtRen_ty_ty xity zetaty Eqty) (upExtRen_ty_vl xivl zetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma extRen'_tm { mty mvl : nat } { nty nvl : nat } { xity : Fin mty -> Fin nty } { xivl : Fin mvl -> Fin nvl } { zetaty : Fin mty -> Fin nty } { zetavl : Fin mvl -> Fin nvl } { s s' : tm mty mvl } (Eqty : forall x, xity x = zetaty x) (Eqvl : forall x, xivl x = zetavl x) (Eqstm : s = s') : ren_tm xity xivl s = ren_tm zetaty zetavl s' :=
have h : _, from extRen_tm xity xivl zetaty zetavl Eqty Eqvl s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma extRen'_vl { mty mvl : nat } { nty nvl : nat } { xity : Fin mty -> Fin nty } { xivl : Fin mvl -> Fin nvl } { zetaty : Fin mty -> Fin nty } { zetavl : Fin mvl -> Fin nvl } { s s' : vl mty mvl } (Eqty : forall x, xity x = zetaty x) (Eqvl : forall x, xivl x = zetavl x) (Eqsvl : s = s') : ren_vl xity xivl s = ren_vl zetaty zetavl s' :=
have h : _, from extRen_vl xity xivl zetaty zetavl Eqty Eqvl s, begin rw <- Eqsvl, assumption end

@[substLemmas]
def upExt_ty_vl { m } { nty nvl : nat } (sigma : Fin m -> vl nty nvl) (tau : Fin m -> vl nty nvl) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_vl sigma) x = (up_ty_vl tau) x :=
  λ n , ap (ren_vl shift id) (Eq n)

@[substLemmas]
def upExt_vl_ty { m } { nty : nat } (sigma : Fin m -> ty nty) (tau : Fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_vl_ty sigma) x = (up_vl_ty tau) x :=
  λ n , ap (ren_ty id) (Eq n)

@[substLemmas]
def upExt_vl_vl { m } { nty nvl : nat } (sigma : Fin m -> vl nty nvl) (tau : Fin m -> vl nty nvl) (Eq : forall x, sigma x = tau x) : forall x, (up_vl_vl sigma) x = (up_vl_vl tau) x :=
  λ n , match n with
  | (some n) := ap (ren_vl id shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def ext_tm,ext_vl
with ext_tm : Π { mty mvl : nat } { nty nvl : nat } (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) (tauty : Fin mty -> ty nty) (tauvl : Fin mvl -> vl nty nvl) (Eqty : forall x, sigmaty x = tauty x) (Eqvl : forall x, sigmavl x = tauvl x) (s : tm mty mvl), subst_tm sigmaty sigmavl s = subst_tm tauty tauvl s
    | mty mvl nty nvl sigmaty sigmavl tauty tauvl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s0) (ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl sigmaty sigmavl tauty tauvl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s0) (ext_ty sigmaty tauty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl sigmaty sigmavl tauty tauvl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (ext_vl sigmaty sigmavl tauty tauvl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with ext_vl : Π { mty mvl : nat } { nty nvl : nat } (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) (tauty : Fin mty -> ty nty) (tauvl : Fin mvl -> vl nty nvl) (Eqty : forall x, sigmaty x = tauty x) (Eqvl : forall x, sigmavl x = tauvl x) (s : vl mty mvl), subst_vl sigmaty sigmavl s = subst_vl tauty tauvl s
    | mty mvl nty nvl sigmaty sigmavl tauty tauvl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl sigmaty sigmavl tauty tauvl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (ext_ty sigmaty tauty Eqty s0) (ext_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (up_vl_ty tauty) (up_vl_vl tauvl) (upExt_vl_ty sigmaty tauty Eqty) (upExt_vl_vl sigmavl tauvl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl sigmaty sigmavl tauty tauvl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (ext_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (up_ty_ty tauty) (up_ty_vl tauvl) (upExt_ty_ty sigmaty tauty Eqty) (upExt_ty_vl sigmavl tauvl Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma ext'_tm { mty mvl : nat } { nty nvl : nat } { sigmaty : Fin mty -> ty nty } { sigmavl : Fin mvl -> vl nty nvl } { tauty : Fin mty -> ty nty } { tauvl : Fin mvl -> vl nty nvl } { s s' : tm mty mvl } (Eqty : forall x, sigmaty x = tauty x) (Eqvl : forall x, sigmavl x = tauvl x) (Eqstm : s = s') : subst_tm sigmaty sigmavl s = subst_tm tauty tauvl s' :=
have h : _, from ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma ext'_vl { mty mvl : nat } { nty nvl : nat } { sigmaty : Fin mty -> ty nty } { sigmavl : Fin mvl -> vl nty nvl } { tauty : Fin mty -> ty nty } { tauvl : Fin mvl -> vl nty nvl } { s s' : vl mty mvl } (Eqty : forall x, sigmaty x = tauty x) (Eqvl : forall x, sigmavl x = tauvl x) (Eqsvl : s = s') : subst_vl sigmaty sigmavl s = subst_vl tauty tauvl s' :=
have h : _, from ext_vl sigmaty sigmavl tauty tauvl Eqty Eqvl s, begin rw <- Eqsvl, assumption end

@[substLemmas]
mutual def compRenRen_tm,compRenRen_vl
with compRenRen_tm : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : Fin mty -> Fin kty) (xivl : Fin mvl -> Fin kvl) (zetaty : Fin kty -> Fin lty) (zetavl : Fin kvl -> Fin lvl) (rhoty : Fin mty -> Fin lty) (rhovl : Fin mvl -> Fin lvl) (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (s : tm mty mvl), ren_tm zetaty zetavl (ren_tm xity xivl s) = ren_tm rhoty rhovl s
    | kty kvl lty lvl mty mvl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s0) (compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s0) (compRenRen_ty xity zetaty rhoty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (compRenRen_vl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with compRenRen_vl : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : Fin mty -> Fin kty) (xivl : Fin mvl -> Fin kvl) (zetaty : Fin kty -> Fin lty) (zetavl : Fin kvl -> Fin lvl) (rhoty : Fin mty -> Fin lty) (rhovl : Fin mvl -> Fin lvl) (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (s : vl mty mvl), ren_vl zetaty zetavl (ren_vl xity xivl s) = ren_vl rhoty rhovl s
    | kty kvl lty lvl mty mvl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from ap (var_vl lty lvl) (Eqvl s), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (upRen_vl_ty rhoty) (upRen_vl_vl rhovl) Eqty (up_ren_ren xivl zetavl rhovl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (compRenRen_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (upRen_ty_ty rhoty) (upRen_ty_vl rhovl) (up_ren_ren xity zetaty rhoty Eqty) Eqvl s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compRenRen'_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { xity : Fin mty -> Fin kty } { xivl : Fin mvl -> Fin kvl } { zetaty : Fin kty -> Fin lty } { zetavl : Fin kvl -> Fin lvl } { rhoty : Fin mty -> Fin lty } { rhovl : Fin mvl -> Fin lvl } { s : tm mty mvl } { s' : tm kty kvl } (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (Eqstm : s' = ren_tm xity xivl s) : ren_tm zetaty zetavl s' = ren_tm rhoty rhovl s :=
have h : _, from compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compRenRen'_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { xity : Fin mty -> Fin kty } { xivl : Fin mvl -> Fin kvl } { zetaty : Fin kty -> Fin lty } { zetavl : Fin kvl -> Fin lvl } { rhoty : Fin mty -> Fin lty } { rhovl : Fin mvl -> Fin lvl } { s : vl mty mvl } { s' : vl kty kvl } (Eqty : forall x, (comp zetaty xity) x = rhoty x) (Eqvl : forall x, (comp zetavl xivl) x = rhovl x) (Eqsvl : s' = ren_vl xity xivl s) : ren_vl zetaty zetavl s' = ren_vl rhoty rhovl s :=
have h : _, from compRenRen_vl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def up_ren_subst_ty_vl { k } { l } { mty mvl : nat } (xi : Fin k -> Fin l) (tau : Fin l -> vl mty mvl) (theta : Fin k -> vl mty mvl) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_ty_vl tau) (upRen_ty_vl xi)) x = (up_ty_vl theta) x :=
  λ n , ap (ren_vl shift id) (Eq n)

@[substLemmas]
def up_ren_subst_vl_ty { k } { l } { mty : nat } (xi : Fin k -> Fin l) (tau : Fin l -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_vl_ty tau) (upRen_vl_ty xi)) x = (up_vl_ty theta) x :=
  λ n , ap (ren_ty id) (Eq n)

@[substLemmas]
def up_ren_subst_vl_vl { k } { l } { mty mvl : nat } (xi : Fin k -> Fin l) (tau : Fin l -> vl mty mvl) (theta : Fin k -> vl mty mvl) (Eq : forall x, (comp tau xi) x = theta x) : forall x, (comp (up_vl_vl tau) (upRen_vl_vl xi)) x = (up_vl_vl theta) x :=
  λ n , match n with
  | (some n) := ap (ren_vl id shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def compRenSubst_tm,compRenSubst_vl
with compRenSubst_tm : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : Fin mty -> Fin kty) (xivl : Fin mvl -> Fin kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) (thetaty : Fin mty -> ty lty) (thetavl : Fin mvl -> vl lty lvl) (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (s : tm mty mvl), subst_tm tauty tauvl (ren_tm xity xivl s) = subst_tm thetaty thetavl s
    | kty kvl lty lvl mty mvl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compRenSubst_ty xity tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (compRenSubst_vl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with compRenSubst_vl : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : Fin mty -> Fin kty) (xivl : Fin mvl -> Fin kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) (thetaty : Fin mty -> ty lty) (thetavl : Fin mvl -> vl lty lvl) (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (s : vl mty mvl), subst_vl tauty tauvl (ren_vl xity xivl s) = subst_vl thetaty thetavl s
    | kty kvl lty lvl mty mvl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (up_vl_ty tauty) (up_vl_vl tauvl) (up_vl_ty thetaty) (up_vl_vl thetavl) (up_ren_subst_vl_ty xity tauty thetaty Eqty) (up_ren_subst_vl_vl xivl tauvl thetavl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (compRenSubst_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (up_ty_ty tauty) (up_ty_vl tauvl) (up_ty_ty thetaty) (up_ty_vl thetavl) (up_ren_subst_ty_ty xity tauty thetaty Eqty) (up_ren_subst_ty_vl xivl tauvl thetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compRenSubst'_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { xity : Fin mty -> Fin kty } { xivl : Fin mvl -> Fin kvl } { tauty : Fin kty -> ty lty } { tauvl : Fin kvl -> vl lty lvl } { thetaty : Fin mty -> ty lty } { thetavl : Fin mvl -> vl lty lvl } { s : tm mty mvl } { s' : tm kty kvl } (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (Eqstm : s' = ren_tm xity xivl s) : subst_tm tauty tauvl s' = subst_tm thetaty thetavl s :=
have h : _, from compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compRenSubst'_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { xity : Fin mty -> Fin kty } { xivl : Fin mvl -> Fin kvl } { tauty : Fin kty -> ty lty } { tauvl : Fin kvl -> vl lty lvl } { thetaty : Fin mty -> ty lty } { thetavl : Fin mvl -> vl lty lvl } { s : vl mty mvl } { s' : vl kty kvl } (Eqty : forall x, (comp tauty xity) x = thetaty x) (Eqvl : forall x, (comp tauvl xivl) x = thetavl x) (Eqsvl : s' = ren_vl xity xivl s) : subst_vl tauty tauvl s' = subst_vl thetaty thetavl s :=
have h : _, from compRenSubst_vl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def up_subst_ren_ty_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : Fin k -> vl lty lvl) (zetaty : Fin lty -> Fin mty) (zetavl : Fin lvl -> Fin mvl) (theta : Fin k -> vl mty mvl) (Eq : forall x, (comp (ren_vl zetaty zetavl) sigma) x = theta x) : forall x, (comp (ren_vl (upRen_ty_ty zetaty) (upRen_ty_vl zetavl)) (up_ty_vl sigma)) x = (up_ty_vl theta) x :=
  λ n , eq.trans (compRenRen_vl shift id (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (comp shift zetaty) (comp id zetavl) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_vl zetaty zetavl shift id (comp shift zetaty) (comp id zetavl) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_vl shift id) (Eq n)))

@[substLemmas]
def up_subst_ren_vl_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (zetaty : Fin lty -> Fin mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (ren_ty zetaty) sigma) x = theta x) : forall x, (comp (ren_ty (upRen_vl_ty zetaty)) (up_vl_ty sigma)) x = (up_vl_ty theta) x :=
  λ n , eq.trans (compRenRen_ty id (upRen_vl_ty zetaty) (comp id zetaty) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_ty zetaty id (comp id zetaty) (λ x , by refl) (sigma n))) (ap (ren_ty id) (Eq n)))

@[substLemmas]
def up_subst_ren_vl_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : Fin k -> vl lty lvl) (zetaty : Fin lty -> Fin mty) (zetavl : Fin lvl -> Fin mvl) (theta : Fin k -> vl mty mvl) (Eq : forall x, (comp (ren_vl zetaty zetavl) sigma) x = theta x) : forall x, (comp (ren_vl (upRen_vl_ty zetaty) (upRen_vl_vl zetavl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenRen_vl id shift (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (comp id zetaty) (comp shift zetavl) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compRenRen_vl zetaty zetavl id shift (comp id zetaty) (comp shift zetavl) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_vl id shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
mutual def compSubstRen_tm,compSubstRen_vl
with compSubstRen_tm : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (zetaty : Fin kty -> Fin lty) (zetavl : Fin kvl -> Fin lvl) (thetaty : Fin mty -> ty lty) (thetavl : Fin mvl -> vl lty lvl) (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (ren_vl zetaty zetavl) sigmavl) x = thetavl x) (s : tm mty mvl), ren_tm zetaty zetavl (subst_tm sigmaty sigmavl s) = subst_tm thetaty thetavl s
    | kty kvl lty lvl mty mvl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compSubstRen_tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s0) (compSubstRen_tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compSubstRen_tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s0) (compSubstRen_ty sigmaty zetaty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (compSubstRen_vl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with compSubstRen_vl : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (zetaty : Fin kty -> Fin lty) (zetavl : Fin kvl -> Fin lvl) (thetaty : Fin mty -> ty lty) (thetavl : Fin mvl -> vl lty lvl) (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (ren_vl zetaty zetavl) sigmavl) x = thetavl x) (s : vl mty mvl), ren_vl zetaty zetavl (subst_vl sigmaty sigmavl s) = subst_vl thetaty thetavl s
    | kty kvl lty lvl mty mvl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (compSubstRen_ty sigmaty zetaty thetaty Eqty s0) (compSubstRen_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (up_vl_ty thetaty) (up_vl_vl thetavl) (up_subst_ren_vl_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_vl_vl sigmavl zetaty zetavl thetavl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (compSubstRen_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (up_ty_ty thetaty) (up_ty_vl thetavl) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_ty_vl sigmavl zetaty zetavl thetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compSubstRen'_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { sigmaty : Fin mty -> ty kty } { sigmavl : Fin mvl -> vl kty kvl } { zetaty : Fin kty -> Fin lty } { zetavl : Fin kvl -> Fin lvl } { thetaty : Fin mty -> ty lty } { thetavl : Fin mvl -> vl lty lvl } { s : tm mty mvl } { s' : tm kty kvl } (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (ren_vl zetaty zetavl) sigmavl) x = thetavl x) (Eqstm : s' = subst_tm sigmaty sigmavl s) : ren_tm zetaty zetavl s' = subst_tm thetaty thetavl s :=
have h : _, from compSubstRen_tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compSubstRen'_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { sigmaty : Fin mty -> ty kty } { sigmavl : Fin mvl -> vl kty kvl } { zetaty : Fin kty -> Fin lty } { zetavl : Fin kvl -> Fin lvl } { thetaty : Fin mty -> ty lty } { thetavl : Fin mvl -> vl lty lvl } { s : vl mty mvl } { s' : vl kty kvl } (Eqty : forall x, (comp (ren_ty zetaty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (ren_vl zetaty zetavl) sigmavl) x = thetavl x) (Eqsvl : s' = subst_vl sigmaty sigmavl s) : ren_vl zetaty zetavl s' = subst_vl thetaty thetavl s :=
have h : _, from compSubstRen_vl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def up_subst_subst_ty_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : Fin k -> vl lty lvl) (tauty : Fin lty -> ty mty) (tauvl : Fin lvl -> vl mty mvl) (theta : Fin k -> vl mty mvl) (Eq : forall x, (comp (subst_vl tauty tauvl) sigma) x = theta x) : forall x, (comp (subst_vl (up_ty_ty tauty) (up_ty_vl tauvl)) (up_ty_vl sigma)) x = (up_ty_vl theta) x :=
  λ n , eq.trans (compRenSubst_vl shift id (up_ty_ty tauty) (up_ty_vl tauvl) (comp (up_ty_ty tauty) shift) (comp (up_ty_vl tauvl) id) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_vl tauty tauvl shift id (comp (ren_ty shift) tauty) (comp (ren_vl shift id) tauvl) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_vl shift id) (Eq n)))

@[substLemmas]
def up_subst_subst_vl_ty { k } { lty : nat } { mty : nat } (sigma : Fin k -> ty lty) (tauty : Fin lty -> ty mty) (theta : Fin k -> ty mty) (Eq : forall x, (comp (subst_ty tauty) sigma) x = theta x) : forall x, (comp (subst_ty (up_vl_ty tauty)) (up_vl_ty sigma)) x = (up_vl_ty theta) x :=
  λ n , eq.trans (compRenSubst_ty id (up_vl_ty tauty) (comp (up_vl_ty tauty) id) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_ty tauty id (comp (ren_ty id) tauty) (λ x , by refl) (sigma n))) (ap (ren_ty id) (Eq n)))

@[substLemmas]
def up_subst_subst_vl_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : Fin k -> vl lty lvl) (tauty : Fin lty -> ty mty) (tauvl : Fin lvl -> vl mty mvl) (theta : Fin k -> vl mty mvl) (Eq : forall x, (comp (subst_vl tauty tauvl) sigma) x = theta x) : forall x, (comp (subst_vl (up_vl_ty tauty) (up_vl_vl tauvl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  λ n , match n with
  | (some n) := eq.trans (compRenSubst_vl id shift (up_vl_ty tauty) (up_vl_vl tauvl) (comp (up_vl_ty tauty) id) (comp (up_vl_vl tauvl) shift) (λ x , by refl) (λ x , by refl) (sigma n)) (eq.trans (eq.symm (compSubstRen_vl tauty tauvl id shift (comp (ren_ty id) tauty) (comp (ren_vl id shift) tauvl) (λ x , by refl) (λ x , by refl) (sigma n))) (ap (ren_vl id shift) (Eq n)))
  | none := by refl
  end

@[substLemmas]
mutual def compSubstSubst_tm,compSubstSubst_vl
with compSubstSubst_tm : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) (thetaty : Fin mty -> ty lty) (thetavl : Fin mvl -> vl lty lvl) (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (subst_vl tauty tauvl) sigmavl) x = thetavl x) (s : tm mty mvl), subst_tm tauty tauvl (subst_tm sigmaty sigmavl s) = subst_tm thetaty thetavl s
    | kty kvl lty lvl mty mvl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (compSubstSubst_vl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with compSubstSubst_vl : Π { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) (thetaty : Fin mty -> ty lty) (thetavl : Fin mvl -> vl lty lvl) (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (subst_vl tauty tauvl) sigmavl) x = thetavl x) (s : vl mty mvl), subst_vl tauty tauvl (subst_vl sigmaty sigmavl s) = subst_vl thetaty thetavl s
    | kty kvl lty lvl mty mvl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (up_vl_ty tauty) (up_vl_vl tauvl) (up_vl_ty thetaty) (up_vl_vl thetavl) (up_subst_subst_vl_ty sigmaty tauty thetaty Eqty) (up_subst_subst_vl_vl sigmavl tauty tauvl thetavl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | kty kvl lty lvl mty mvl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (compSubstSubst_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (up_ty_ty tauty) (up_ty_vl tauvl) (up_ty_ty thetaty) (up_ty_vl thetavl) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) (up_subst_subst_ty_vl sigmavl tauty tauvl thetavl Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma compSubstSubst'_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { sigmaty : Fin mty -> ty kty } { sigmavl : Fin mvl -> vl kty kvl } { tauty : Fin kty -> ty lty } { tauvl : Fin kvl -> vl lty lvl } { thetaty : Fin mty -> ty lty } { thetavl : Fin mvl -> vl lty lvl } { s : tm mty mvl } { s' : tm kty kvl } (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (subst_vl tauty tauvl) sigmavl) x = thetavl x) (Eqstm : s' = subst_tm sigmaty sigmavl s) : subst_tm tauty tauvl s' = subst_tm thetaty thetavl s :=
have h : _, from compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s, begin rw Eqstm, assumption end

@[substLemmas]
lemma compSubstSubst'_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } { sigmaty : Fin mty -> ty kty } { sigmavl : Fin mvl -> vl kty kvl } { tauty : Fin kty -> ty lty } { tauvl : Fin kvl -> vl lty lvl } { thetaty : Fin mty -> ty lty } { thetavl : Fin mvl -> vl lty lvl } { s : vl mty mvl } { s' : vl kty kvl } (Eqty : forall x, (comp (subst_ty tauty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp (subst_vl tauty tauvl) sigmavl) x = thetavl x) (Eqsvl : s' = subst_vl sigmaty sigmavl s) : subst_vl tauty tauvl s' = subst_vl thetaty thetavl s :=
have h : _, from compSubstSubst_vl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s, begin rw Eqsvl, assumption end

@[substLemmas]
def rinstInst_up_ty_vl { m } { nty nvl : nat } (xi : Fin m -> Fin nvl) (sigma : Fin m -> vl nty nvl) (Eq : forall x, (comp (var_vl nty nvl) xi) x = sigma x) : forall x, (comp (var_vl (nat.succ nty) nvl) (upRen_ty_vl xi)) x = (up_ty_vl sigma) x :=
  λ n , ap (ren_vl shift id) (Eq n)

@[substLemmas]
def rinstInst_up_vl_ty { m } { nty : nat } (xi : Fin m -> Fin nty) (sigma : Fin m -> ty nty) (Eq : forall x, (comp (var_ty nty) xi) x = sigma x) : forall x, (comp (var_ty nty) (upRen_vl_ty xi)) x = (up_vl_ty sigma) x :=
  λ n , ap (ren_ty id) (Eq n)

@[substLemmas]
def rinstInst_up_vl_vl { m } { nty nvl : nat } (xi : Fin m -> Fin nvl) (sigma : Fin m -> vl nty nvl) (Eq : forall x, (comp (var_vl nty nvl) xi) x = sigma x) : forall x, (comp (var_vl nty (nat.succ nvl)) (upRen_vl_vl xi)) x = (up_vl_vl sigma) x :=
  λ n , match n with
  | (some n) := ap (ren_vl id shift) (Eq n)
  | none := by refl
  end

@[substLemmas]
mutual def rinst_inst_tm,rinst_inst_vl
with rinst_inst_tm : Π { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqvl : forall x, (comp (var_vl nty nvl) xivl) x = sigmavl x) (s : tm mty mvl), ren_tm xity xivl s = subst_tm sigmaty sigmavl s
    | mty mvl nty nvl xity xivl sigmaty sigmavl Eqty Eqvl 
    (app _ s0 s1) := have h : _, from congr_app (rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s0) (rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl sigmaty sigmavl Eqty Eqvl 
    (tapp _ s0 s1) := have h : _, from congr_tapp (rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s0) (rinst_inst_ty xity sigmaty Eqty s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl sigmaty sigmavl Eqty Eqvl 
    (vt _ s0) := have h : _, from congr_vt (rinst_inst_vl xity xivl sigmaty sigmavl Eqty Eqvl s0), begin simp with substLemmas at *, assumption end
 with rinst_inst_vl : Π { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqvl : forall x, (comp (var_vl nty nvl) xivl) x = sigmavl x) (s : vl mty mvl), ren_vl xity xivl s = subst_vl sigmaty sigmavl s
    | mty mvl nty nvl xity xivl sigmaty sigmavl Eqty Eqvl 
    (var_vl _ _ s) := have h : _, from Eqvl s, begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl sigmaty sigmavl Eqty Eqvl 
    (lam _ s0 s1) := have h : _, from congr_lam (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (up_vl_ty sigmaty) (up_vl_vl sigmavl) (rinstInst_up_vl_ty xity sigmaty Eqty) (rinstInst_up_vl_vl xivl sigmavl Eqvl) s1), begin simp with substLemmas at *, assumption end
    | mty mvl nty nvl xity xivl sigmaty sigmavl Eqty Eqvl 
    (tlam _ s0) := have h : _, from congr_tlam (rinst_inst_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (up_ty_ty sigmaty) (up_ty_vl sigmavl) (rinstInst_up_ty_ty xity sigmaty Eqty) (rinstInst_up_ty_vl xivl sigmavl Eqvl) s0), begin simp with substLemmas at *, assumption end
using_well_founded wf_single_arg

@[substLemmas]
lemma rinst_inst'_tm { mty mvl : nat } { nty nvl : nat } { sigmaty : Fin mty -> ty nty } { sigmavl : Fin mvl -> vl nty nvl } { xity : Fin mty -> Fin nty } { xivl : Fin mvl -> Fin nvl } { s s' : tm mty mvl } (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqvl : forall x, (comp (var_vl nty nvl) xivl) x = sigmavl x) (Eqstm : s = s') : ren_tm xity xivl s' = subst_tm sigmaty sigmavl s :=
have h : _, from rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s, begin rw <- Eqstm, assumption end

@[substLemmas]
lemma rinst_inst'_vl { mty mvl : nat } { nty nvl : nat } { sigmaty : Fin mty -> ty nty } { sigmavl : Fin mvl -> vl nty nvl } { xity : Fin mty -> Fin nty } { xivl : Fin mvl -> Fin nvl } { s s' : vl mty mvl } (Eqty : forall x, (comp (var_ty nty) xity) x = sigmaty x) (Eqvl : forall x, (comp (var_vl nty nvl) xivl) x = sigmavl x) (Eqsvl : s = s') : ren_vl xity xivl s' = subst_vl sigmaty sigmavl s :=
have h : _, from rinst_inst_vl xity xivl sigmaty sigmavl Eqty Eqvl s, begin rw <- Eqsvl, assumption end

@[substLemmas]
lemma instId_tm { mty mvl : nat } : subst_tm (var_ty mty) (var_vl mty mvl) = id :=
 (λ x , idSubst_tm (var_ty mty) (var_vl mty mvl) (λ n , by refl) (λ n , by refl) (id x))

@[substLemmas]
lemma instId_vl { mty mvl : nat } : subst_vl (var_ty mty) (var_vl mty mvl) = id :=
 (λ x , idSubst_vl (var_ty mty) (var_vl mty mvl) (λ n , by refl) (λ n , by refl) (id x))

@[substLemmas]
lemma varL_vl { mty mvl : nat } { nty nvl : nat } (sigmaty : Fin mty -> ty nty) (sigmavl : Fin mvl -> vl nty nvl) : comp (subst_vl sigmaty sigmavl) (var_vl mty mvl) = sigmavl :=
 (λ x , by refl)

@[substLemmas]
lemma rinstInst_tm { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) : ren_tm xity xivl = subst_tm (comp (var_ty nty) xity) (comp (var_vl nty nvl) xivl) :=
 (λ x , rinst_inst_tm xity xivl (_) (_) (λ n , by refl) (λ n , by refl) x)

@[substLemmas]
lemma rinstInst_vl { mty mvl : nat } { nty nvl : nat } (xity : Fin mty -> Fin nty) (xivl : Fin mvl -> Fin nvl) : ren_vl xity xivl = subst_vl (comp (var_ty nty) xity) (comp (var_vl nty nvl) xivl) :=
 (λ x , rinst_inst_vl xity xivl (_) (_) (λ n , by refl) (λ n , by refl) x)

@[substLemmas]
lemma compComp_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) (s : tm mty mvl) : subst_tm tauty tauvl (subst_tm sigmaty sigmavl s) = subst_tm (comp (subst_ty tauty) sigmaty) (comp (subst_vl tauty tauvl) sigmavl) s :=
compSubstSubst_tm sigmaty sigmavl tauty tauvl (_) (_) (λ n , by refl) (λ n , by refl) s

@[substLemmas]
lemma compComp_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) (s : vl mty mvl) : subst_vl tauty tauvl (subst_vl sigmaty sigmavl s) = subst_vl (comp (subst_ty tauty) sigmaty) (comp (subst_vl tauty tauvl) sigmavl) s :=
compSubstSubst_vl sigmaty sigmavl tauty tauvl (_) (_) (λ n , by refl) (λ n , by refl) s

@[substLemmas]
lemma compComp'_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) : comp (subst_tm tauty tauvl) (subst_tm sigmaty sigmavl) = subst_tm (comp (subst_ty tauty) sigmaty) (comp (subst_vl tauty tauvl) sigmavl) :=
 (λ n , compComp_tm sigmaty sigmavl tauty tauvl n)

@[substLemmas]
lemma compComp'_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : Fin mty -> ty kty) (sigmavl : Fin mvl -> vl kty kvl) (tauty : Fin kty -> ty lty) (tauvl : Fin kvl -> vl lty lvl) : comp (subst_vl tauty tauvl) (subst_vl sigmaty sigmavl) = subst_vl (comp (subst_ty tauty) sigmaty) (comp (subst_vl tauty tauvl) sigmavl) :=
 (λ n , compComp_vl sigmaty sigmavl tauty tauvl n)

meta def is_app_tm : expr -> bool
  | `(app _ _ %%s0 %%s1) := tt
  | `(congr_app' %%s0 %%s1) := tt
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_app_tm : expr -> tactic (expr × expr)
  | `(app _ _ %%s0 %%s1) := return (s0, s1)
  | `(congr_app' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_app_tm Eqstm
  | e := return (e, e)

meta def is_tapp_tm : expr -> bool
  | `(tapp _ _ %%s0 %%s1) := tt
  | `(congr_tapp' %%s0 %%s1) := tt
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_tapp_tm : expr -> tactic (expr × expr)
  | `(tapp _ _ %%s0 %%s1) := return (s0, s1)
  | `(congr_tapp' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_tapp_tm Eqstm
  | e := return (e, e)

meta def is_vt_tm : expr -> bool
  | `(vt _ _ %%s0) := tt
  | `(congr_vt' %%s0) := tt
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_vt_tm : expr -> tactic expr
  | `(vt _ _ %%s0) := return s0
  | `(congr_vt' %%s0) := return s0
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_vt_tm Eqstm
  | e := return e

meta def is_lam_vl : expr -> bool
  | `(lam _ _ %%s0 %%s1) := tt
  | `(congr_lam' %%s0 %%s1) := tt
  | `(idSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | _ := ff

meta def destruct_lam_vl : expr -> tactic (expr × expr)
  | `(lam _ _ %%s0 %%s1) := return (s0, s1)
  | `(congr_lam' %%s0 %%s1) := return (s0, s1)
  | `(idSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_lam_vl Eqsvl
  | e := return (e, e)

meta def is_tlam_vl : expr -> bool
  | `(tlam _ _ %%s0) := tt
  | `(congr_tlam' %%s0) := tt
  | `(idSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | _ := ff

meta def destruct_tlam_vl : expr -> tactic expr
  | `(tlam _ _ %%s0) := return s0
  | `(congr_tlam' %%s0) := return s0
  | `(idSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_tlam_vl Eqsvl
  | e := return e

meta def is_InstRen_tm : expr -> bool
  | `(extRen'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | `(compRenRen'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_InstRen_tm : expr -> tactic (expr × expr)
  | `(extRen'_tm %%Eqty %%Eqvl %%Eqstm) := return (Eqty, Eqvl, Eqstm)
  | `(compRenRen'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_InstRen_tm Eqstm
  | `(%%e) := return (e,e)

meta def is_InstRen_vl : expr -> bool
  | `(extRen'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | `(compRenRen'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | _ := ff

meta def destruct_InstRen_vl : expr -> tactic (expr × expr)
  | `(extRen'_vl %%Eqty %%Eqvl %%Eqsvl) := return (Eqty, Eqvl, Eqsvl)
  | `(compRenRen'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_InstRen_vl Eqsvl
  | `(%%e) := return (e,e)

meta def is_Inst_tm : expr -> bool
  | `(compRenRen'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | `(compRenSubst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | `(compSubstRen'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | `(compSubstSubst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | `(rinst_inst'_tm %%Eqty %%Eqvl %%Eqstm) := tt
  | _ := ff

meta def destruct_Inst_tm : expr -> tactic (expr × expr)
  | `(ext'_tm %%Eqty %%Eqvl %%Eqstm) := return (Eqty, Eqvl, Eqstm)
  | `(compRenRen'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compRenSubst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstRen'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(compSubstSubst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(idSubst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(rinst_inst'_tm %%Eqty %%Eqvl %%Eqstm) := destruct_Inst_tm Eqstm
  | `(%%e) := return (e,e)

meta def is_Inst_vl : expr -> bool
  | `(compRenRen'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | `(compRenSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | `(compSubstRen'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | `(compSubstSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | `(idSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | `(rinst_inst'_vl %%Eqty %%Eqvl %%Eqsvl) := tt
  | _ := ff

meta def destruct_Inst_vl : expr -> tactic (expr × expr)
  | `(ext'_vl %%Eqty %%Eqvl %%Eqsvl) := return (Eqty, Eqvl, Eqsvl)
  | `(compRenRen'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_Inst_vl Eqsvl
  | `(compRenSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_Inst_vl Eqsvl
  | `(compSubstRen'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_Inst_vl Eqsvl
  | `(compSubstSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_Inst_vl Eqsvl
  | `(idSubst'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_Inst_vl Eqsvl
  | `(rinst_inst'_vl %%Eqty %%Eqvl %%Eqsvl) := destruct_Inst_vl Eqsvl
  | `(%%e) := return (e,e)

meta def is_scons : expr -> bool
  | `(up_ren_subst_ty_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_ren_vl_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_ren_vl_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_subst_ty_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(up_subst_subst_vl_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(up_subst_subst_vl_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(scons %%H1 %%H2) := tt
  | _ := ff

meta def is_comp : expr -> bool
  | `(up_ren_subst_ty_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_ren_vl_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_ren_vl_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_subst_ty_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(up_subst_subst_vl_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(up_subst_subst_vl_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(comp %%H1 %%H2) := tt
  | _ := ff

meta def is_var_tm : expr -> bool
  | `(up_ren_subst_ty_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_ren_vl_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_ren_vl_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_subst_ty_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(up_subst_subst_vl_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(up_subst_subst_vl_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(var_tm %%H1 %%H2) := tt
  | _ := ff

meta def is_var_vl : expr -> bool
  | `(up_ren_subst_ty_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_ty %%xi %%tau %%theta %%Eq) := tt
  | `(up_ren_subst_vl_vl %%xi %%tau %%theta %%Eq) := tt
  | `(up_subst_ren_ty_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_ren_vl_ty %%sigma %%zetaty %%theta %%Eq) := tt
  | `(up_subst_ren_vl_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := tt
  | `(up_subst_subst_ty_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(up_subst_subst_vl_ty %%sigma %%tauty %%theta %%Eq) := tt
  | `(up_subst_subst_vl_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := tt
  | `(var_vl %%H1 %%H2) := tt
  | _ := ff

meta def destruct_scons : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_vl %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_ren_subst_vl_ty %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_ren_subst_vl_vl %%xi %%tau %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_ty_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_vl_ty %%sigma %%zetaty %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_ren_vl_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_ty_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_vl_ty %%sigma %%tauty %%theta %%Eq) := destruct_scons Eq
  | `(up_subst_subst_vl_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := destruct_scons Eq
  | `(scons %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def destruct_comp : expr -> tactic (expr × expr)
  | `(up_ren_subst_ty_vl %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_ren_subst_vl_ty %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_ren_subst_vl_vl %%xi %%tau %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_ty_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_vl_ty %%sigma %%zetaty %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_ren_vl_vl %%sigma %%zetaty %%zetavl %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_ty_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_vl_ty %%sigma %%tauty %%theta %%Eq) := destruct_comp Eq
  | `(up_subst_subst_vl_vl %%sigma %%tauty %%tauvl %%theta %%Eq) := destruct_comp Eq
  | `(comp %%H1 %%H2) := return (H1, H2)
  | `(%%e) := return (e,e)

meta def normalize'_tm : nat -> expr -> expr -> tactic expr
  | n e1 e2 := if (n = Inst)
  then if (e2 = var_tm)
  then idSubst'_tm e1 e2
  else subst_tm e1 e2
  else if (n = Rinst)
  then if (e2 = var_tm)
  then idSubst'_tm e1 e2
  else if (is_App_tm e1)
  then TODO
  else subst_tm e1 e2
  else if (n = Comp)
  then TODO
  else if (n = Up)
  then TODO
  else ex ``(%%e1=%%e2)

meta def normalize'_vl : nat -> expr -> expr -> tactic expr
  | n e1 e2 := if (n = Inst)
  then if (e2 = var_vl)
  then idSubst'_vl e1 e2
  else subst_vl e1 e2
  else if (n = Rinst)
  then if (e2 = var_vl)
  then idSubst'_vl e1 e2
  else if (is_App_vl e1)
  then TODO
  else subst_vl e1 e2
  else if (n = Comp)
  then TODO
  else if (n = Up)
  then TODO
  else ex ``(%%e1=%%e2)
