Require Export fintype.

Inductive ty (nty : nat) : Type :=
  | var_ty  : fin nty -> ty nty
  | arr  : ty nty -> ty nty -> ty nty
  | all  : ty (S nty) -> ty nty.

Lemma congr_arr { mty : nat } { s0 : ty mty } { s1 : ty mty } { t0 : ty mty } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> arr mty s0 s1 = arr mty t0 t1 .
Proof. congruence. Qed.

Lemma congr_all { mty : nat } { s0 : ty (S mty) } { t0 : ty (S mty) } : s0 = t0 -> all mty s0 = all mty t0 .
Proof. congruence. Qed.

Definition upRen_ty_ty { m } { n } (xi : fin m -> fin n) : _ :=
  up_ren xi.

Fixpoint ren_ty { mty : nat } { nty : nat } (xity : fin mty -> fin nty) (s : ty mty) : _ :=
    match s with
    | var_ty s => (var_ty nty) (xity s)
    | arr s0 s1 => arr nty (ren_ty xity s0) (ren_ty xity s1)
    | all s0 => all nty (ren_ty (upRen_ty_ty xity) s0)
    end.

Definition up_ty_ty { m } { nty : nat } (sigma : fin m -> ty nty) : _ :=
  scons ((var_ty (S nty)) var_zero) (comp tt (ren_ty shift) sigma).

Fixpoint subst_ty { mty : nat } { nty : nat } (sigmaty : fin mty -> ty nty) (s : ty mty) : _ :=
    match s with
    | var_ty s => sigmaty s
    | arr s0 s1 => arr nty (subst_ty sigmaty s0) (subst_ty sigmaty s1)
    | all s0 => all nty (subst_ty (up_ty_ty sigmaty) s0)
    end.

Definition upId_ty_ty { mty : nat } (sigma : fin mty -> ty mty) (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_ty_ty sigma) x = (var_ty (S mty)) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint idSubst_ty { mty : nat } (sigmaty : fin mty -> ty mty) (Eqty : forall x, sigmaty x = (var_ty mty) x) (s : ty mty) : subst_ty sigmaty s = s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (idSubst_ty sigmaty Eqty s0) (idSubst_ty sigmaty Eqty s1)
    | all s0 => congr_all (idSubst_ty (up_ty_ty sigmaty) (upId_ty_ty (_) Eqty) s0)
    end.

Definition upExtRen_ty_ty { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_ty xi) x = (upRen_ty_ty zeta) x :=
  fun n => match n with
  | Some n => ap shift (Eq n)
  | None => eq_refl
  end.

Fixpoint extRen_ty { mty : nat } { nty : nat } (xity : fin mty -> fin nty) (zetaty : fin mty -> fin nty) (Eqty : forall x, xity x = zetaty x) (s : ty mty) : ren_ty xity s = ren_ty zetaty s :=
    match s with
    | var_ty s => ap (var_ty nty) (Eqty s)
    | arr s0 s1 => congr_arr (extRen_ty xity zetaty Eqty s0) (extRen_ty xity zetaty Eqty s1)
    | all s0 => congr_all (extRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upExtRen_ty_ty xity zetaty Eqty) s0)
    end.

Definition upExt_ty_ty { m } { nty : nat } (sigma : fin m -> ty nty) (tau : fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_ty sigma) x = (up_ty_ty tau) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint ext_ty { mty : nat } { nty : nat } (sigmaty : fin mty -> ty nty) (tauty : fin mty -> ty nty) (Eqty : forall x, sigmaty x = tauty x) (s : ty mty) : subst_ty sigmaty s = subst_ty tauty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (ext_ty sigmaty tauty Eqty s0) (ext_ty sigmaty tauty Eqty s1)
    | all s0 => congr_all (ext_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (upExt_ty_ty sigmaty tauty Eqty) s0)
    end.

Fixpoint compRenRen_ty { kty : nat } { lty : nat } { mty : nat } (xity : fin mty -> fin kty) (zetaty : fin kty -> fin lty) (rhoty : fin mty -> fin lty) (Eqty : forall x, (comp tt zetaty xity) x = rhoty x) (s : ty mty) : ren_ty zetaty (ren_ty xity s) = ren_ty rhoty s :=
    match s with
    | var_ty s => ap (var_ty lty) (Eqty s)
    | arr s0 s1 => congr_arr (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_ty xity zetaty rhoty Eqty s1)
    | all s0 => congr_all (compRenRen_ty (upRen_ty_ty xity) (upRen_ty_ty zetaty) (upRen_ty_ty rhoty) (up_ren_ren xity zetaty rhoty Eqty) s0)
    end.

Definition up_ren_subst_ty_ty { k } { l } { mty : nat } (xi : fin k -> fin l) (tau : fin l -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_ty_ty tau) (upRen_ty_ty xi)) x = (up_ty_ty theta) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint compRenSubst_ty { kty : nat } { lty : nat } { mty : nat } (xity : fin mty -> fin kty) (tauty : fin kty -> ty lty) (thetaty : fin mty -> ty lty) (Eqty : forall x, (comp tt tauty xity) x = thetaty x) (s : ty mty) : subst_ty tauty (ren_ty xity s) = subst_ty thetaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_ty xity tauty thetaty Eqty s1)
    | all s0 => congr_all (compRenSubst_ty (upRen_ty_ty xity) (up_ty_ty tauty) (up_ty_ty thetaty) (up_ren_subst_ty_ty xity tauty thetaty Eqty) s0)
    end.

Definition up_subst_ren_ty_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (zetaty : fin lty -> fin mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (ren_ty zetaty) sigma) x = theta x) : forall x, (comp tt (ren_ty (upRen_ty_ty zetaty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenRen_ty shift (upRen_ty_ty zetaty) (comp tt shift zetaty) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_ty zetaty shift (comp tt shift zetaty) (fun x => eq_refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstRen__ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : fin mty -> ty kty) (zetaty : fin kty -> fin lty) (thetaty : fin mty -> ty lty) (Eqty : forall x, (comp tt (ren_ty zetaty) sigmaty) x = thetaty x) (s : ty mty) : ren_ty zetaty (subst_ty sigmaty s) = subst_ty thetaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (compSubstRen__ty sigmaty zetaty thetaty Eqty s0) (compSubstRen__ty sigmaty zetaty thetaty Eqty s1)
    | all s0 => congr_all (compSubstRen__ty (up_ty_ty sigmaty) (upRen_ty_ty zetaty) (up_ty_ty thetaty) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) s0)
    end.

Definition up_subst_subst_ty_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (tauty : fin lty -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (subst_ty tauty) sigma) x = theta x) : forall x, (comp tt (subst_ty (up_ty_ty tauty)) (up_ty_ty sigma)) x = (up_ty_ty theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenSubst_ty shift (up_ty_ty tauty) (comp tt (up_ty_ty tauty) shift) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen__ty tauty shift (comp tt (ren_ty shift) tauty) (fun x => eq_refl) (sigma n))) (ap (ren_ty shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstSubst_ty { kty : nat } { lty : nat } { mty : nat } (sigmaty : fin mty -> ty kty) (tauty : fin kty -> ty lty) (thetaty : fin mty -> ty lty) (Eqty : forall x, (comp tt (subst_ty tauty) sigmaty) x = thetaty x) (s : ty mty) : subst_ty tauty (subst_ty sigmaty s) = subst_ty thetaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1)
    | all s0 => congr_all (compSubstSubst_ty (up_ty_ty sigmaty) (up_ty_ty tauty) (up_ty_ty thetaty) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) s0)
    end.

Definition rinstInst_up_ty_ty { m } { nty : nat } (xi : fin m -> fin nty) (sigma : fin m -> ty nty) (Eq : forall x, (comp tt (var_ty nty) xi) x = sigma x) : forall x, (comp tt (var_ty (S nty)) (upRen_ty_ty xi)) x = (up_ty_ty sigma) x :=
  fun n => match n with
  | Some n => ap (ren_ty shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint rinst_inst_ty { mty : nat } { nty : nat } (xity : fin mty -> fin nty) (sigmaty : fin mty -> ty nty) (Eqty : forall x, (comp tt (var_ty nty) xity) x = sigmaty x) (s : ty mty) : ren_ty xity s = subst_ty sigmaty s :=
    match s with
    | var_ty s => Eqty s
    | arr s0 s1 => congr_arr (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_ty xity sigmaty Eqty s1)
    | all s0 => congr_all (rinst_inst_ty (upRen_ty_ty xity) (up_ty_ty sigmaty) (rinstInst_up_ty_ty xity sigmaty Eqty) s0)
    end.

Inductive tm (nty nvl : nat) : Type :=
  | app  : tm nty nvl -> tm nty nvl -> tm nty nvl
  | tapp  : tm nty nvl -> ty nty -> tm nty nvl
  | vt  : vl nty nvl -> tm nty nvl
 with vl (nty nvl : nat) : Type :=
  | var_vl  : fin nvl -> vl nty nvl
  | lam  : ty nty -> tm nty (S nvl) -> vl nty nvl
  | tlam  : tm (S nty) nvl -> vl nty nvl.

Lemma congr_app { mty mvl : nat } { s0 : tm mty mvl } { s1 : tm mty mvl } { t0 : tm mty mvl } { t1 : tm mty mvl } : s0 = t0 -> s1 = t1 -> app mty mvl s0 s1 = app mty mvl t0 t1 .
Proof. congruence. Qed.

Lemma congr_tapp { mty mvl : nat } { s0 : tm mty mvl } { s1 : ty mty } { t0 : tm mty mvl } { t1 : ty mty } : s0 = t0 -> s1 = t1 -> tapp mty mvl s0 s1 = tapp mty mvl t0 t1 .
Proof. congruence. Qed.

Lemma congr_vt { mty mvl : nat } { s0 : vl mty mvl } { t0 : vl mty mvl } : s0 = t0 -> vt mty mvl s0 = vt mty mvl t0 .
Proof. congruence. Qed.

Lemma congr_lam { mty mvl : nat } { s0 : ty mty } { s1 : tm mty (S mvl) } { t0 : ty mty } { t1 : tm mty (S mvl) } : s0 = t0 -> s1 = t1 -> lam mty mvl s0 s1 = lam mty mvl t0 t1 .
Proof. congruence. Qed.

Lemma congr_tlam { mty mvl : nat } { s0 : tm (S mty) mvl } { t0 : tm (S mty) mvl } : s0 = t0 -> tlam mty mvl s0 = tlam mty mvl t0 .
Proof. congruence. Qed.

Definition upRen_ty_vl { m } { n } (xi : fin m -> fin n) : _ :=
  xi.

Definition upRen_vl_ty { m } { n } (xi : fin m -> fin n) : _ :=
  xi.

Definition upRen_vl_vl { m } { n } (xi : fin m -> fin n) : _ :=
  up_ren xi.

Fixpoint ren_tm { mty mvl : nat } { nty nvl : nat } (xity : fin mty -> fin nty) (xivl : fin mvl -> fin nvl) (s : tm mty mvl) : _ :=
    match s with
    | app s0 s1 => app nty nvl (ren_tm xity xivl s0) (ren_tm xity xivl s1)
    | tapp s0 s1 => tapp nty nvl (ren_tm xity xivl s0) (ren_ty xity s1)
    | vt s0 => vt nty nvl (ren_vl xity xivl s0)
    end
 with ren_vl { mty mvl : nat } { nty nvl : nat } (xity : fin mty -> fin nty) (xivl : fin mvl -> fin nvl) (s : vl mty mvl) : _ :=
    match s with
    | var_vl s => (var_vl nty nvl) (xivl s)
    | lam s0 s1 => lam nty nvl (ren_ty xity s0) (ren_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) s1)
    | tlam s0 => tlam nty nvl (ren_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) s0)
    end.

Definition up_ty_vl { m } { nty nvl : nat } (sigma : fin m -> vl nty nvl) : _ :=
  comp tt (ren_vl shift id) sigma.

Definition up_vl_ty { m } { nty : nat } (sigma : fin m -> ty nty) : _ :=
  comp tt (ren_ty id) sigma.

Definition up_vl_vl { m } { nty nvl : nat } (sigma : fin m -> vl nty nvl) : _ :=
  scons ((var_vl nty (S nvl)) var_zero) (comp tt (ren_vl id shift) sigma).

Fixpoint subst_tm { mty mvl : nat } { nty nvl : nat } (sigmaty : fin mty -> ty nty) (sigmavl : fin mvl -> vl nty nvl) (s : tm mty mvl) : _ :=
    match s with
    | app s0 s1 => app nty nvl (subst_tm sigmaty sigmavl s0) (subst_tm sigmaty sigmavl s1)
    | tapp s0 s1 => tapp nty nvl (subst_tm sigmaty sigmavl s0) (subst_ty sigmaty s1)
    | vt s0 => vt nty nvl (subst_vl sigmaty sigmavl s0)
    end
 with subst_vl { mty mvl : nat } { nty nvl : nat } (sigmaty : fin mty -> ty nty) (sigmavl : fin mvl -> vl nty nvl) (s : vl mty mvl) : _ :=
    match s with
    | var_vl s => sigmavl s
    | lam s0 s1 => lam nty nvl (subst_ty sigmaty s0) (subst_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) s1)
    | tlam s0 => tlam nty nvl (subst_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) s0)
    end.

Definition upId_ty_vl { mty mvl : nat } (sigma : fin mvl -> vl mty mvl) (Eq : forall x, sigma x = (var_vl mty mvl) x) : forall x, (up_ty_vl sigma) x = (var_vl (S mty) mvl) x :=
  fun n => ap (ren_vl shift id) (Eq n).

Definition upId_vl_ty { mty : nat } (sigma : fin mty -> ty mty) (Eq : forall x, sigma x = (var_ty mty) x) : forall x, (up_vl_ty sigma) x = (var_ty mty) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition upId_vl_vl { mty mvl : nat } (sigma : fin mvl -> vl mty mvl) (Eq : forall x, sigma x = (var_vl mty mvl) x) : forall x, (up_vl_vl sigma) x = (var_vl mty (S mvl)) x :=
  fun n => match n with
  | Some n => ap (ren_vl id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint idSubst_tm { mty mvl : nat } (sigmaty : fin mty -> ty mty) (sigmavl : fin mvl -> vl mty mvl) (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqvl : forall x, sigmavl x = (var_vl mty mvl) x) (s : tm mty mvl) : subst_tm sigmaty sigmavl s = s :=
    match s with
    | app s0 s1 => congr_app (idSubst_tm sigmaty sigmavl Eqty Eqvl s0) (idSubst_tm sigmaty sigmavl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (idSubst_tm sigmaty sigmavl Eqty Eqvl s0) (idSubst_ty sigmaty Eqty s1)
    | vt s0 => congr_vt (idSubst_vl sigmaty sigmavl Eqty Eqvl s0)
    end
 with idSubst_vl { mty mvl : nat } (sigmaty : fin mty -> ty mty) (sigmavl : fin mvl -> vl mty mvl) (Eqty : forall x, sigmaty x = (var_ty mty) x) (Eqvl : forall x, sigmavl x = (var_vl mty mvl) x) (s : vl mty mvl) : subst_vl sigmaty sigmavl s = s :=
    match s with
    | var_vl s => Eqvl s
    | lam s0 s1 => congr_lam (idSubst_ty sigmaty Eqty s0) (idSubst_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (upId_vl_ty (_) Eqty) (upId_vl_vl (_) Eqvl) s1)
    | tlam s0 => congr_tlam (idSubst_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (upId_ty_ty (_) Eqty) (upId_ty_vl (_) Eqvl) s0)
    end.

Definition upExtRen_ty_vl { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_ty_vl xi) x = (upRen_ty_vl zeta) x :=
  fun n => Eq n.

Definition upExtRen_vl_ty { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_vl_ty xi) x = (upRen_vl_ty zeta) x :=
  fun n => Eq n.

Definition upExtRen_vl_vl { m } { n } (xi : fin m -> fin n) (zeta : fin m -> fin n) (Eq : forall x, xi x = zeta x) : forall x, (upRen_vl_vl xi) x = (upRen_vl_vl zeta) x :=
  fun n => match n with
  | Some n => ap shift (Eq n)
  | None => eq_refl
  end.

Fixpoint extRen_tm { mty mvl : nat } { nty nvl : nat } (xity : fin mty -> fin nty) (xivl : fin mvl -> fin nvl) (zetaty : fin mty -> fin nty) (zetavl : fin mvl -> fin nvl) (Eqty : forall x, xity x = zetaty x) (Eqvl : forall x, xivl x = zetavl x) (s : tm mty mvl) : ren_tm xity xivl s = ren_tm zetaty zetavl s :=
    match s with
    | app s0 s1 => congr_app (extRen_tm xity xivl zetaty zetavl Eqty Eqvl s0) (extRen_tm xity xivl zetaty zetavl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (extRen_tm xity xivl zetaty zetavl Eqty Eqvl s0) (extRen_ty xity zetaty Eqty s1)
    | vt s0 => congr_vt (extRen_vl xity xivl zetaty zetavl Eqty Eqvl s0)
    end
 with extRen_vl { mty mvl : nat } { nty nvl : nat } (xity : fin mty -> fin nty) (xivl : fin mvl -> fin nvl) (zetaty : fin mty -> fin nty) (zetavl : fin mvl -> fin nvl) (Eqty : forall x, xity x = zetaty x) (Eqvl : forall x, xivl x = zetavl x) (s : vl mty mvl) : ren_vl xity xivl s = ren_vl zetaty zetavl s :=
    match s with
    | var_vl s => ap (var_vl nty nvl) (Eqvl s)
    | lam s0 s1 => congr_lam (extRen_ty xity zetaty Eqty s0) (extRen_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (upExtRen_vl_ty xity zetaty Eqty) (upExtRen_vl_vl xivl zetavl Eqvl) s1)
    | tlam s0 => congr_tlam (extRen_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (upExtRen_ty_ty xity zetaty Eqty) (upExtRen_ty_vl xivl zetavl Eqvl) s0)
    end.

Definition upExt_ty_vl { m } { nty nvl : nat } (sigma : fin m -> vl nty nvl) (tau : fin m -> vl nty nvl) (Eq : forall x, sigma x = tau x) : forall x, (up_ty_vl sigma) x = (up_ty_vl tau) x :=
  fun n => ap (ren_vl shift id) (Eq n).

Definition upExt_vl_ty { m } { nty : nat } (sigma : fin m -> ty nty) (tau : fin m -> ty nty) (Eq : forall x, sigma x = tau x) : forall x, (up_vl_ty sigma) x = (up_vl_ty tau) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition upExt_vl_vl { m } { nty nvl : nat } (sigma : fin m -> vl nty nvl) (tau : fin m -> vl nty nvl) (Eq : forall x, sigma x = tau x) : forall x, (up_vl_vl sigma) x = (up_vl_vl tau) x :=
  fun n => match n with
  | Some n => ap (ren_vl id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint ext_tm { mty mvl : nat } { nty nvl : nat } (sigmaty : fin mty -> ty nty) (sigmavl : fin mvl -> vl nty nvl) (tauty : fin mty -> ty nty) (tauvl : fin mvl -> vl nty nvl) (Eqty : forall x, sigmaty x = tauty x) (Eqvl : forall x, sigmavl x = tauvl x) (s : tm mty mvl) : subst_tm sigmaty sigmavl s = subst_tm tauty tauvl s :=
    match s with
    | app s0 s1 => congr_app (ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s0) (ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (ext_tm sigmaty sigmavl tauty tauvl Eqty Eqvl s0) (ext_ty sigmaty tauty Eqty s1)
    | vt s0 => congr_vt (ext_vl sigmaty sigmavl tauty tauvl Eqty Eqvl s0)
    end
 with ext_vl { mty mvl : nat } { nty nvl : nat } (sigmaty : fin mty -> ty nty) (sigmavl : fin mvl -> vl nty nvl) (tauty : fin mty -> ty nty) (tauvl : fin mvl -> vl nty nvl) (Eqty : forall x, sigmaty x = tauty x) (Eqvl : forall x, sigmavl x = tauvl x) (s : vl mty mvl) : subst_vl sigmaty sigmavl s = subst_vl tauty tauvl s :=
    match s with
    | var_vl s => Eqvl s
    | lam s0 s1 => congr_lam (ext_ty sigmaty tauty Eqty s0) (ext_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (up_vl_ty tauty) (up_vl_vl tauvl) (upExt_vl_ty sigmaty tauty Eqty) (upExt_vl_vl sigmavl tauvl Eqvl) s1)
    | tlam s0 => congr_tlam (ext_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (up_ty_ty tauty) (up_ty_vl tauvl) (upExt_ty_ty sigmaty tauty Eqty) (upExt_ty_vl sigmavl tauvl Eqvl) s0)
    end.

Fixpoint compRenRen_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : fin mty -> fin kty) (xivl : fin mvl -> fin kvl) (zetaty : fin kty -> fin lty) (zetavl : fin kvl -> fin lvl) (rhoty : fin mty -> fin lty) (rhovl : fin mvl -> fin lvl) (Eqty : forall x, (comp tt zetaty xity) x = rhoty x) (Eqvl : forall x, (comp tt zetavl xivl) x = rhovl x) (s : tm mty mvl) : ren_tm zetaty zetavl (ren_tm xity xivl s) = ren_tm rhoty rhovl s :=
    match s with
    | app s0 s1 => congr_app (compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s0) (compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (compRenRen_tm xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s0) (compRenRen_ty xity zetaty rhoty Eqty s1)
    | vt s0 => congr_vt (compRenRen_vl xity xivl zetaty zetavl rhoty rhovl Eqty Eqvl s0)
    end
 with compRenRen_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : fin mty -> fin kty) (xivl : fin mvl -> fin kvl) (zetaty : fin kty -> fin lty) (zetavl : fin kvl -> fin lvl) (rhoty : fin mty -> fin lty) (rhovl : fin mvl -> fin lvl) (Eqty : forall x, (comp tt zetaty xity) x = rhoty x) (Eqvl : forall x, (comp tt zetavl xivl) x = rhovl x) (s : vl mty mvl) : ren_vl zetaty zetavl (ren_vl xity xivl s) = ren_vl rhoty rhovl s :=
    match s with
    | var_vl s => ap (var_vl lty lvl) (Eqvl s)
    | lam s0 s1 => congr_lam (compRenRen_ty xity zetaty rhoty Eqty s0) (compRenRen_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (upRen_vl_ty rhoty) (upRen_vl_vl rhovl) Eqty (up_ren_ren xivl zetavl rhovl Eqvl) s1)
    | tlam s0 => congr_tlam (compRenRen_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (upRen_ty_ty rhoty) (upRen_ty_vl rhovl) (up_ren_ren xity zetaty rhoty Eqty) Eqvl s0)
    end.

Definition up_ren_subst_ty_vl { k } { l } { mty mvl : nat } (xi : fin k -> fin l) (tau : fin l -> vl mty mvl) (theta : fin k -> vl mty mvl) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_ty_vl tau) (upRen_ty_vl xi)) x = (up_ty_vl theta) x :=
  fun n => ap (ren_vl shift id) (Eq n).

Definition up_ren_subst_vl_ty { k } { l } { mty : nat } (xi : fin k -> fin l) (tau : fin l -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_vl_ty tau) (upRen_vl_ty xi)) x = (up_vl_ty theta) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition up_ren_subst_vl_vl { k } { l } { mty mvl : nat } (xi : fin k -> fin l) (tau : fin l -> vl mty mvl) (theta : fin k -> vl mty mvl) (Eq : forall x, (comp tt tau xi) x = theta x) : forall x, (comp tt (up_vl_vl tau) (upRen_vl_vl xi)) x = (up_vl_vl theta) x :=
  fun n => match n with
  | Some n => ap (ren_vl id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint compRenSubst_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : fin mty -> fin kty) (xivl : fin mvl -> fin kvl) (tauty : fin kty -> ty lty) (tauvl : fin kvl -> vl lty lvl) (thetaty : fin mty -> ty lty) (thetavl : fin mvl -> vl lty lvl) (Eqty : forall x, (comp tt tauty xity) x = thetaty x) (Eqvl : forall x, (comp tt tauvl xivl) x = thetavl x) (s : tm mty mvl) : subst_tm tauty tauvl (ren_tm xity xivl s) = subst_tm thetaty thetavl s :=
    match s with
    | app s0 s1 => congr_app (compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (compRenSubst_tm xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compRenSubst_ty xity tauty thetaty Eqty s1)
    | vt s0 => congr_vt (compRenSubst_vl xity xivl tauty tauvl thetaty thetavl Eqty Eqvl s0)
    end
 with compRenSubst_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (xity : fin mty -> fin kty) (xivl : fin mvl -> fin kvl) (tauty : fin kty -> ty lty) (tauvl : fin kvl -> vl lty lvl) (thetaty : fin mty -> ty lty) (thetavl : fin mvl -> vl lty lvl) (Eqty : forall x, (comp tt tauty xity) x = thetaty x) (Eqvl : forall x, (comp tt tauvl xivl) x = thetavl x) (s : vl mty mvl) : subst_vl tauty tauvl (ren_vl xity xivl s) = subst_vl thetaty thetavl s :=
    match s with
    | var_vl s => Eqvl s
    | lam s0 s1 => congr_lam (compRenSubst_ty xity tauty thetaty Eqty s0) (compRenSubst_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (up_vl_ty tauty) (up_vl_vl tauvl) (up_vl_ty thetaty) (up_vl_vl thetavl) (up_ren_subst_vl_ty xity tauty thetaty Eqty) (up_ren_subst_vl_vl xivl tauvl thetavl Eqvl) s1)
    | tlam s0 => congr_tlam (compRenSubst_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (up_ty_ty tauty) (up_ty_vl tauvl) (up_ty_ty thetaty) (up_ty_vl thetavl) (up_ren_subst_ty_ty xity tauty thetaty Eqty) (up_ren_subst_ty_vl xivl tauvl thetavl Eqvl) s0)
    end.

Definition up_subst_ren_ty_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : fin k -> vl lty lvl) (zetaty : fin lty -> fin mty) (zetavl : fin lvl -> fin mvl) (theta : fin k -> vl mty mvl) (Eq : forall x, (comp tt (ren_vl zetaty zetavl) sigma) x = theta x) : forall x, (comp tt (ren_vl (upRen_ty_ty zetaty) (upRen_ty_vl zetavl)) (up_ty_vl sigma)) x = (up_ty_vl theta) x :=
  fun n => eq_trans (compRenRen_vl shift id (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (comp tt shift zetaty) (comp tt id zetavl) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_vl zetaty zetavl shift id (comp tt shift zetaty) (comp tt id zetavl) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_vl shift id) (Eq n))).

Definition up_subst_ren_vl_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (zetaty : fin lty -> fin mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (ren_ty zetaty) sigma) x = theta x) : forall x, (comp tt (ren_ty (upRen_vl_ty zetaty)) (up_vl_ty sigma)) x = (up_vl_ty theta) x :=
  fun n => eq_trans (compRenRen_ty id (upRen_vl_ty zetaty) (comp tt id zetaty) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_ty zetaty id (comp tt id zetaty) (fun x => eq_refl) (sigma n))) (ap (ren_ty id) (Eq n))).

Definition up_subst_ren_vl_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : fin k -> vl lty lvl) (zetaty : fin lty -> fin mty) (zetavl : fin lvl -> fin mvl) (theta : fin k -> vl mty mvl) (Eq : forall x, (comp tt (ren_vl zetaty zetavl) sigma) x = theta x) : forall x, (comp tt (ren_vl (upRen_vl_ty zetaty) (upRen_vl_vl zetavl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenRen_vl id shift (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (comp tt id zetaty) (comp tt shift zetavl) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compRenRen_vl zetaty zetavl id shift (comp tt id zetaty) (comp tt shift zetavl) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_vl id shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstRen__tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : fin mty -> ty kty) (sigmavl : fin mvl -> vl kty kvl) (zetaty : fin kty -> fin lty) (zetavl : fin kvl -> fin lvl) (thetaty : fin mty -> ty lty) (thetavl : fin mvl -> vl lty lvl) (Eqty : forall x, (comp tt (ren_ty zetaty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp tt (ren_vl zetaty zetavl) sigmavl) x = thetavl x) (s : tm mty mvl) : ren_tm zetaty zetavl (subst_tm sigmaty sigmavl s) = subst_tm thetaty thetavl s :=
    match s with
    | app s0 s1 => congr_app (compSubstRen__tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s0) (compSubstRen__tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (compSubstRen__tm sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s0) (compSubstRen__ty sigmaty zetaty thetaty Eqty s1)
    | vt s0 => congr_vt (compSubstRen__vl sigmaty sigmavl zetaty zetavl thetaty thetavl Eqty Eqvl s0)
    end
 with compSubstRen__vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : fin mty -> ty kty) (sigmavl : fin mvl -> vl kty kvl) (zetaty : fin kty -> fin lty) (zetavl : fin kvl -> fin lvl) (thetaty : fin mty -> ty lty) (thetavl : fin mvl -> vl lty lvl) (Eqty : forall x, (comp tt (ren_ty zetaty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp tt (ren_vl zetaty zetavl) sigmavl) x = thetavl x) (s : vl mty mvl) : ren_vl zetaty zetavl (subst_vl sigmaty sigmavl s) = subst_vl thetaty thetavl s :=
    match s with
    | var_vl s => Eqvl s
    | lam s0 s1 => congr_lam (compSubstRen__ty sigmaty zetaty thetaty Eqty s0) (compSubstRen__tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (upRen_vl_ty zetaty) (upRen_vl_vl zetavl) (up_vl_ty thetaty) (up_vl_vl thetavl) (up_subst_ren_vl_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_vl_vl sigmavl zetaty zetavl thetavl Eqvl) s1)
    | tlam s0 => congr_tlam (compSubstRen__tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (upRen_ty_ty zetaty) (upRen_ty_vl zetavl) (up_ty_ty thetaty) (up_ty_vl thetavl) (up_subst_ren_ty_ty sigmaty zetaty thetaty Eqty) (up_subst_ren_ty_vl sigmavl zetaty zetavl thetavl Eqvl) s0)
    end.

Definition up_subst_subst_ty_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : fin k -> vl lty lvl) (tauty : fin lty -> ty mty) (tauvl : fin lvl -> vl mty mvl) (theta : fin k -> vl mty mvl) (Eq : forall x, (comp tt (subst_vl tauty tauvl) sigma) x = theta x) : forall x, (comp tt (subst_vl (up_ty_ty tauty) (up_ty_vl tauvl)) (up_ty_vl sigma)) x = (up_ty_vl theta) x :=
  fun n => eq_trans (compRenSubst_vl shift id (up_ty_ty tauty) (up_ty_vl tauvl) (comp tt (up_ty_ty tauty) shift) (comp tt (up_ty_vl tauvl) id) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen__vl tauty tauvl shift id (comp tt (ren_ty shift) tauty) (comp tt (ren_vl shift id) tauvl) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_vl shift id) (Eq n))).

Definition up_subst_subst_vl_ty { k } { lty : nat } { mty : nat } (sigma : fin k -> ty lty) (tauty : fin lty -> ty mty) (theta : fin k -> ty mty) (Eq : forall x, (comp tt (subst_ty tauty) sigma) x = theta x) : forall x, (comp tt (subst_ty (up_vl_ty tauty)) (up_vl_ty sigma)) x = (up_vl_ty theta) x :=
  fun n => eq_trans (compRenSubst_ty id (up_vl_ty tauty) (comp tt (up_vl_ty tauty) id) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen__ty tauty id (comp tt (ren_ty id) tauty) (fun x => eq_refl) (sigma n))) (ap (ren_ty id) (Eq n))).

Definition up_subst_subst_vl_vl { k } { lty lvl : nat } { mty mvl : nat } (sigma : fin k -> vl lty lvl) (tauty : fin lty -> ty mty) (tauvl : fin lvl -> vl mty mvl) (theta : fin k -> vl mty mvl) (Eq : forall x, (comp tt (subst_vl tauty tauvl) sigma) x = theta x) : forall x, (comp tt (subst_vl (up_vl_ty tauty) (up_vl_vl tauvl)) (up_vl_vl sigma)) x = (up_vl_vl theta) x :=
  fun n => match n with
  | Some n => eq_trans (compRenSubst_vl id shift (up_vl_ty tauty) (up_vl_vl tauvl) (comp tt (up_vl_ty tauty) id) (comp tt (up_vl_vl tauvl) shift) (fun x => eq_refl) (fun x => eq_refl) (sigma n)) (eq_trans (eq_sym (compSubstRen__vl tauty tauvl id shift (comp tt (ren_ty id) tauty) (comp tt (ren_vl id shift) tauvl) (fun x => eq_refl) (fun x => eq_refl) (sigma n))) (ap (ren_vl id shift) (Eq n)))
  | None => eq_refl
  end.

Fixpoint compSubstSubst_tm { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : fin mty -> ty kty) (sigmavl : fin mvl -> vl kty kvl) (tauty : fin kty -> ty lty) (tauvl : fin kvl -> vl lty lvl) (thetaty : fin mty -> ty lty) (thetavl : fin mvl -> vl lty lvl) (Eqty : forall x, (comp tt (subst_ty tauty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp tt (subst_vl tauty tauvl) sigmavl) x = thetavl x) (s : tm mty mvl) : subst_tm tauty tauvl (subst_tm sigmaty sigmavl s) = subst_tm thetaty thetavl s :=
    match s with
    | app s0 s1 => congr_app (compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (compSubstSubst_tm sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s0) (compSubstSubst_ty sigmaty tauty thetaty Eqty s1)
    | vt s0 => congr_vt (compSubstSubst_vl sigmaty sigmavl tauty tauvl thetaty thetavl Eqty Eqvl s0)
    end
 with compSubstSubst_vl { kty kvl : nat } { lty lvl : nat } { mty mvl : nat } (sigmaty : fin mty -> ty kty) (sigmavl : fin mvl -> vl kty kvl) (tauty : fin kty -> ty lty) (tauvl : fin kvl -> vl lty lvl) (thetaty : fin mty -> ty lty) (thetavl : fin mvl -> vl lty lvl) (Eqty : forall x, (comp tt (subst_ty tauty) sigmaty) x = thetaty x) (Eqvl : forall x, (comp tt (subst_vl tauty tauvl) sigmavl) x = thetavl x) (s : vl mty mvl) : subst_vl tauty tauvl (subst_vl sigmaty sigmavl s) = subst_vl thetaty thetavl s :=
    match s with
    | var_vl s => Eqvl s
    | lam s0 s1 => congr_lam (compSubstSubst_ty sigmaty tauty thetaty Eqty s0) (compSubstSubst_tm (up_vl_ty sigmaty) (up_vl_vl sigmavl) (up_vl_ty tauty) (up_vl_vl tauvl) (up_vl_ty thetaty) (up_vl_vl thetavl) (up_subst_subst_vl_ty sigmaty tauty thetaty Eqty) (up_subst_subst_vl_vl sigmavl tauty tauvl thetavl Eqvl) s1)
    | tlam s0 => congr_tlam (compSubstSubst_tm (up_ty_ty sigmaty) (up_ty_vl sigmavl) (up_ty_ty tauty) (up_ty_vl tauvl) (up_ty_ty thetaty) (up_ty_vl thetavl) (up_subst_subst_ty_ty sigmaty tauty thetaty Eqty) (up_subst_subst_ty_vl sigmavl tauty tauvl thetavl Eqvl) s0)
    end.

Definition rinstInst_up_ty_vl { m } { nty nvl : nat } (xi : fin m -> fin nvl) (sigma : fin m -> vl nty nvl) (Eq : forall x, (comp tt (var_vl nty nvl) xi) x = sigma x) : forall x, (comp tt (var_vl (S nty) nvl) (upRen_ty_vl xi)) x = (up_ty_vl sigma) x :=
  fun n => ap (ren_vl shift id) (Eq n).

Definition rinstInst_up_vl_ty { m } { nty : nat } (xi : fin m -> fin nty) (sigma : fin m -> ty nty) (Eq : forall x, (comp tt (var_ty nty) xi) x = sigma x) : forall x, (comp tt (var_ty nty) (upRen_vl_ty xi)) x = (up_vl_ty sigma) x :=
  fun n => ap (ren_ty id) (Eq n).

Definition rinstInst_up_vl_vl { m } { nty nvl : nat } (xi : fin m -> fin nvl) (sigma : fin m -> vl nty nvl) (Eq : forall x, (comp tt (var_vl nty nvl) xi) x = sigma x) : forall x, (comp tt (var_vl nty (S nvl)) (upRen_vl_vl xi)) x = (up_vl_vl sigma) x :=
  fun n => match n with
  | Some n => ap (ren_vl id shift) (Eq n)
  | None => eq_refl
  end.

Fixpoint rinst_inst_tm { mty mvl : nat } { nty nvl : nat } (xity : fin mty -> fin nty) (xivl : fin mvl -> fin nvl) (sigmaty : fin mty -> ty nty) (sigmavl : fin mvl -> vl nty nvl) (Eqty : forall x, (comp tt (var_ty nty) xity) x = sigmaty x) (Eqvl : forall x, (comp tt (var_vl nty nvl) xivl) x = sigmavl x) (s : tm mty mvl) : ren_tm xity xivl s = subst_tm sigmaty sigmavl s :=
    match s with
    | app s0 s1 => congr_app (rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s0) (rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s1)
    | tapp s0 s1 => congr_tapp (rinst_inst_tm xity xivl sigmaty sigmavl Eqty Eqvl s0) (rinst_inst_ty xity sigmaty Eqty s1)
    | vt s0 => congr_vt (rinst_inst_vl xity xivl sigmaty sigmavl Eqty Eqvl s0)
    end
 with rinst_inst_vl { mty mvl : nat } { nty nvl : nat } (xity : fin mty -> fin nty) (xivl : fin mvl -> fin nvl) (sigmaty : fin mty -> ty nty) (sigmavl : fin mvl -> vl nty nvl) (Eqty : forall x, (comp tt (var_ty nty) xity) x = sigmaty x) (Eqvl : forall x, (comp tt (var_vl nty nvl) xivl) x = sigmavl x) (s : vl mty mvl) : ren_vl xity xivl s = subst_vl sigmaty sigmavl s :=
    match s with
    | var_vl s => Eqvl s
    | lam s0 s1 => congr_lam (rinst_inst_ty xity sigmaty Eqty s0) (rinst_inst_tm (upRen_vl_ty xity) (upRen_vl_vl xivl) (up_vl_ty sigmaty) (up_vl_vl sigmavl) (rinstInst_up_vl_ty xity sigmaty Eqty) (rinstInst_up_vl_vl xivl sigmavl Eqvl) s1)
    | tlam s0 => congr_tlam (rinst_inst_tm (upRen_ty_ty xity) (upRen_ty_vl xivl) (up_ty_ty sigmaty) (up_ty_vl sigmavl) (rinstInst_up_ty_ty xity sigmaty Eqty) (rinstInst_up_ty_vl xivl sigmavl Eqvl) s0)
    end.
