import .utlc

meta def normalize_expr : expr → tactic expr
| t := do ty ← tactic.infer_type t, match ty with 
  |`(tm) := 
    if (is_app_tm t) then 
    do (e1,e2) ← destruct_app_tm t,
    n1 ← normalize_expr e1, n2 ← normalize_expr e2, 
    ex ``(congr_app' %%e1 %%n1 %%e2 %%n2) else

    if (is_lam_tm t) then 
    do e ← destruct_lam_tm t, n ← normalize_expr e,
    ex ``(congr_lam' %%e %%n) else

    if (is_ren_tm t) then 
    do (e1,e2) ← destruct_ren_tm t,
    n1 ← normalize_expr e1, n2 ← normalize_expr e2, 
    normalize_tm Rinst n1 n2 else

    if (is_subst_tm t) then 
    do (e1,e2) ← destruct_subst_tm t,
    n1 ← normalize_expr e1, n2 ← normalize_expr e2, 
    normalize_tm Inst n1 n2 else   

     ex ``(eq.refl %%t)

  |`(Fin %%m -> Fin %%n) := 

    if (is_scons t) then 
    do (e1,e2) ← destruct_scons t,
    n1 ← normalize_expr e1, n2 ← normalize_expr e2, 
    normalize_tm Cons n1 n2 else
    
    if (is_comp t) then 
    do (e1,e2) ← destruct_comp t,
    n1 ← normalize_expr e1, n2 ← normalize_expr e2, 
    normalize_tm Comp n1 n2 else
  
    ex ``(eq.refl %%t)

  | _ := do tactic.trace (ty), ex ``(0) end


def f : Fin 1 → Fin 2 := some
def g : Fin 1 → Fin 2 := f

def h : nat → Prop := λ n, true
variable (n:nat)
def some_prop (f : Fin 1 → Fin 2) : Prop := true
open expr

#check pi
meta def simplify_function: expr → tactic expr 
| (app f a) := 
  do tyf ← tactic.infer_type f, tya ← tactic.infer_type a,
  match tya with 
    | `(tm) :=  normalize_expr a 
    | `(Fin %%k → Fin %%l) :=  normalize_expr a
    |  _ := match tyf with 
    | `(tm) :=  normalize_expr f 
    | `(Fin %%k → Fin %%l) := do n ← normalize_expr f,  ex ``((forall x, app %%n x ) %%a) -- TODO
    | _ := tactic.failed end end | t := tactic.failed 
--| (pi x _ _ eq) := do tactic.trace(x), tactic.trace(eq), tactic.skip


meta def testing  : tactic unit := 
do t ← tactic.target, match t with 
|`(%%t = %%d) := 
  (do t ← simplify_function t, e1 ← normalize_expr t, d ← simplify_function d, e2 ← normalize_expr d,
 tactic.refine ```(%%e1)) <|>
  do e1 ← normalize_expr t,  e2 ← normalize_expr d,
  tactic.trace ("left: "), tactic.trace(e1), 
  tactic.trace ("right: "), tactic.trace(e2), 
  tactic.refine ```(%%e1) <|> tactic.failed
 | t := do t ← simplify_function t, tactic.skip end

lemma test1 : f = g := begin testing, end
lemma test2 (n: Fin 1): (g n) = (f n) := begin testing, end
lemma test' : some_prop f := begin testing, trivial end

meta def simplify_eq : tactic unit :=
do t ← tactic.target, match t with 
|`(%%t = %%d) := 
  do e1 ← normalize_expr t, 
  e2 ← normalize_expr d,
  tactic.trace ("left: "), tactic.trace(e1), 
  tactic.trace ("right: "), tactic.trace(e2), 
  tactic.refine ```(%%e1) <|> tactic.failed
 | t := tactic.skip end
